/* drivers/barcelona/bat/bat.c
 *
 * Implementation of the battery driver.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Author: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* Includes */
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/hardware/clock.h>
#include <barcelona/Barc_Battery.h>
#include <barcelona/Barc_adc.h>
#include <barcelona/debug.h>
#include <barcelona/gopins.h>
#include <barcelona/gotype.h>
#include <barcelona/goboard.h>
#include <barcelona/usbmode.h>

/* defines */
#define PFX "bat: "
#define PK_DBG PK_DBG_FUNC

#define SAMPLES_PER_SECOND			100		/* samples per second */

static int bat_disabled;

static char g_current_charge_mode;
static char g_chargeAccordingTemperature;
static char g_chargeAccordingUSB;
static char g_chargeAccordingVoltage;

#define BATTERY_AVERAGE_SIZE		50
#define REF_ADC				1225
#define ADC_MARIGOT_DIVIDER		9
#define REF_MARIGOT			(REF_ADC * ADC_MARIGOT_DIVIDER)
#define	REF_OTHER			2450	/* bit of a magic number eh? */

#define	VOLTAGE_3V			3000	/* minimal voltage for battery charging */
#define	VOLTAGE_60_MINUTE_DELAY		(60 * SAMPLES_PER_SECOND * 60)	/* 60 minutes */
#define	VOLTAGE_ADC_BUF_SIZE		16		/* samples to use for average calculation */
static int undervoltageWait = 0;			/* start checking after 60 minutes of consecutive charging */

static unsigned short batteryVoltageBuffer[BATTERY_AVERAGE_SIZE];
static unsigned short batteryCurrentBuffer[BATTERY_AVERAGE_SIZE];
static int batteryIndex = 0;
static int batteryWait = BATTERY_AVERAGE_SIZE;
static int batteryVoltageSum = 0;
static int batteryCurrentSum = 0;
static int batteryCurrentLimit = 0;
static int ref_adc_gotype = REF_OTHER;
static int voltageSource = VOLTAGE_SOURCE_BATTERY;

#define TEMP_20MINUTES_DELAY		(20 * 60 * SAMPLES_PER_SECOND)
#define TEMP_INITIAL_DELAY
#define TEMP_ADC_MINUS10_DEGREES	1023	/* adc value for minus 10 degrees centigrade DEFAULT:off */
#define TEMP_ADC_0_DEGREES		1023	/* adc value for 0 degrees centigrade DEFAULT:off */
#define TEMP_ADC_MINUS10_DEGREES_B1	762	/* adc value for minus 10 degrees centigrade Dublin B1 sample */
#define TEMP_ADC_0_DEGREES_B1		687	/* adc value for 0 degrees centigrade Dublin B1 sample */
#define TEMP_ADC_MINUS10_DEGREES_B2	985	/* adc value for minus 10 degrees centigrade Dublin B2 sample */
#define TEMP_ADC_0_DEGREES_B2		910	/* adc value for 0 degrees centigrade Dublin B2 sample */

#define TEMP_ADC_BUF_SIZE			16		/* samples to use for average calculation */
static unsigned short tempSampleBuffer[TEMP_ADC_BUF_SIZE];
static int tempIndex = 0;			/* current sampled value */
static int tempSum = 0;				/* current running average */
static int tempWait = TEMP_ADC_BUF_SIZE;	/* wait for a filled buffer begore calculating average */
static int temp_adc_minus10_degrees = 0;	/* adc level at -10 degrees, depending on type of hardware */
static int temp_adc_0_degrees = 0;		/* adc level at 0 degrees, depending on type of hardware */

#define CHARGE_NONE	0x00
#define	CHARGE_ENABLE	0x01
#define	CHARGE_HIGH	0x02
#define	CHARGE_DONTCARE	(CHARGE_ENABLE | CHARGE_HIGH)
#define	CHARGE_EVALUATE	0x80

#define	LOGGING_INTERVAL (100)	// 1000 is 10 seconds
/* Just comment out the line below to enable battery logging */
#undef	LOGGING_INTERVAL

#if defined(LOGGING_INTERVAL)
#define DEBUG_MSG
int	loggingTime = LOGGING_INTERVAL;
#endif	/* LOGGING_INTERVAL */

#ifdef DEBUG_MSG
	#define DBG_MSG(fmt, arg...) printk( fmt, ##arg )
#else
	#define DBG_MSG(msg...) { }
#endif

#define GOBOARD_DUBLIN_SUBID ((IO_GetSubId()<<16) | GOBOARD_DUBLIN )

#if defined(LOGGING_INTERVAL)
/**
 * Calculate pnd temperature from ADC value of TEMP_SENS, for Dublin only !!
 * Temperature line is NOT linear, levels for -10 and 0 degrees are fixed ADC-values
 **/
int calculate_pnd_temperature(int adc )
{
	int slope  = 0;
	int offset = 0;
	if (IO_GetModelId() == GOTYPE_DUBLIN)
	{
		switch(GOBOARD_DUBLIN_SUBID)
		{
			/* Temperature = slope x ADC + offset
			 * slope and offset are multiplied by 1000 for integer sake
			 */
			case GOBOARD_DUBLIN_B2:	//B2
			case GOBOARD_DUBLIN_C1:	//C1  
			case GOBOARD_DUBLIN_C2:	//C2
				slope  = -179;
				offset = 164473;
			break;
			case 0xFFFF:
			case GOBOARD_DUBLIN_B1:	//A and B1
				slope  = -135;
				offset = 93047;
			break;
			default:
			break;
		}
	}	
	return	(((adc * slope ) + offset )/1000 );
}
#endif

/**
 * In fact there are 4 charge levels when LTC3455 Charger is used:
 * Disabled	:USB_SUSPEND_OUT set active, USB_PWR_BYPASS set inactive
 * Very low	:USB_HP set inactive, USB_SUSPEND_OUT set inactive: charge-current is limited to 100 mA
 * Low		:USB_HP set active: charge-current is limited to 500 mA
 * High		:USB_PWR_BYPASS set active, charge-current is limited by hardware
 * Keep in mind that the charge-current depends on the pnd-load, the charge-current is reduced when the load is high !
 **/
 
 void bat_set_charge_disable(void) {
	// Disable charging :
	DBG_MSG("%s\n",__FUNCTION__);
	IO_Deactivate(USB_PWR_BYPASS);
	IO_Deactivate(USB_HP);
	IO_Activate(USB_SUSPEND_OUT);
}

void bat_set_charge_very_low(void)
{
	// Very Low power charge
	DBG_MSG(PFX "%s\n",__FUNCTION__);
	IO_Deactivate(USB_HP);
	IO_Deactivate(USB_PWR_BYPASS);
	IO_Deactivate(USB_SUSPEND_OUT);
}

void bat_set_charge_low(void)
{
	// Low power charge
	DBG_MSG(PFX "%s\n",__FUNCTION__);
	IO_Activate(USB_HP);
	IO_Deactivate(USB_PWR_BYPASS);
	IO_Deactivate(USB_SUSPEND_OUT);
}

void bat_set_charge_high(void) {
	// High power charge :
	DBG_MSG(PFX "%s\n",__FUNCTION__);
	IO_Activate(USB_PWR_BYPASS);
	IO_Deactivate(USB_HP);
	IO_Deactivate(USB_SUSPEND_OUT);
}

static void bat_set_hw_according_to_charge_states(void) {
	int charge;
	
	/* determine charge mode from requested charge modes */
	charge = g_chargeAccordingTemperature & g_chargeAccordingVoltage & g_chargeAccordingUSB;
	
	/* only effectuate charge mode if it actually really really really changed.. */
	/* effectuate charge current settings only for dublin at the moment.. */
	if (charge != g_current_charge_mode) {
		/* effectuate charge mode depending on temperature and battery voltage */
		if (charge & CHARGE_ENABLE) {
			DBG_MSG(KERN_INFO PFX "enabling charge current\n");
			/* we are charging so check if we should charge with low or high current */
			IO_SetInput(BATT_TEMP_OVER);
			if (charge & CHARGE_HIGH) {
				DBG_MSG(KERN_INFO PFX "switching to high current charge\n");
				bat_set_charge_high();
			} else {
				DBG_MSG(KERN_INFO PFX "switching to low current charge\n");
				/* low current charging */
				bat_set_charge_low();
			} /* else */
			
		} else {
			DBG_MSG(KERN_INFO PFX "disabling charge current\n");
			/* disable charge current */
			bat_set_charge_disable();
			/* reset the 60 minute battery undervoltage count */
			undervoltageWait = VOLTAGE_60_MINUTE_DELAY;
		} /* else */

		/* save current charge mode */
		g_current_charge_mode = charge;
		
	} /* charge != g_current_charge_mode */
}



/**
 * Based on the usb mode change the power bypass
 **/
void  usb_state_change_listener(USB_STATE previous_state,USB_STATE current_state, void* arg){
	if (IO_HaveUsbBusPowered()) {
		if (IO_GetModelId() == GOTYPE_DUBLIN) {
			// Only enable high power charge in CLA mode !!
			// In all other cases we can select low power charge
			if (( previous_state != USB_STATE_CLA )&&( current_state == USB_STATE_CLA ))
			{
				DBG_MSG(KERN_INFO PFX " %s HIGH POWER CHARGE \n", __FUNCTION__);
				g_chargeAccordingUSB = CHARGE_HIGH |CHARGE_ENABLE;
			} else {
				// We also end here if USB is disconnected, but it won't harm,
				// no power, no charge !
				DBG_MSG(KERN_INFO PFX " %s LOW POWER CHARGE \n", __FUNCTION__);
				g_chargeAccordingUSB = CHARGE_ENABLE;
			}
		} else {
			switch (previous_state) {
			case USB_STATE_INITIAL:
			case USB_STATE_IDLE:
			case USB_STATE_DEVICE_DETECT:
			case USB_STATE_DEVICE:
				/* if we where in a non powered mode and are switching to a powered mode...*/
				if (	current_state == USB_STATE_HOST_DETECT ||
					current_state == USB_STATE_HOST ||
					current_state == USB_STATE_CLA) {
					DBG_MSG(KERN_INFO PFX " %s ENABLE POWER CHARGE \n", __FUNCTION__);
					IO_Activate(USB_PWR_BYPASS);
				}
				break;
			case USB_STATE_HOST_DETECT:
			case USB_STATE_HOST:
			case USB_STATE_CLA:
				if (	current_state == USB_STATE_INITIAL ||
					current_state == USB_STATE_IDLE ||
					current_state == USB_STATE_DEVICE_DETECT ||
					current_state == USB_STATE_DEVICE ) {
					DBG_MSG(KERN_INFO PFX " %s DISABLE POWER CHARGE \n", __FUNCTION__);
					IO_Deactivate(USB_PWR_BYPASS);
				}
				break;
			}
		}
	}
}

static void bat_hw_init(void)
{
	int i;
	int current_limit = 1000;
	
	IO_SetInput(BATT_TEMP_OVER);

	bat_set_charge_low();
	
	g_chargeAccordingTemperature = CHARGE_DONTCARE;
	g_chargeAccordingUSB         = CHARGE_DONTCARE;
	g_chargeAccordingVoltage     = CHARGE_DONTCARE;
	
	g_current_charge_mode = g_chargeAccordingTemperature & g_chargeAccordingVoltage & g_chargeAccordingUSB;

	/* reset voltage and current measurements */
	for (i = 0; i < BATTERY_AVERAGE_SIZE; i++) {
		batteryCurrentBuffer[i] = 0;
		batteryVoltageBuffer[i] = 0;
	}
	batteryIndex = 0;
	batteryVoltageSum = 0;
	batteryCurrentSum = 0;
	
	/* Set device specific parameters */
	switch (IO_GetModelId())
	{
		case GOTYPE_NEWCASTLE:/* For Newcastle, the charge current is limited to 493 mA */
			current_limit = 493;
			ref_adc_gotype = REF_OTHER;
		break;
		case GOTYPE_DUBLIN:
			/* For Dublin, the charge current is limited to 691 mA
			 * and the TEMP_SENS - ADC value is different for B1 and B2 samples
			 */
			current_limit = 691;
			ref_adc_gotype = REF_OTHER;
			switch (GOBOARD_DUBLIN_SUBID)
			{
				case 0xFFFF:
				case GOBOARD_DUBLIN_B1:	// A and B1
					temp_adc_minus10_degrees = TEMP_ADC_MINUS10_DEGREES_B1;
					temp_adc_0_degrees       = TEMP_ADC_0_DEGREES_B1;
				break;
				case GOBOARD_DUBLIN_B2:	// B2
				case GOBOARD_DUBLIN_C1:	// C1
				case GOBOARD_DUBLIN_C2:	// C2
					temp_adc_minus10_degrees = TEMP_ADC_MINUS10_DEGREES_B2;
					temp_adc_0_degrees       = TEMP_ADC_0_DEGREES_B2;
				break;
				default:
					temp_adc_minus10_degrees = TEMP_ADC_MINUS10_DEGREES;
					temp_adc_0_degrees       = TEMP_ADC_0_DEGREES;
				break;
			}
		break;
		case GOTYPE_MARIGOT:
			current_limit = 942;
			ref_adc_gotype = REF_MARIGOT;
		break;
		default:
			current_limit = 942;
			ref_adc_gotype = REF_OTHER;
		break;
	}

	batteryCurrentLimit = current_limit;
	
	batteryWait = BATTERY_AVERAGE_SIZE;
	undervoltageWait = VOLTAGE_60_MINUTE_DELAY;

	/* reset temperature measurements */
	for (i = 0; i < TEMP_ADC_BUF_SIZE; i++) {
		tempSampleBuffer[i] = 0;
	}
	tempIndex = 0;
	tempSum = 0;
	tempWait = TEMP_ADC_BUF_SIZE;
	
	/* force charge mode to be reevaluated */
	g_current_charge_mode |= CHARGE_EVALUATE;
}

static int bat_set_charge_state_according_to_temp(int temp)
{
	int charge;

	/* initial charge action is 'do nothing' */
	charge = 0x00;

	if (temp > temp_adc_minus10_degrees) {
                DBG_MSG(KERN_INFO PFX " Temperature below -10 degrees, disable CHARGE !\n");		
		charge &= ~CHARGE_ENABLE;	/* disable charge current */
		charge &= ~CHARGE_HIGH;		/* disable high charge current */
	} else if (temp > temp_adc_0_degrees) {
                DBG_MSG(KERN_INFO PFX " Temperature between -10 and 0 degrees, charge LOW !\n");		
		charge |= CHARGE_ENABLE;	/* enable charge current */
		charge &= ~CHARGE_HIGH;		/* disable high charge current */
	} else {
		charge = CHARGE_DONTCARE;	/* whatever! use the max/best/optimal mode */
	}

	/* let caller decide what charge mode to use, we just hint
	 * in a direction depending on supplied battery temperature.. */
	return (charge);
}



static int bat_set_charge_state_according_to_voltage(int voltage) {
	int charge;

	/* initial charge action is 'do nothing' */
	charge = 0x00;

	if (voltage < VOLTAGE_3V) {
                DBG_MSG(KERN_INFO PFX " Voltage too low ! : Disable CHARGE !\n");			
		charge &= ~CHARGE_ENABLE;	/* disable charge current */
		charge &= ~CHARGE_HIGH;		/* disable high charge current */
	} else {
		charge = CHARGE_DONTCARE;	/* whatever! use the max/best/optimal mode */
	}

	/* let caller decide what charge mode to use, we just hint
	 * in a direction depending on supplied battery voltage.. */
	return (charge);
}





static void bat_adc_poll(short adcBuffer[ADC_CHANNELS], void* arg)
{
	unsigned int sample;
	unsigned int lightsensor_adc = 0 ;

	/* global indicator of voltage adc source */
	voltageSource = VOLTAGE_SOURCE_BATTERY;

	if (!batteryWait) {
		/* get battery voltage sample */
		sample = (adcBuffer[ADC_BAT] * ref_adc_gotype ) / adcBuffer[ADC_REF];
		/* (re)calculate running sum */
		batteryVoltageSum -= batteryVoltageBuffer[batteryIndex];
		batteryVoltageSum += sample;
		batteryVoltageBuffer[batteryIndex] = sample;
		
		/* get battery charge current sample */
		sample = (adcBuffer[ADC_CHRG] * batteryCurrentLimit) / adcBuffer[ADC_REF];
		/* (re)calculate running sum */
		batteryCurrentSum -= batteryCurrentBuffer[batteryIndex];
		batteryCurrentSum += sample;
		batteryCurrentBuffer[batteryIndex] = sample;

		if (++batteryIndex >= BATTERY_AVERAGE_SIZE)
			batteryIndex = 0;
	} else {
		batteryWait--;
	}

	/* voltage = (1.2 * strBatteryStatus.u16BatteryVoltage * 2 / strBatteryStatus.u16ReferenceVoltage);
	 * ampere = (1200.0 * strBatteryStatus.u16ChargeCurrent / (strBatteryStatus.u16ReferenceVoltage * 1500)); */

	/* gather and calculate temperature stuff */
	sample = adcBuffer[ADC_TEMP];
	tempSum -= tempSampleBuffer[tempIndex];
	tempSum += sample;
	tempSampleBuffer[tempIndex] = sample;

	if (++tempIndex >= TEMP_ADC_BUF_SIZE)
		tempIndex = 0;

	/* check if battery temperature is ok.. */
	if (!tempWait) {
		/* get suggested charge mode depending on measured average temperature */
		g_chargeAccordingTemperature = bat_set_charge_state_according_to_temp(tempSum / TEMP_ADC_BUF_SIZE);
		/* reinstate the temperature check delay */
		tempWait = TEMP_20MINUTES_DELAY;
	} else {
		/* the clock is ticking.. */
		--tempWait;
	}

	/* check if battery voltage below charge threshold..
	 * ..and only if we can calculate a valid average voltage..! */
	if ((!batteryWait) && (!undervoltageWait)) {
		/* get suggested charge mode depending on measured average temperature */
		g_chargeAccordingVoltage = bat_set_charge_state_according_to_voltage(batteryVoltageSum / VOLTAGE_ADC_BUF_SIZE);
	} else {
		/* we don't care what charge mode is going to be
		 * we are still busy waiting for an average measurement */
		g_chargeAccordingVoltage = CHARGE_DONTCARE;
	}


	/* effectuate charge current settings.. */
	bat_set_hw_according_to_charge_states();  /* After this function "g_current_charge_mode" will reflect the set charge state */

		/* only measure chargetime if we are actually charging !!!*/
	if (undervoltageWait>0)
	{
		if (g_current_charge_mode & CHARGE_ENABLE)
				--undervoltageWait;
//		/* less time when charging faster */
//		if (g_current_charge_mode & (CHARGE_HIGH|CHARGE_ENABLE))
//				--undervoltageWait;
	}
#if defined(LOGGING_INTERVAL)
	if (!loggingTime) {
		/* some logging */
		int dummy = batteryCurrentSum / BATTERY_AVERAGE_SIZE;
		int temperature = 0;
		
		temperature = calculate_pnd_temperature(tempSum / TEMP_ADC_BUF_SIZE);

		if (IO_GetInput(BATT_TEMP_OVER)) // Charge-current is miscalculated at BATT_TEMP_OVER 
		{
			dummy = 0;
		}
		lightsensor_adc = adcBuffer[ADC_LX_OUT];
		printk(KERN_INFO PFX "ModelId:%d SubId:%d T:%d(%d), V:%d mV, C:%d mA, CM:T%d V%d U%d -> %s%s%s%s (%s) %s Light:%d LowDC:%d, HU_PWR_ON: %d\n",
			IO_GetModelId(),
			IO_GetSubId(),
			tempSum / TEMP_ADC_BUF_SIZE,temperature,
			batteryVoltageSum / BATTERY_AVERAGE_SIZE,
			dummy,
			g_chargeAccordingTemperature,
			g_chargeAccordingVoltage,
			g_chargeAccordingUSB,
			(g_current_charge_mode==0x03)?"CHARGE_HIGH  ":"",	
			(g_current_charge_mode==0x02)?"CHARGE_ENABLE":"",	
			(g_current_charge_mode==0x01)?"CHARGE_LOW   ":"",
			(g_current_charge_mode==0x80)?"CHARGE_EVAL  ":"",
			IO_GetInput(CHARGING)?"CHARGING":"NoCharging",
			(IO_GetInput(CHARGING)&&IO_GetInput(BATT_TEMP_OVER))?"!! BATT_TEMP_OVER !!":"",
			lightsensor_adc,
			IO_GetInput(LOW_DC_VCC),
			IO_GetInput(HU_PWR_ON)
			);
		loggingTime = LOGGING_INTERVAL;
	} else {
		--loggingTime;
	}
#endif	/* LOGGING_INTERVAL */
}

static int bat_get_voltage(void)
{
	return batteryVoltageSum / BATTERY_AVERAGE_SIZE;
}

static int bat_get_current(void)
{
	int chargecurrent = batteryCurrentSum / BATTERY_AVERAGE_SIZE;
	
	if (IO_GetModelId() == GOTYPE_DUBLIN)
	{
		if (IO_GetInput(BATT_TEMP_OVER)) 
		{
			chargecurrent = 0;
		}
	}
	
	return chargecurrent ;
}

static int bat_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int ret;

	if (bat_disabled)	// Suspend mode
		return -ENODEV;

	switch (cmd) {
	case IO_ENABLE_CHARGING:
		DBG_MSG(PFX " %s:IO_ENABLE_CHARGING\n",__FUNCTION__);
		IO_SetInput(BATT_TEMP_OVER);
		ret = 0;
		break;
	case IO_DISABLE_CHARGING:
		DBG_MSG(PFX " %s:IO_DISABLE_CHARGING\n",__FUNCTION__);
		IO_Activate(BATT_TEMP_OVER);
		ret = 0;
		break;
	case IOR_BATTERY_STATUS:
		if (!access_ok(VERIFY_WRITE, (void __user *) arg, sizeof(BATTERY_STATUS))) {
			ret = -EFAULT;
		} else {
			BATTERY_STATUS bs;
			bs.u16BatteryVoltage   = bat_get_voltage();
			bs.u16ChargeCurrent    = bat_get_current();
			bs.u8VoltageSource     = voltageSource;

			/* Power state */
			switch (IO_GetChargerType()) {
				case GOCHARGER_LTC3455:
					/* Use ADC to detect charge current */
					if (IO_GetInput(ACPWR))
					{
						bs.u8ChargeStatus = CHARGE_STATE_COMPLETE;
						if (IO_GetInput(BATT_TEMP_OVER)) {
							bs.u8ChargeStatus = CHARGE_STATE_FAULT;
						}
						if (bs.u16ChargeCurrent > 50 ) bs.u8ChargeStatus = CHARGE_STATE_TOPPINGUP;
						if (bs.u16ChargeCurrent > 500) bs.u8ChargeStatus = CHARGE_STATE_FAST_CHARGING;
					} else {
						bs.u8ChargeStatus = CHARGE_STATE_ACPWR_OFF;
					}
					break;
				case GOCHARGER_LTC1733:
					/* Use CHARGE_OUT to detect charging */
					if (IO_GetInput(ACPWR)) {
						if (IO_GetInput(CHARGEFAULT)) {
							bs.u8ChargeStatus = CHARGE_STATE_FAULT;
						}
						else if (IO_GetInput(CHARGING)) {
							IO_Activate(CHARGE_OUT);
							if (IO_GetInput(CHARGING)) {
								bs.u8ChargeStatus = CHARGE_STATE_TOPPINGUP;
							} else {
								bs.u8ChargeStatus = CHARGE_STATE_FAST_CHARGING;
							}
							IO_Deactivate(CHARGE_OUT);
						} else {
							bs.u8ChargeStatus = CHARGE_STATE_COMPLETE;
						}
					} else {
						bs.u8ChargeStatus = CHARGE_STATE_ACPWR_OFF;
					}
					break;
				default:
					bs.u8ChargeStatus = CHARGE_STATE_ACPWR_OFF;
			}

			ret = __copy_to_user((void __user *) arg, &bs, sizeof(BATTERY_STATUS)) ? -EFAULT : 0;
		}
		break;
	default:
		PK_WARN("Invalid ioctl command %u\n", cmd);
		ret = -EINVAL;
		break;
	}

	return ret;
}

static int bat_open(struct inode *inode, struct file *file)
{
	return nonseekable_open(inode, file);
}

static int bat_release(struct inode *inode, struct file *file)
{
	return 0;
}

/* Kernel interface */
static struct file_operations bat_fops = {
	.owner		= THIS_MODULE,
	.ioctl		= bat_ioctl,
	.open		= bat_open,
	.release	= bat_release,
};

static int bat_probe(struct device *dev)
{
	int ret;

	bat_hw_init();

	PK_DBG("Registering ADC pollfunc\n");
	ret = adc_register_poll(bat_adc_poll, NULL );
	if (ret < 0) {
		PK_ERR("Unable to register ADC pollfunc (%d)\n", ret);
		return ret;
	}

	PK_DBG("Registering chardev\n");
	ret = register_chrdev(BATTERY_MAJOR, BATTERY_DEVNAME, &bat_fops);
	if (ret != 0) {
		PK_ERR("Unable to register chardev on major=%d (%d)\n", BATTERY_MAJOR, ret);
		return ret;
	}

	PK_DBG("Registering usbmode callback\n");
        ret = add_usb_state_change_listener(usb_state_change_listener,0);
        if (ret != 0){
                printk(KERN_INFO PFX "%s ERROR failed to register usbmode callback\n", __FUNCTION__);
                return ret;
        }
	PK_DBG("Done\n");
	return 0;
}

static int bat_remove(struct device *dev)
{
	PK_DBG("Unregistering chardev\n");
	unregister_chrdev(BATTERY_MAJOR, BATTERY_DEVNAME);
	remove_usb_state_change_listener(usb_state_change_listener,0);

	PK_DBG("Done\n");
	return 0;
}

static void bat_shutdown(struct device *dev)
{
	PK_DBG("Shutting down\n");
	/* Nothing yet */
}

#ifdef CONFIG_PM

static int bat_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);
	if (level == SUSPEND_POWER_DOWN) {
		bat_disabled = 1;
		if (IO_HaveUsbBusPowered()) {
			IO_Deactivate(USB_PWR_BYPASS);
		}
	}
	return 0;
}

static int bat_resume(struct device *dev, u32 level)
{
	PK_DBG("dev = %p, level = %u\n", dev, level);
	if (level == RESUME_POWER_ON) {
		bat_hw_init();
		bat_disabled = 0;
		if (IO_HaveUsbBusPowered()){
			/*make sure we don't blow up our friend PC's */
			IO_Deactivate(USB_PWR_BYPASS);
		}
	}
	return 0;
}

#else /* CONFIG_PM */
#define bat_suspend NULL
#define bat_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver bat_driver = {
	.name		= "tomtomgo-bat",
	.bus		= &platform_bus_type,
	.probe		= bat_probe,
	.remove		= bat_remove,
	.shutdown	= bat_shutdown,
	.suspend	= bat_suspend,
	.resume		= bat_resume,
};

static int __init bat_mod_init(void)
{
	int ret;

	printk(KERN_INFO "TomTom GO Battery Driver, (C) 2004,2005,2006,2007 TomTom BV\n");
	PK_DBG("Registering driver\n");
	ret = driver_register(&bat_driver);
	if (ret) {
		PK_ERR("Unable to register driver (%d)\n", ret);
		return ret;
	}
	PK_DBG("Done\n");
	return 0;
}

static void __exit bat_mod_exit(void)
{
	PK_DBG("Unregistering driver\n");
	driver_unregister(&bat_driver);
	PK_DBG("Done\n");
}

module_init(bat_mod_init);
module_exit(bat_mod_exit);

MODULE_AUTHOR("Dimitry Andric <dimitry.andric@tomtom.com>");
MODULE_DESCRIPTION("TomTom GO Battery Driver");
MODULE_LICENSE("GPL");

/* EOF */
