/* drivers/barcelona/diag/diag.c
 *
 * Implementation of the diagnostics driver.
 *
 * Copyright (C) 2007 TomTom BV <http://www.tomtom.com/>
 * Author 0: Bjorn Martens <bjorn.martens@tomtom.com>
 * Author 1: Patrique Haidar <patrique.haidar@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* Includes */ 
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/hardware/clock.h>
#include <barcelona/Barc_diag.h>
#include <barcelona/Barc_adc.h>
#include <barcelona/debug.h>
#include <barcelona/gotype.h>
#include <barcelona/gopins.h>

/* defines */
#define PFX "diag: "
#define PK_DBG PK_DBG_FUNC


#define REF_VOLTAGE     12250    // Reference voltage = 1.225V (in 0.1 mV units)

typedef struct temp_conversion 
{
	int  temperature ;  // Temperature in degrees Celsius
	int  voltage ;      // Voltage in 0.1 mV units
} temp_conversion;

static const temp_conversion temp_conversion_table[] = 
{
{-30,  24380},
{-29,  24280},
{-28,  24170},
{-27,  24070},
{-26,  23970},
{-25,  23860},
{-24,  23760},
{-23,  23650},
{-22,  23550},
{-21,  23440},
{-20,  23340},
{25,   18560},
{30,   18020},
{35,   17470},
{40,   16930},
{45,   16380},
{50,   15840},
{75,   13080},
{76,   12960},
{77,   12850},
{78,   12740},
{79,   12630},
{80,   12520},
{81,   12410},
{82,   12290},
{83,   12180},
{84,   12070},
{85,   11960},
{86,   11850},
{87,   11730},
{88,   11620},
{89,   11510},
{90,   11400},
{91,   11290},
{92,   11170},
{93,   11060},
{94,   10950},
{95,   10840},
{96,   10720},
{97,   10610},
{98,   10500},
{99,   10380},
{100,  10270}
};

#define CONVERSION_TABLE_SIZE (sizeof(temp_conversion_table) / sizeof(temp_conversion))

static int get_temperature_deg(void)
{
  int temperature;
  int adc_voltage;
  int i;
  
  adc_voltage = (adc_get_channel(ADC_TEMP) * REF_VOLTAGE) / adc_get_channel(ADC_REF);

  i = 0;
  while ((i < CONVERSION_TABLE_SIZE) && temp_conversion_table[i].voltage > adc_voltage)
  {
    i++;
  }
  
  if (i < CONVERSION_TABLE_SIZE)
  {
    if (i > 0)
    {
      // Linear interpolation
      temperature = temp_conversion_table[i-1].temperature +
                    (((temp_conversion_table[i].temperature - temp_conversion_table[i-1].temperature) *
                      (adc_voltage - temp_conversion_table[i-1].voltage)) / 
                      (temp_conversion_table[i].voltage - temp_conversion_table[i-1].voltage));
    }
    else
    {
       temperature = temp_conversion_table[0].temperature; // Clip to minimum temperature
    }
  }
  else
  {
     temperature = temp_conversion_table[CONVERSION_TABLE_SIZE-1].temperature; // Clip to maximum temperature
  }
   
  return temperature;
}


static int diag_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
          unsigned long arg)
{
  int ret = 0;
  
  /* TODO ADC_MUX_DIAGNOSTICS_SETTING, ADC_CAR_BATTERY, ADC_REFERENCE and ADC_TEMPERATURE should 
   * be defined in Barc_adc.h. 
   * Until the diagnostics port number is known we define 
   * it here. For testing purposes we set it to the Touchscreen 
   * Y value, ADC_BAT, ADC_TS_Y and ADC_TEMP.
   */
  const int ADC_MUX_DIAGNOSTICS_SETTING = ADC_CHRG; // AIN6
  const int ADC_CAR_BATTERY = ADC_BAT;              // AIN0
  const int ADC_REFERENCE = ADC_REF;                // AIN4
  const int ADC_TEMPERATURE = ADC_TEMP;             // AIN1


  /* Reading internal samsung ADC */
  int adc_value =  -EINVAL;
  int temperature_value =  -EINVAL;

  switch (cmd) {
  case MUX_GPIO_0:
    /* Deactivate or Activate the GPIO pin according to the  
     * argument.
     */
    switch (arg) {
    case 0:
      IO_Deactivate(GPIO_DIAG0);
      break;
    case 1:
      IO_Activate(GPIO_DIAG0);
      break;
    default:
      ret = -EINVAL;
      break;
    }
    break;
  case MUX_GPIO_1:
    /* Deactivate or Activate the GPIO pin according to the  
     * argument.
     */
    switch (arg) {
    case 0:
      IO_Deactivate(GPIO_DIAG1);
      break;
    case 1:
      IO_Activate(GPIO_DIAG1);
      break;
    default:
      ret = -EINVAL;
      break;
    }
    break;
  case MUX_GPIO_2:
    /* Deactivate or Activate the GPIO pin according to the  
     * argument.
     */
    switch (arg) {
    case 0:
      IO_Deactivate(GPIO_DIAG2);
      break;
    case 1:
      IO_Activate(GPIO_DIAG2);
      break;
    default:
      ret = -EINVAL;
      break;
    }
    break;
  case GPIO_GPS_EXT_ANT:
    /* Deactivate or Activate the GPIO pin according to the  
     * argument.
     */
    switch (arg) {
    case 0:
      IO_Deactivate(GPS_EXT_ANT);
      break;
    case 1:
      IO_Activate(GPS_EXT_ANT);
      break;
    default:
      ret = -EINVAL;
      break;
    }
    break;
  case IO_USB_POWERLINE:
    /* FIXME the 1 should be changed to a read GPIO pin */
    ret = put_user(1, (int __user *)arg);
    break;
  case IO_DIAGNOSTICS_READ_ADC:
    /* Reading internal samsung ADC */
    adc_value = adc_get_channel(ADC_MUX_DIAGNOSTICS_SETTING);

    ret = put_user(adc_value, (int __user *)arg);
    break;
  case IO_DIAGNOSTICS_GET_ADC_RATE:
    ret = put_user(ADC_RATE, (int __user *)arg);
    break;
  case IO_CAR_BATTERY:
    adc_value = adc_get_channel(ADC_CAR_BATTERY);

    ret = put_user(adc_value, (int __user *)arg);
    break;
  case IO_TEMPERATURE:
    adc_value = adc_get_channel(ADC_TEMPERATURE);

    ret = put_user(adc_value, (int __user *)arg);
    break;
  case IO_TEMPERATURE_DEG:
    if (IO_HaveTemperatureSensor())
    {
        temperature_value = get_temperature_deg();
    }
    else
    {
        temperature_value = 0;
    }
    ret = put_user(temperature_value, (int __user *)arg);
    break;
  case IO_REFERENCE:
    adc_value = adc_get_channel(ADC_REFERENCE);

    ret = put_user(adc_value, (int __user *)arg);
    break;
  case IO_SDCARD:
    ret = put_user(IO_GetInput(CD_SD), (int __user *)arg);
    break;
  case IO_MUTE:
    IO_SetInput(NAV_MUTE_DIAG);
    ret = put_user(IO_GetInput(NAV_MUTE_DIAG), (int __user *)arg);
    break;
  default:
    PK_WARN("Invalid ioctl command %u\n", cmd);
    ret = -EINVAL;
    break;
  }
  return ret;
}

static int diag_open(struct inode *inode, struct file *file)
{
  return nonseekable_open(inode, file);
}

static int diag_release(struct inode *inode, struct file *file)
{
  return 0;
}

/* Kernel interface */
static struct file_operations diag_fops = {
  .owner    = THIS_MODULE,
  .ioctl    = diag_ioctl,
  .open    = diag_open,
  .release  = diag_release,
};

static int diag_probe(struct device *dev)
{
  return 0;
}

static int diag_remove(struct device *dev)
{
  return 0;
}

static void diag_shutdown(struct device *dev)
{
  PK_DBG("Shutting down diag\n");
}

#ifdef CONFIG_PM

static int diag_suspend(struct device *dev, u32 state, u32 level)
{
  return 0;
}

static int diag_resume(struct device *dev, u32 level)
{
  return 0;
}

#else /* CONFIG_PM */
#define diag_suspend NULL
#define diag_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver diag_driver = {
  .name    = "tomtomgo-diag",
  .bus    = &platform_bus_type,
  .probe    = diag_probe,
  .remove    = diag_remove,
  .shutdown  = diag_shutdown,
  .suspend  = diag_suspend,
  .resume    = diag_resume,
};

static int __init diag_mod_init(void)
{
  int ret;

  printk(KERN_INFO "TomTom Diagnostics Driver, (C) 2007 TomTom BV\n");
  
  PK_DBG("Registering driver\n");
  ret = driver_register(&diag_driver);
  if (ret) {
    PK_ERR("Unable to register driver (%d)\n", ret);
    return ret;
  }

  PK_DBG("Registering chardev\n");
  ret = register_chrdev(DIAGNOSTICS_MAJOR, DIAGNOSTICS_DEVNAME, &diag_fops);
  if (ret != 0) {
    PK_ERR("Unable to register chardev on major=%d (%d)\n", DIAGNOSTICS_MAJOR, ret);
    return ret;
  }
    
  return ret;
}

static void __exit diag_mod_exit(void)
{
  PK_DBG("Unregistering chardev\n");
  unregister_chrdev(DIAGNOSTICS_MAJOR, DIAGNOSTICS_DEVNAME);

  PK_DBG("Unregistering driver\n");
  driver_unregister(&diag_driver);
}

module_init(diag_mod_init);
module_exit(diag_mod_exit);

MODULE_AUTHOR("Bjorn Martens <bjorn.martens@tomtom.com>");
MODULE_DESCRIPTION("TomTom Diagnostics Driver");
MODULE_LICENSE("GPL");

/* EOF */ 
