/* drivers/barcelona/gpio/buspower.c
 *
 * Implementation of the TomTom VBus power detection driver.
 *
 * Copyright (C) 2004,2005,2006,2007 TomTom BV <http://www.tomtom.com/>
 * Authors:	Ithamar Adema <Ithamar.adema@tomtom.com>
 *			Dimitry Andric <dimitry.andric@tomtom.com>
 *			Koen Martens <kmartens@sonologic.nl>
 *			Jeroen Taverne <jeroen.taverne@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <asm/hardware.h>
#include <asm/mach-types.h>
#include <asm/hardware/clock.h>
#include <asm/io.h>
#include <linux/kernel.h>
#include <linux/err.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/interrupt.h>

#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-clock.h>
#include <asm/arch/regs-hsudc.h>
#include <asm/arch/regs-udc.h>
#include <asm/arch/regs-dyn.h>

#include <barcelona/gopins.h>

#include "buspower.h"

/* usb state machine states 
UNDEF		== the default state undefined
HOST		== usb is in host mode
CLA		== car lighter adapter mode , in this mode
the device is receiving power from the USB, in this mode
some changes to the battery loader are performed because we
know we can charge at faster rate then the defined 500mA for
usb clients.
NOVBUS		== NO VBUS
CONFIGURED	== ???
BUSY		== ???
*/

#define USB_STATE_UNDEF		0
#define USB_STATE_HOST			4
#define USB_STATE_CLA			5
#define USB_STATE_NOVBUS		6
#define USB_STATE_CONFIGURED		7
#define USB_STATE_BUSY			8


static int usb_flag       = 0;
/* usb H(ost) state? */
static int usb_hstate     = USB_STATE_UNDEF;

/* This value is set to 1 if the device is a usb 2.0
device. the s3c2443 for example is such a device. usb 2.0
has as different default wireing. One example is that in
usb 1.0 the client must have a 1.5K pulldown resitors. In
usb 2.0 this is done differently. buspower also contains
code to enable and disable the 1.5K pulldown if de device switches
to host mode. We don't use client mode at all at present time since the usb client mode
is implemented in the bootloader */
static int has_usb_v2	  = 0;

//USB_INT_REG / USB_INT_EN_REG
#define SUSPEND_INT            	          0x01
#define RESUME_INT                        0x02
#define RESET_INT                         0x04

int usbResetTick;

#define PFX "buspower: "

/* We have two irq handlers based on the usb implementation if we are using usb 1.0
we use the buspower_irq_handler callback and if we have usb 2.0 we use the usb2_irq_handler
callback */
static irqreturn_t buspower_irq_handler(int irq, void *dev_id, struct pt_regs *regs);
static irqreturn_t usb2_irq_handler(int irq, void *dev_id, struct pt_regs *regs);
static void buspower_UnSetDevice(void); 

static struct clk *clk;

void buspower_SetUsbCurrent(int cur);

static void usb2_hw_init(void);

// static 
int buspower_probe(struct device *_dev)
{
	int ret;

	has_usb_v2	= (IO_GetCpuType() == GOCPU_S3C2443);
	usb_flag 	= 0;
	usb_hstate 	= USB_STATE_UNDEF;

	ret = request_irq(IRQ_USBD, has_usb_v2 ? usb2_irq_handler : buspower_irq_handler, 0,  "Buspower IRQ driver", _dev);
	if (ret != 0) {
		printk(KERN_INFO PFX "failed to install irq (%d)\n", ret);
		return ret;
	}

	clk = clk_get(NULL, "usb-device");
	if (IS_ERR(clk)) {
		printk(KERN_INFO PFX "cannot get usb-device clock\n");
		ret = -ENOENT;
		goto err;
	}
	clk_use(clk);
	buspower_hw_init();

	return 0;

err:
	free_irq(IRQ_USBD, _dev);
	return ret;
		
}

// static 
int buspower_remove(struct device *_dev)
{
	printk(KERN_INFO PFX "%s ()\n", __FUNCTION__);
	
	buspower_hw_exit();
	
	printk(KERN_INFO PFX "Freeing IRQ\n");
	free_irq(IRQ_USBD, _dev);

	/* no unregistration of chardev */

	printk(KERN_INFO PFX "Done\n");
	return 0;
}

static irqreturn_t
buspower_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	int handled;

	do {
		int saveIdx = __raw_readb(S3C_UDC_INDEX_REG);
		int usb_status = __raw_readb(S3C_UDC_USB_INT_REG);
		int usbd_status = __raw_readb(S3C_UDC_EP_INT_REG);
	
		handled = 0;
				
		/* ReSeT Interrupt Request - USB reset */
		if (usb_status & S3C_UDC_USBINT_RESET) {
			/* clear interrupt */
			__raw_writeb(S3C_UDC_USBINT_RESET,
				S3C_UDC_USB_INT_REG);
			buspower_ResetDetected();
			handled = 1;
		}

		/* RESume Interrupt Request */
		if (usb_status & S3C_UDC_USBINT_RESUME) {
			/* clear interrupt */
			__raw_writel(S3C_UDC_USBINT_RESUME,
				S3C_UDC_USB_INT_REG);
			handled = 1;
		}

		/* SUSpend Interrupt Request */
		if (usb_status & S3C_UDC_USBINT_SUSPEND) {
			/* clear interrupt */
			__raw_writel(S3C_UDC_USBINT_SUSPEND,
				S3C_UDC_USB_INT_REG);
				
			handled = 1;
		}

		/* EP interrupts */
	   	if (usbd_status) {
  			int	i;

  			/* control traffic */
  			if (usbd_status & S3C_UDC_INT_EP0) {
  				/* Clear the interrupt bit by setting it to 1 */
  				__raw_writel(S3C_UDC_INT_EP0,S3C_UDC_EP_INT_REG);
  				handled = 1;
  			}

  			/* endpoint data transfers */
  			for (i = 1; i < 5; i++) {
  				u32 tmp = 1 << i;
  				if (usbd_status & tmp) {
  					/* Clear the interrupt bit by setting it to 1 */
  					__raw_writel(tmp, S3C_UDC_EP_INT_REG);
  					handled = 1;
  				}
  			}
		}
		/* restore index reg */
		__raw_writeb(saveIdx, S3C_UDC_INDEX_REG);
	} while (handled);
	
	return IRQ_HANDLED;	
}

#define S3C2410_CLKSLOW_USB_CLK_DISABLE		(1 << 7)

void buspower_hw_init(void)
{
	int tmp;

	usb_flag 	= 0;
	usb_hstate 	= USB_STATE_UNDEF;

	switch (IO_GetCpuType()) {
		case GOCPU_S3C2443:
			usb2_hw_init();
			break;

		case GOCPU_S3C2410:
		case GOCPU_S3C2440:
		case GOCPU_S3C2442:
			/* 2410/2440/2442, not 2412/2413 */
			tmp = __raw_readb(S3C2410_CLKSLOW);
			tmp &= ~S3C2410_CLKSLOW_USB_CLK_DISABLE; /* enable usb device clock */
			__raw_writeb(tmp, S3C2410_CLKSLOW);
			/* intentional fallthrough */

		default:
			/* disable EP0-4 USBD interrupts for now */
			__raw_writeb(0x00, S3C_UDC_USB_INT_EN_REG);
	
			/* Select USBD instead of USBH1, enable USB port 1 */
			tmp = __raw_readb(s3c24xx_misccr);
			tmp &= ~S3C2410_MISCCR_USBHOST; // (1<<3);
			tmp &= ~S3C2410_MISCCR_USBSUSPND1; // (1<<13);
			__raw_writeb(tmp, s3c24xx_misccr);
	
			/* just clear */
			tmp = __raw_readb(S3C_UDC_EP_INT_REG);
			tmp = __raw_readb(S3C_UDC_USB_INT_REG);

			/* Enable USB interrupts for RESET and SUSPEND/RESUME on USBD */
			tmp  = S3C_UDC_USBINT_RESET | S3C_UDC_USBINT_SUSPEND;
			__raw_writeb(tmp, S3C_UDC_USB_INT_EN_REG);
			break;
	}
}


static irqreturn_t
usb2_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned int s, ep_int, tr;
	unsigned int fnr;

	/* Get interrupt source */
	s = __raw_readl(S3C_HSUDC_SYS_STATUS_REG);
	ep_int = __raw_readl(S3C_HSUDC_EP_INT_REG);
	tr = __raw_readl(S3C_HSUDC_TEST_REG);
	fnr = __raw_readl( S3C_HSUDC_FRAME_NUM_REG );

	/* Ack any endpoint activity */
	if (ep_int) {
		__raw_writel(ep_int, S3C_HSUDC_EP_INT_REG);
	}

	/* If a SOF condition is detected, flag this. */
	if( fnr & (1 << 14) ){
		buspower_SOFDetected( );
		buspower_UnSetDevice(); /* disable pullup etc */
	}

	if (s) {
		if (s & S3C_HSUDC_INT_RESET) {
			buspower_ResetDetected();
			buspower_UnSetDevice(); /* disable pullup etc */
		}
		if (s & S3C_HSUDC_INT_VBUSON) {
		}
		if (s & S3C_HSUDC_INT_VBUSOFF) {
		}

		__raw_writel(s, S3C_HSUDC_SYS_STATUS_REG);
	}

	if (tr) {
		__raw_writel(tr, S3C_HSUDC_TEST_REG);
	}

	return IRQ_HANDLED;
}

static __inline__ void
usb2_write(u32 val, u32 port, u8 ind)
{
	__raw_writel(ind, S3C_HSUDC_INDEX_REG);
	__raw_writel(val, port);
}

/* Setup new USB2 controller in 2443 (and probably later) */
static void
usb2_hw_init(void)
{
	if (__raw_readl(S3C2443_RSTSTAT) & S3C2443_RSTSTAT_SLEEP) {
		__raw_writel(__raw_readl(S3C2443_RSTCON) | S3C2443_RSTCON_PWROFF_SLEEP, S3C2443_RSTCON);
	}
	 
	/* put USB port in normal mode (not suspend) */
	__raw_writel(__raw_readl(s3c24xx_misccr) & ~(1 << 12), s3c24xx_misccr);

	/* enable external USB PHY power regulator */
	IO_Activate(USB_PHY_PWR_EN);
	mdelay(1);

	__raw_writel(__raw_readl(S3C2443_PWRCFG) | S3C2443_PWRCFG_nSW_PHY_OFF_USB, S3C2443_PWRCFG);

	/* USB device 2.0 must reset like below,
	 * 1st phy reset and after at least 10us, func_reset & host reset
	 * phy reset can reset below registers.
	 */

	__raw_writel( S3C2443_URSTCON_PHY_RESET, S3C2443_URSTCON);
	udelay( 20 ); /* phy reset must be asserted for at least 10us */
 
	/*Function 2.0, Host 1.1 S/W reset, deassert PHY reset */
	__raw_writel( S3C2443_URSTCON_FUNC_RESET | S3C2443_URSTCON_HOST_RESET, S3C2443_URSTCON);
	udelay( 20 );
	__raw_writel(0, S3C2443_URSTCON);	/* deassert all resets	*/
 
	mdelay(10); 
 
	/* 12Mhz clock on ,PHY2.0 analog block power on
	 * XO block power on,XO block power in suspend mode,
	 * PHY 2.0 Pll power on ,suspend signal for save mode disable
	 */
	__raw_writel( S3C2443_PHYPWR_COMMON_ON_N, S3C2443_PHYPWR );
 
	/* D+ pull up disable(VBUS detect), USB2.0 Function clock Enable,
	 * USB1.1 HOST disable, USB2.0 PHY test disable
	 */
	__raw_writel(	S3C2443_UCLKCON_DETECT_VBUS	|	/* detect VBUS */
			S3C2443_UCLKCON_HOST_CLK_TEST	|	/* DISable HOST_CLK_TEST*/
			S3C2443_UCLKCON_FUNC_CLK_EN,	/* USB 2 clock enable */ 
			S3C2443_UCLKCON);


	/* EP0-2 Interrupt enable */
	__raw_writel(0x7, S3C_HSUDC_EP_INT_EN_REG);

	__raw_writel(	0x0000, S3C_HSUDC_TEST_REG );
 
	/* error interrupt enable, 16bit bus, Little format,
	 * suspend&reset enable
	 */
	__raw_writew(	S3C_HSUDC_DTZIEN_EN |
			S3C_HSUDC_RRD_EN |	/* low byte first NOT supported...	*/
			S3C_HSUDC_SUS_EN | 
			S3C_HSUDC_RST_EN,	/* | S3C2443_UDC_SCR_BIS;	*/
 			S3C_HSUDC_SYS_CON_REG);

	__raw_writel(0, S3C_HSUDC_EP0_CON_REG);	
}


void buspower_hw_exit(void)
{
	/*	simple case: suspend always 500 mA mode	*/
	buspower_SetUsbCurrent(500); 
}

/* external or internal (s3c) callbacks */
void buspower_ResetDetected(void)
{
	/* callback from USB ISR suspend/resume/reset signalling */
	usb_flag = 1;
}

void buspower_SOFDetected( void )
{
	/* callback from USB ISR SOF signalling */
	usb_flag = 1;
}

void buspower_SetConfiguredCurrent(void)
{
	/* can called from interrupt context */
	/* callback from USBd if we can draw the full 500 mA (current) */
	// usb_flags = USB_STATE_CONFIGURED;
	
	/* no PM calls, let USB_PollCLA and PM_Poll() take care of that */
}

void buspower_SetDevice(void)
{
	unsigned i;

	IO_SetInput(USB_HOST_DETECT);
	
	if (!has_usb_v2) {
		/* Select USBD instead of USBH1, enable USB port 1 */
		i = __raw_readl(s3c24xx_misccr);
		i = i & ~(1<<3);
		i = i & ~(1<<13);
		__raw_writel(i, s3c24xx_misccr);
			
		mdelay(10);

		/* Enable USB interrupts for RESET and SUSPEND/RESUME on USBD	*/
		__raw_writeb(0x04 | 0x01, S3C_UDC_USB_INT_REG);
		
		/* Signal to host that USBD is here using D+ pullup (full-speed) */
		IO_Activate(USB_PULL_EN);
	
	} else {
		/* Enable pull up	*/
		unsigned long uclkcon;
		uclkcon = __raw_readl( S3C2443_UCLKCON );
		uclkcon |= S3C2443_UCLKCON_DETECT_VBUS;
		__raw_writel(uclkcon, S3C2443_UCLKCON);
	}	
}

void buspower_UnSetDevice(void)
{
	/* Disable USB interrupts for RESET and SUSPEND/RESUME on USBD	*/
	// __raw_writeb(0x00, S3C_UDC_USB_INT_REG);
	
	if( has_usb_v2){
		/* Disable pull up	*/
		unsigned long uclkcon;
		uclkcon = __raw_readl( S3C2443_UCLKCON );
		uclkcon &= ~( S3C2443_UCLKCON_DETECT_VBUS );
		__raw_writel(uclkcon, S3C2443_UCLKCON);
	}
	else{
		IO_Deactivate(USB_PULL_EN);
	}
}

/* end s3c specific stuff */
int buspower_PollCLA(void)
{
	static int	from_cla=0;

	switch (usb_hstate) {
		case USB_STATE_UNDEF:
			if (IO_HaveUsbBusPowered()) {
				buspower_SetUsbCurrent(100);
				if (IO_GetInput(USB_HOST_DETECT)) {
					buspower_SetDevice(); /* wave with carrot to host (enable pullup) */
					usbResetTick = 3 * 5; /* 3 secs no USB activity/reset -> assume CLA */
					usb_hstate = USB_STATE_BUSY;
					printk("Starting USB vbus probe (USB charge supported)\n");
				} else {
					printk("No USB power (1) (USB charge supported)\n");
					buspower_SetUsbCurrent(0);
					usb_hstate = USB_STATE_NOVBUS;
				}
			} else {
				buspower_SetUsbCurrent(0);
				if (IO_GetInput(USB_HOST_DETECT)) {
					printk("USB Host detected (simple USB)\n");
					usb_hstate = USB_STATE_HOST;
				} else {
					printk("USB Host disconnected detected (simple USB)\n");
					usb_hstate = USB_STATE_NOVBUS;
				}
			}
			break;
		case USB_STATE_BUSY:
			if (!IO_GetInput(USB_HOST_DETECT)) {
				printk("No USB power (2) (USB charge supported)\n");
				usb_hstate = USB_STATE_NOVBUS;
				buspower_SetUsbCurrent(100);
			}
			/* timeout */
			if (usbResetTick <= 0) {
				buspower_UnSetDevice(); /* disable pullup etc */
				usb_hstate = USB_STATE_CLA; /* timeout */
				printk("USB CLA power detected (USB charge supported)\n");
			} else {
				if (usb_flag) {
					printk("USB Host detected (USB charge supported)\n");

					usb_hstate = USB_STATE_HOST;
					buspower_UnSetDevice();
				}
			}
			break;
		case USB_STATE_HOST:
			/* Check if we came from CLA mode. */
			if( from_cla )
			{
				from_cla=0;
				printk( "USB CLA power detected in Host mode (Increased USB charge supported)\n" );
				buspower_SetUsbCurrent( 2000 );
			}

			/* end of host mode ? */
			if (!IO_GetInput(USB_HOST_DETECT)) {
				printk("Lost USB host power (3)\n");
				usb_hstate = USB_STATE_NOVBUS;
				buspower_SetUsbCurrent(100);
			}
			break;
		case USB_STATE_CLA:
			/* Ensure the from_cla flag is cleared. */
			if( from_cla )
			{
				printk( "Transitioning from USB CLA mode to USB CLA mode\n" );
				from_cla=0;
			}

			if (!IO_GetInput(USB_HOST_DETECT)) {
				printk("Lost USB CLA power (4)\n");
				usb_hstate = USB_STATE_NOVBUS;
				buspower_SetUsbCurrent(0);
			}
			else
			{
				/* Buspower is still here... We might get a cable insert, generating a reset. */
				if( usb_flag )
				{
					/* Set from_cla. This to indicate that we transition from CLA mode to Host mode. */
					/* this is so we can keep the charging current high. */
					from_cla=1;
					usb_hstate=USB_STATE_NOVBUS;
				}
			}
			break;
		case USB_STATE_NOVBUS:
			if (IO_GetInput(USB_HOST_DETECT)) {
				/* start polling again */
				usb_flag   = 0;
				usb_hstate = USB_STATE_UNDEF;
			}
			break;
	}
	return usb_hstate;
}

void buspower_SetUsbCurrent(int cur)
{
	IO_Deactivate(USB_SUSPEND_OUT);
	IO_Activate(USB_HP);
	if (cur > 500)
		IO_Activate(USB_PWR_BYPASS);
	else
		IO_Deactivate(USB_PWR_BYPASS);
}

int buspower_ModeDetect()
{
	/* Note: can be called from timer interupt context ! */
	if (IO_GetInput(USB_HOST_DETECT)) {
		/* Vbus present */
		if (!IO_HaveUsbBusPowered()) {
			/* no buspower feature, simple VBUS detect case */
			return USB_MODE_DEVICE;
		} else {
			/* check if we receive reset from host */
			switch (usb_hstate) {
				case USB_STATE_UNDEF:
					buspower_SetUsbCurrent(100); // Host detected, no conf yet
					return 0;
				case USB_STATE_HOST:
					/* USB reset has been received -> host has been detected */
					/* but, do no extra power until user chooses to reboot */
					return USB_MODE_HOST;
				case USB_STATE_CONFIGURED:
					buspower_SetUsbCurrent(500); 	// Host configured
					return USB_MODE_DEVICE;
				case USB_STATE_CLA:
					/* Connected to external power without USB host, set low current when docked to Newcastle else high current */
					if (IO_GetDockState() == GODOCK_RADIO)
						buspower_SetUsbCurrent(500);
					else
						buspower_SetUsbCurrent(2000);
					return USB_MODE_CLA;
				case USB_STATE_NOVBUS:
					buspower_SetUsbCurrent(100);
					return USB_MODE_NONE;
				default:
					buspower_SetUsbCurrent(100);
					return 0;
			}
		}
	} else {
		usb_hstate = USB_STATE_NOVBUS;
		/* Vbus power not present */
		return USB_MODE_NONE;
	}
}

void buspower_poll(void)
{
	buspower_PollCLA();
}

int buspower_suspend(struct device *_dev)
{
	/* simple case: suspend always 500 mA mode */
	buspower_SetUsbCurrent(500); 
	
	return 0;
}

int buspower_resume( struct device *_dev )
{
	buspower_hw_init( );
	return 0;
}
