/* drivers/barcelona/gps/gps.c
 *
 * Implementation of the GPS driver.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Authors:
 * Jeroen Taverne <jeroen.taverne@tomtom.com>
 * Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* Includes */ 
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/hardware/clock.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/tomtomgo-irq.h>
#include <barcelona/Barc_gps.h>
#include <barcelona/gopins.h>
#include <barcelona/gotype.h>
#include <barcelona/debug.h>
#include <linux/delay.h>
/* Defines */
#define PFX "gps: "
#define PK_DBG PK_DBG_FUNC
#define IH_DBG PK_DBG_NONE

/* Kernel resources */
static int gps_irq;

/* Local variables */
static struct timeval gps_stamp;
static rwlock_t gps_stamp_lock = RW_LOCK_UNLOCKED;

void gps_get_timestamp(struct timeval *tv)
{
	unsigned long flags;

	read_lock_irqsave(&gps_stamp_lock, flags);
	memcpy(tv, &gps_stamp, sizeof(struct timeval));
	read_unlock_irqrestore(&gps_stamp_lock, flags);
}
EXPORT_SYMBOL(gps_get_timestamp);

irqreturn_t gps_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned long flags;

	write_lock_irqsave(&gps_stamp_lock, flags);
	do_gettimeofday(&gps_stamp);
	write_unlock_irqrestore(&gps_stamp_lock, flags);

	IH_DBG("%8ld.%06ld\n", gps_stamp.tv_sec, gps_stamp.tv_usec);

	return IRQ_HANDLED;
}

static inline void gps_power(unsigned on)
{
	PK_DBG("Switching GPS device %s\n", on ? "on" : "off");

	if (on) IO_Activate(GPS_ON); else IO_Deactivate(GPS_ON);
}

static inline void gps_reset(void)
{
	PK_DBG("Resetting GPS device\n");

	IO_Activate(GPS_RESET);	/* will reset BT as well on classc,malaga,atlanta */
	mdelay(50);
	IO_Deactivate(GPS_RESET);	
}

static inline void gps_update(unsigned on)
{
	PK_DBG("%sabling GPS firmware update\n", on ? "En" : "Dis");

	if (on)
		IO_Activate(GPS_REPRO);
	else
		IO_Deactivate(GPS_REPRO);
	gps_reset();
}

static int gps_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int ret = 0;
	unsigned long flags;

	switch (cmd) {
	case IOW_GPS_PROGRAM_MODE:
		gps_update(1);
		break;
	case IOW_GPS_NORMAL_MODE:
		gps_update(0);
		break;
	case IOW_GPS_ON:
		PK_WARN("WARNING: Ignoring IOW_GPS_ON\n");
		//gps_power(1);
		break;
	case IOW_GPS_OFF:
		PK_WARN("WARNING: Ignoring IOW_GPS_OFF\n");
		//gps_power(0);
		break;
	case IOW_GPS_RESET:
		PK_WARN("WARNING: Not Ignoring IOW_GPS_RESET\n");
		gps_reset();
		break;
	case IOW_GPS_UPDATE_ON:
		PK_WARN("WARNING: Ignoring IOW_GPS_UPDATE_ON\n");
		//gps_update(1);
		break;
	case IOW_GPS_UPDATE_OFF:
		PK_WARN("WARNING: Ignoring IOW_GPS_UPDATE_OFF\n");
		//gps_update(0);
		break;
	case IOR_GET_TIME_STAMP:
		if (!access_ok(VERIFY_WRITE, (void __user *) arg, sizeof gps_stamp)) {
			ret = -EFAULT;
		} else {
			read_lock_irqsave(&gps_stamp_lock, flags);
			ret = __copy_to_user((void __user *) arg, &gps_stamp, sizeof gps_stamp) ? -EFAULT : 0;
			read_unlock_irqrestore(&gps_stamp_lock, flags);
		}
		break;
	default:
		PK_WARN("Invalid ioctl command %u\n", cmd);
		ret = -EINVAL;
		break;
	}

	return ret;
}

static int gps_open(struct inode *inode, struct file *file)
{
	return nonseekable_open(inode, file);
}

static int gps_release(struct inode *inode, struct file *file)
{
	return 0;
}

/* Kernel interface */
static struct file_operations gps_fops = {
	.owner		= THIS_MODULE,
	.ioctl		= gps_ioctl,
	.open		= gps_open,
	.release	= gps_release,
};

extern unsigned int reset_state;

static void gps_hw_init(void)
{
	/* Config bluetooth UART signals */
	IO_SetFunction(TXD_BT);
	IO_SetFunction(RXD_BT);

	if (IO_HaveGlDetected()) {
		IO_SetFunction2(RTS_GPS);     /* Deactivate signal, i.e. enable RTS  */
		IO_SetFunction2(CTS_GPS);

      		gps_power(1);

		/* As Cork only has this pin, it will only be activated for it */
		IO_Activate(GPS_POWERON_OUT);
	} else {
		//GPS_REPRO == reprogramming is people will reprogram
		//the gps they will use the ioctl calls to program the device
		IO_Deactivate(GPS_REPRO);

		IO_SetFunction(RTS_BT);
		IO_SetFunction(CTS_BT);

		/* Turn on all GPS pins */
		IO_SetInterruptOnActivation(GPS_1PPS);

      		gps_power(1);

		PK_DBG(KERN_INFO "gps_hw_init:IO_HaveGlDetected=%d SiRF detected\n",IO_HaveGlDetected());
	}

	IO_SetFunction(RXD_GPS);
	IO_SetFunction(TXD_GPS);

	if (!IO_HaveGlDetected()) {
		gps_reset();
	} else {
		static int firsttime = 1;
		if (firsttime) {
			firsttime = 0;

			if (reset_state) {
				gps_reset();
                   		PK_DBG(KERN_INFO "gps_hw_init: the reset button pressed,GPS_RESET is on\n");
			}
		} else {
			IO_Deactivate(GPS_RESET);
		}
	}
}

static void gps_hw_exit(void)
{
    /* Turn off all GPS pins, exept GPS-nRESET_RF pin,
     * so Trident board is hooked now it neen not resets the
     * Real Time Clock that we have in the chip.
     */

	if (IO_HaveGlDetected()) {
                IO_Deactivate(GPS_POWERON_OUT);

                IO_Deactivate(RTS_GPS);
                IO_Deactivate(TXD_GPS);

		IO_Deactivate(GPS_ON);
	} else {
		IO_Deactivate(GPS_REPRO);
		IO_Deactivate(RTS_GPS);
		IO_Deactivate(TXD_GPS);
		IO_Deactivate(GPS_ON);
		IO_Deactivate(RXD_GPS);
		IO_Deactivate(GPS_1PPS);
	}
}

static int gps_probe(struct device *dev)
{
	int ret;

	gps_hw_init();

        if (!IO_HaveGlDetected() && IO_HasPin(GPS_1PPS)) {
		PK_DBG("Requesting GPS_PPS IRQ\n");
		ret = request_irq(IO_GetInterruptNumber(GPS_1PPS), gps_irq_handler, 0, "GPS PPS", dev);
		if (ret != 0) {
			PK_ERR("Failed to request timer IRQ (%d)\n", ret);
			return ret;
		}
		gps_irq = 1;
	} else {
		gps_irq = 0;
           	PK_DBG("gps_probe: no IRQ!\n");
	}

	PK_DBG("Registering chardev\n");
	ret = register_chrdev(GPS_MAJOR, GPS_DEVNAME, &gps_fops);
	if (ret != 0) {
		PK_ERR("Unable to register chardev on major=%d (%d)\n", GPS_MAJOR, ret);
		return ret;
	}

	PK_DBG("Done\n");
	return 0;
}

static int gps_remove(struct device *dev)
{
	PK_DBG("Unregistering chardev\n");
	unregister_chrdev(GPS_MAJOR, GPS_DEVNAME);

	if (gps_irq != 0) {
		PK_DBG("Freeing GPS_PPS IRQ\n");
		free_irq(TOMTOMGO_IRQ_GPS_PPS, dev);
		gps_irq = 0;
	}

	gps_hw_exit();

	PK_DBG("Done\n");
	return 0;
}

static void gps_shutdown(struct device *dev)
{
	PK_DBG("Shutting down\n");
	gps_hw_exit();
}

#ifdef CONFIG_PM

static int gps_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);
	if (level == SUSPEND_POWER_DOWN) {
		gps_hw_exit();
	}
	return 0;
}

static int gps_resume(struct device *dev, u32 level)
{
	PK_DBG("dev = %p, level = %u\n", dev, level);
	if (level == RESUME_POWER_ON) {

    if (IO_GetModelId() == GOTYPE_MARIGOT)
    {
      //! In marigot plateform the power supply of VDD_5V does not support high load
      //! of current at startup. This problem is due to the C1 sample.
      //! In case of overload VDD_5V = 2.6 Volts
      //! As a result the CAN driver is off.
      //! A delay authorizes a decrease of this current overload. 
      //! An other delay is added after power up : see below.
      mdelay(20);
    }
    gps_hw_init();

    if (IO_GetModelId() == GOTYPE_MARIGOT)
    {
      //! Current ovelaod limitation :See above for the explanation
      mdelay(20);
    }

  }
	return 0;
}

#else /* CONFIG_PM */
#define gps_suspend NULL
#define gps_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver gps_driver = {
	.name		= "tomtomgo-gps",
	.bus		= &platform_bus_type,
	.probe		= gps_probe,
	.remove		= gps_remove,
	.shutdown	= gps_shutdown,
	.suspend	= gps_suspend,
	.resume		= gps_resume,
};

static int __init gps_mod_init(void)
{
	int ret;

	printk(KERN_INFO "TomTom GO GPS Driver, (C) 2004,2005 TomTom BV\n");
	PK_DBG("Registering driver\n");
	ret = driver_register(&gps_driver);
	if (ret) {
		PK_ERR("Unable to register driver (%d)\n", ret);
		return ret;
	}
	PK_DBG("Done\n");
	return 0;
}

static void __exit gps_mod_exit(void)
{
	PK_DBG("Unregistering driver\n");
	driver_unregister(&gps_driver);
	PK_DBG("Done\n");
}

module_init(gps_mod_init);
module_exit(gps_mod_exit);

MODULE_AUTHOR("Dimitry Andric <dimitry.andric@tomtom.com>");
MODULE_DESCRIPTION("TomTom GO GPS Driver");
MODULE_LICENSE("GPL");

/* EOF */
