/* drivers/barcelona/micro/micro.c
*
* Implementation of the CAN-microprocessor driver.
*
* Copyright (C) 2007 TomTom BV <http://www.tomtom.com/>
* Author  Peter Wouters <peter.wouters.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*/

/* Includes */ 
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/hardware/clock.h>
#include <asm/delay.h>
#include <barcelona/Barc_micro.h>
#include <barcelona/debug.h>
#include <barcelona/gopins.h>

// Defines
#define PFX "micro: "
#define PK_DBG PK_DBG_FUNC

// Function prototypes
static int activate_micro(void);
static int deactivate_micro(void);

static struct fasync_struct *fasync_queue;
static int irqIsSet = 0;

static int activate_micro(void)
{
  // Nothing to do (yet)
  return 0;
}

static int deactivate_micro(void)
{
  // Nothing to do (yet)
  return 0;
}

static int micro_ioctl(struct inode *inode, 
                       struct file *file, 
                       unsigned int cmd,
                       unsigned long arg)
{ 
  int ret = 0;
    
  switch (cmd) {
    case IO_nCAN_RST:
    /* Deactivate or Activate the GPIO pin according to the	
    * argument.
    */
    switch (arg) {
    case 0:
      IO_Activate(CAN_RESETOUT);
      break;
    case 1:
      IO_Deactivate(CAN_RESETOUT);
      break;
    default:
      ret = -EINVAL;
      break;
    }
    break;
    
    case IO_CAN_SYNC:
    /* Pulse the SYNC pin
    */
      IO_Deactivate(CAN_SYNC);
      udelay(1); // Low Pulse width of 1 usec 
      IO_Activate(CAN_SYNC);
      break;
  
    case IO_CAN_BT_MD:
    /* Deactivate or Activate the GPIO pin according to the	
      * argument.
    */
      switch (arg) {
        case 0:
          IO_Deactivate(CAN_BT_MODE);
          break;
        case 1:
          IO_Activate(CAN_BT_MODE);
          break;
        default:
          ret = -EINVAL;
          break;
      }
      break;

    case IO_PWR_5V_ON_MCU:
    /* Deactivate or Activate the GPIO pin according to the	
      * argument.
    */
      switch (arg) {
        case 0:
          IO_Deactivate(PWR_5V_ON_MCU);
          break;
        case 1:
          IO_Activate(PWR_5V_ON_MCU);
          break;
        default:
          ret = -EINVAL;
          break;
      }
      break;

    case IO_MCU_nPWR_ON:
    /* Deactivate or Activate the GPIO pin according to the	
      * argument.
    */
      switch (arg) {
        case 0:
          IO_Deactivate(MCU_PWR_ON);
          break;
        case 1:
          IO_Activate(MCU_PWR_ON);
          break;
        default:
          ret = -EINVAL;
          break;
      }
      break;
   
  case IO_CAN_nRST:
    ret = put_user(IO_GetInput(CAN_RESETIN), (int __user *)arg);
    break;

  case IO_CAN_RES:
    ret = put_user(IO_GetInput(CAN_RESERV), (int __user *)arg);
    break;
    
  default:
    PK_WARN("Micro: Invalid ioctl command %u\n", cmd);
    ret = -EINVAL;
    break;
  }
  return ret;
}

irqreturn_t micro_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  kill_fasync(&fasync_queue, SIGIO, POLL_IN);

  return IRQ_HANDLED;
}

static int micro_fasync(int fd, struct file *fp, int mode)
{
  return fasync_helper(fd, fp, mode, &fasync_queue);
}

static int micro_open(struct inode *inode, struct file *file)
{
  int ret;

  /* install interrupt handler(s) for GPIO interrupt(s)                   */
  /* note: currently only CAN_RESETIN to detect CAN microcontroller reset */
  if (!irqIsSet) {
    ret = request_irq( IO_GetInterruptNumber(CAN_RESETIN), micro_interrupt, 
                       0,  "tomtomgo-canmicro", NULL );
    if (ret != 0) {
      PK_ERR("Unable to install CAN micro interrupt handler(%d)\n", ret);
    }
    else
    {
      /* now tell the hardware to generate interrupts when pin is activated */
      IO_SetInterruptOnActivation(CAN_RESETIN);

      irqIsSet = 1;
    }
  }

  return nonseekable_open(inode, file);
}

static int micro_release(struct inode *inode, struct file *file)
{
  if (irqIsSet) {
    /* here we should disable interrupt generation in the hardware, */
    /* unfortunately there is no functionality to do this.          */
    /* Linux will ignore the interrupt.                             */
    ;

    free_irq(IO_GetInterruptNumber(CAN_RESETIN), NULL);

    irqIsSet = 0;
  }

  micro_fasync(-1, file, 0);

  return 0;
}

/* Kernel interface */
static struct file_operations micro_fops = {
  .owner   = THIS_MODULE,
  .ioctl   = micro_ioctl,
  .open	   = micro_open,
  .release = micro_release,
  .fasync  = micro_fasync,
};

static int micro_probe(struct device *dev)
{
  return 0;
}

static int micro_remove(struct device *dev)
{
  deactivate_micro();
  return 0;
}

static void micro_shutdown(struct device *dev)
{
  PK_DBG("Shutting down micro\n");
  deactivate_micro();
}

#ifdef CONFIG_PM

static int micro_suspend(struct device *dev, u32 state, u32 level)
{
  PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);	
  if (level == SUSPEND_POWER_DOWN) 
  {
    deactivate_micro();
  }
  return 0;
}

static int micro_resume(struct device *dev, u32 level)
{
  PK_DBG("dev = %p, level = %u\n", dev, level);	
  if (level == RESUME_POWER_ON) 
  { 
    activate_micro();
  }
  return 0;
}

#else /* CONFIG_PM */
#define micro_suspend NULL
#define micro_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver micro_driver = {
  .name		  = "tomtomgo-canmicro",
  .bus	    = &platform_bus_type,
  .probe		= micro_probe,
  .remove		= micro_remove,
  .shutdown	= micro_shutdown,
  .suspend	= micro_suspend,
  .resume		= micro_resume,
};

static int __init micro_mod_init(void)
{
  int ret;

  fasync_queue = NULL;

  printk(KERN_INFO "TomTom CAN-Microprocessor Driver, (C) 2008 TomTom BV\n");

  PK_DBG("Registering driver\n");
  ret = driver_register(&micro_driver);
  if (ret) {
    PK_ERR("Unable to register driver (%d)\n", ret);
    return ret;
  }
  

  PK_DBG("Registering chardev\n");
  ret = register_chrdev(MICRO_MAJOR, MICRO_DEVNAME, &micro_fops);
  if (ret != 0) {
    PK_ERR("Unable to register chardev on major=%d (%d)\n", MICRO_MAJOR, ret);
    return ret;
  }

  activate_micro();
  
  return ret;
}

static void __exit micro_mod_exit(void)
{
  deactivate_micro();
  
  PK_DBG("Unregistering chardev\n");
  unregister_chrdev(MICRO_MAJOR, MICRO_DEVNAME);

  PK_DBG("Unregistering driver\n");
  driver_unregister(&micro_driver);
}

module_init(micro_mod_init);
module_exit(micro_mod_exit);

MODULE_AUTHOR("Peter Wouters <peter.wouters@tomtom.com>");
MODULE_DESCRIPTION("TomTom CAN-Microprocessor Driver");
MODULE_LICENSE("GPL");

/* EOF */ 
