/* si4703.c
 *
 * Control driver for Silicon Lab 4703 chip.
 *
 * Copyright (C) 2006 TomTom BV <http://www.tomtom.com/>
 * Authors: Xander Hover <Xander.Hover@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/i2c.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <asm/uaccess.h>
#include <linux/fmreceiver.h>
#include <linux/si4703.h>
#include <linux/device.h>
#include <barcelona/gopins.h>
#include <linux/delay.h>
#include <asm/semaphore.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <asm/irq.h>
#include <asm/arch/irqs.h>
#include <linux/wait.h> /* needed to inmplement blocking read and wait queues */
#include <linux/poll.h> /* needed for poll_wait() call */
#include <linux/kfifo.h> /* needed for the kfifo ringbuffer */


#define PFX "SI4703: "
#define PK_DBG(fmt, arg...)    printk(KERN_DEBUG PFX "%s: " fmt, __FUNCTION__ ,##arg)
#define PK_ERR(fmt, arg...)    printk(KERN_ERR PFX "%s: " fmt, __FUNCTION__ ,##arg)
#define PK_ENTRY	PK_DBG("----------------------%s-------------------\n", __func__)

#define NO_ENTRY_PRINTS

#ifdef NO_ENTRY_PRINTS
#ifdef PK_ENTRY
#undef PK_ENTRY
#define PK_ENTRY do { } while (0)
#endif
#endif

struct si4703dev {
	wait_queue_head_t wq; /* wait queue for read */
	struct cdev cdev;
	struct i2c_client *client;
	struct semaphore sem; /* it may be possible that this is not needed */
	spinlock_t sp_lock;
	struct kfifo* fifo;
};

/* there shall only be one user of the device */
static atomic_t si4703_available = ATOMIC_INIT(1);
static int data_avail = 0 ;

/* pins */
//#warning XH dock_i2c_en_pin does not belong here.
static gopin_t tmcRX_dock_i2c_en_pin; /* this is temporary until the dock driver is updated */
static gopin_t tmcRX_reset_pin;	/* reset is active low */
static gopin_t tmcRX_sen_pin;	/* sen is active low */
static dev_t tmcRX_devno;
static unsigned int tmcRX_irq; /* the interrupt number of the device */
static u16 s_device_id = 0xFFFF;	/* value when device not recognized */

/* Forward declarations */
static struct i2c_client client_template;

/* Spec says SI4703 sits on address 0x10 */
static unsigned short normal_i2c[] =
    { SI4703_I2C_SLAVE_ADDR, I2C_CLIENT_END };

/* Magic definition of all other variables and things */
I2C_CLIENT_INSMOD;

static int si4703_attach(struct i2c_adapter *adapter);
static int si4703_detach(struct i2c_client *client);
static int si4703_i2c_probe(struct device *dev);

#ifdef CONFIG_PM
static int si4703_i2c_suspend(struct device *dev, pm_message_t state,
			      u32 level);
static int si4703_i2c_resume(struct device *dev, u32 level);
#else
#    define si4710_i2c_suspend  (NULL)
#    define si4710_i2c_resume   (NULL)
#endif				/* CONFIG_PM     */

static int si4703_powersuspend(struct i2c_client *client);
static int si4703_powerresume(struct i2c_client *client);

static int si4703_read_registers(struct i2c_client *c, const u8 n);
static int si4703_write_registers(struct i2c_client *c, const u8 n);
static int si4703_read_all_registers(struct i2c_client *c, u16* buf);
static int si4703_sync_registers(void);

static struct i2c_driver driver = {
	.id = I2C_DRIVERID_SI4703,
	.owner = THIS_MODULE,
	.name = "si4703_i2c",
	.flags = I2C_DF_NOTIFY,	/* I2C_DF_NOTIFY is gone in recent kernels */
	.attach_adapter = si4703_attach,
	.detach_client = si4703_detach,
	.driver = {
		   .probe = si4703_i2c_probe,
		   .suspend = si4703_i2c_suspend,
		   .resume = si4703_i2c_resume,
		   .name = SI4703_DEVNAME,
		   .owner = THIS_MODULE,
		   .bus = &platform_bus_type,
		   },
};

static struct i2c_client client_template = {
	.name = "si4703_i2c_client",
	.flags = I2C_CLIENT_ALLOW_USE,
	.usage_count = 0,
	.driver = &driver,
	.addr = SI4703_I2C_SLAVE_ADDR,
};


/* prototypes for the tuner commands */
static int si4703_get_freq(struct i2c_client *c, __u16 * freq,
			   __u16 * rssi);
static int si4703_set_freq(struct i2c_client *c, __u16 freq);
static int si4703_set_volume(struct i2c_client *c, __u16 volume);
static int si4703_set_band(struct i2c_client *c, __u16 band);
static int si4703_powerup(struct i2c_client *c);
static int si4703_powerdown(struct i2c_client *c);
static int si4703_seek(struct i2c_client *c,
		       struct fmr_seek_param_type seek_param);

static int si4703_enable_rds(struct i2c_client *c);
static int si4703_disable_rds(struct i2c_client* c);

static int si4703_fifo_alloc(struct si4703dev* dev, int size);
static void si4703_fifo_free(struct si4703dev* dev);
static void si4703_fifo_purge(struct si4703dev* dev);
static int si4703_fifo_len(struct si4703dev* dev);
static int si4703_fifo_put(struct si4703dev* dev, unsigned char* data, int len );
static int si4703_fifo_get(struct si4703dev* dev, unsigned char* data, int len );

/* The silly thing is that the read layout is different from
 * the write layout. The horror, the horror..
 */
struct si4703_read_type {
	u16 statusrssi;
	u16 readchan;
	u16 rdsa;
	u16 rdsb;
	u16 rdsc;
	u16 rdsd;
	u16 deviceid;
	u16 chipid;
	u16 powercfg;
	u16 channel;
	u16 sysconfig1;
	u16 sysconfig2;
	u16 sysconfig3;
	u16 test1;
	u16 test2;
	u16 bootconfig;
} __attribute__ ((__packed__));

struct si4703_write_type {
	u16 powercfg;
	u16 channel;
	u16 sysconfig1;
	u16 sysconfig2;
	u16 sysconfig3;
	u16 test1;
	u16 test2;
	u16 bootconfig;
	u16 statusrssi;
	u16 readchan;
	u16 rdsa;
	u16 rdsb;
	u16 rdsc;
	u16 rdsd;
	u16 deviceid;
	u16 chipid;
} __attribute__ ((__packed__));

/* have these packed */
static u16 si4703_read_regs[16];
static u16 si4703_write_regs[16];

static u16 current_freq = 0;
static u16 current_rssi = 0;

static struct si4703dev* gdev; /* global pointer to dereference irq later */

static void si4703_workqueue_handler(void* arg );

DECLARE_WORK(rds_data_grabber, si4703_workqueue_handler, (void*) NULL);

static void si4703_workqueue_handler(void* arg )
{
	u16 rdsdata[16];
	u16 lfreq = 0;
	u16 lrssi = 0;
	struct fmr_rds_type rdsoutdata;
	struct si4703_read_type* pdata = (struct si4703_read_type*) rdsdata;
	u16* pwreg = &rdsdata[8];

	do { } while ( si4703_read_all_registers(gdev->client, rdsdata) );
	{
		u8 seektune_complete = (pdata->statusrssi & 0x4000 ) ? 1: 0;
		u8 rdssync = (pdata->statusrssi & 0x0800 )?1:0;

		u8 band = (u8) ((pdata->sysconfig2 >> 6) & 0x03);
		u8 spacing = (u8) ((pdata->sysconfig2 >> 4) & 0x03);
		u16 channel = (u16) (pdata->readchan & 0x0003ff);

		u16 offset = 0;
		u16 delta = 0;
		u16 maxfreq = 0;

		rdsoutdata.sf_bandlimit = (pdata->statusrssi & 0x2000) ? 1: 0;
		rdsoutdata.rssi = (pdata->statusrssi & 0x00ff ) ;
		lrssi = rdsoutdata.rssi;

		switch (band) {
			default:
			case 0x0:
				offset = 8750;
				maxfreq = 10800;
				break;
			case 0x1:
				offset = 7600;
				maxfreq = 10800;
				break;
			case 0x2:
				offset = 7600;
				maxfreq = 9000;
				break;
		}

		switch (spacing) {
			default:
			case 0x0:
				delta = 20;
				break;
			case 0x1:
				delta = 10;
				break;
			case 0x2:
				delta = 5;
				break;
		}

		lfreq = channel * delta + offset;
		if ( lfreq < maxfreq ) rdsoutdata.sf_bandlimit = 0; /* remove false positives or seek_fails */

		rdsoutdata.freq = lfreq;
		rdsoutdata.blocka = 0x00FF;
		rdsoutdata.blockb = 0x0000;
		rdsoutdata.blockc = 0x0000;
		rdsoutdata.blockd = 0x0000;

		if ( seektune_complete || rdsoutdata.sf_bandlimit )
		{
			/* unset tune or seek */
			si4703_fifo_purge(gdev);
			si4703_fifo_put(gdev, (u8*) &rdsoutdata, sizeof(struct fmr_rds_type));

			pdata->powercfg = pdata->powercfg & ~0x0100; /*  disable seek bit */
			pdata->channel = pdata->channel & ~0x8000;   /* disable tune bit */

			int i = 0;
			for (i = 0; i < 16; i++) {
				rdsdata[i] = cpu_to_be16(rdsdata[i]);
			}
			i2c_master_send(gdev->client, (u8 *) pwreg, 4);

			goto exit_irq_handler;

		}

		/* RDS ready bit not working in verbose mode, so use rds sync only. */
		if ( rdssync ) /* if rds sync */
		{
			if ( !((rdsdata[0] & 0x0600) | (rdsdata[1] & 0xFC00)) )
			{ 	/* if we have any error in block a,b,c or d, discard whole group */
				if ( si4703_fifo_len(gdev) < (54 * sizeof(struct fmr_rds_type))) /* else so put group in fifo if it still fits */
				{
					rdsoutdata.blocka = pdata->rdsa;
					rdsoutdata.blockb = pdata->rdsb;
					rdsoutdata.blockc = pdata->rdsc;
					rdsoutdata.blockd = pdata->rdsd;
					si4703_fifo_put(gdev, (u8*) &rdsoutdata, sizeof(struct fmr_rds_type));
				}
			}
		}
	}

exit_irq_handler:
	down(&gdev->sem);

	data_avail = (si4703_fifo_len(gdev) > 0 ) ;
	current_freq = lfreq;
	current_rssi =  lrssi;

	wake_up(&gdev->wq);
	up(&gdev->sem);
}

static irqreturn_t si4703_interrupthandler(int irq, void* dev_id, struct pt_regs* regs )
{
	schedule_work(&rds_data_grabber);
	return IRQ_HANDLED;
}

static int si4703_i2c_probe(struct device *dev)
{
	struct fm_receiver_info *pdata = dev->platform_data;
	struct platform_device* pdev = to_platform_device(dev);

	PK_ENTRY;

  if(IO_GetTMCReceiverType() != GOTMC_SI4703)
  {
    PK_DBG("wrong type: %d vs %d\n", IO_GetTMCReceiverType(), GOTMC_SI4703);
    return -1;        
  }
        
        tmcRX_devno = pdata->device_nr;

	tmcRX_reset_pin = pdata->fm_reset_pin;
	tmcRX_sen_pin = pdata->fm_sen_pin;
	tmcRX_irq = (unsigned int) platform_get_irq(pdev, 0);
#warning XH: This should move to the cagliari dock driver
	tmcRX_dock_i2c_en_pin = pdata->fm_dock_i2c_en_pin;

	PK_DBG("SI4703 uses irq %d.\n", tmcRX_irq);

	return i2c_add_driver(&driver);
}

/* --- powermanagement functions --- */

#ifdef CONFIG_PM
static int
si4703_i2c_suspend(struct device *dev, pm_message_t state, u32 level)
{
#if 0
	struct i2c_client *c = to_i2c_client(dev);
	int rc = 0;

	if (level != SUSPEND_POWER_DOWN)
		return 0;

	if (c == NULL) {
		PK_ERR("Can't find I2C driver!\n");
		rc = -ENODEV;
	}
#endif
	PK_DBG
		("----------------------%s----------------pm_msg=%d, level=%d\n",
		 __func__, state, level);

#warning Quickfix for suspend crash in kernel
	//if (!rc) {
	IOP_Activate(tmcRX_reset_pin);
	msleep(10);

	if (tmcRX_dock_i2c_en_pin) {
		IOP_Deactivate(tmcRX_dock_i2c_en_pin);
		udelay(500);
	}
	//}

	return 0;
}


static int si4703_i2c_resume(struct device *dev, u32 level)
{
	struct i2c_client *c = to_i2c_client(dev);
	int rc = 0;

	if (level != RESUME_POWER_ON)
		return 0;

	if (c == NULL) {
		PK_ERR("Can't find I2C driver!\n");
		rc = -ENODEV;
	}

	PK_DBG("----------------------%s----------------level=%d\n",
	       __func__, level);
	if (!rc) {
		rc = si4703_powerresume(c);
	} else {
		PK_ERR("error while doing powerresume: rc = %d\n", rc);
	}
	return 0;
}

#endif

static int si4703_powersuspend(struct i2c_client *c)
{
/*
 * Powerdown sequence
 * 1. Write register 07h (example: 0x3C04, 0xBC04 or 0x7C04)
 * 2. Write register 02h
 * 	- Clear DMUTE
 * 	- Set ENABLE bit high to set the powerdown state
 * 	- Set DISABLE bit high to set the powerdown state
 * After the disable bit is set high, the device performs a powerdown sequence and
 * then sets the ENABLE and DISABLE bits low.
 * Example: Write 0x0041
 */

	/* store registers */
	struct si4703_write_type *pw =
	    (struct si4703_write_type *) si4703_write_regs;
	si4703_read_registers(c, 16);
	si4703_sync_registers();

	pw->test1 = 0xBC04;
	si4703_write_registers(c, 6);

	/* clear DMUTE, set ENABLE, set DISABLE */
	pw->powercfg &= ~0x4000;
	pw->powercfg |= 0x0041;
	si4703_write_registers(c, 1);	/* powercfg is the first written register so we only write one */

	si4703_read_registers(c, 16);
	si4703_sync_registers();

	return 0;
}

static int si4703_powerresume(struct i2c_client *c)
{
/*
 * Powerup sequence
 * 1.  Set XOSCEN (write 0x8100 or 0xBC04 to the register)
 * 2.  wait 500 ms for the oscillator to set
 * 3.  Set DMUTE bit to disable mute
 * 4.  Set ENABLE bit to high to set the powerup state
 * 5.  Set DISABLE bit to low to set the powerup state ( ex. 3..5: write 0x4001 )
 * 6.  Wait for device powerup
 * 7.  Read addresses 00h - 01h -> device_id, chip_id
 * 8.  Read addresses 02h - 0Fh -> store all registers in a local shadow copy.
 * 9.  Write remaining hardware config registers
 * 10. Write the general config registers
 * 11. Write the regional config registers
 */
	/* get the write registers */
	struct si4703_write_type *pw =
	    (struct si4703_write_type *) si4703_write_regs;
	/* read all regs */

	if (tmcRX_dock_i2c_en_pin) {
		IOP_Activate(tmcRX_dock_i2c_en_pin);
		udelay(500);
	}

	if (tmcRX_sen_pin) {
		IOP_Deactivate(tmcRX_sen_pin);
		udelay(500);
	}

	IOP_Activate(tmcRX_reset_pin);
	msleep(100);
	IOP_Deactivate(tmcRX_reset_pin);
	msleep(100);

	PK_ENTRY;
	si4703_read_registers(c, 16);
	si4703_sync_registers();

	/* manipulate for powerup */
	/* enable internal oscillator */
	pw->test1 |= 0x8100;
	si4703_write_registers(c, 6);
	PK_DBG("Waiting for oscillator to settle ... \n");
	msleep(500);

	si4703_read_registers(c, 16);
	si4703_sync_registers();
	pw->powercfg |= 0x4001 | 0xB000;
	pw->powercfg &= ~0x0040;
	si4703_write_registers(c, 1);
	msleep(100);		/* wait for the device to powerup -- 500 ms is overkill */

	si4703_read_registers(c, 16);
	si4703_sync_registers();

	/* reconfigure device after powerup */

	return 0;
}

/*--------- utility functions for reading ----------------------------------*/

/* read n registers and transform them to cpu endianness */
/* result is stored in global variable: si4703_read_regs */
/* note: global variable has cpu endian register values */
static int si4703_read_registers(struct i2c_client *c, const u8 n)
{
	int count = i2c_master_recv(c, (u8 *) si4703_read_regs, 2 * n);
	if (count < 0)
		return -EBUSY;
	else {
		int i;
		for (i = 0; i < n; i++) {
			si4703_read_regs[i] =
			    be16_to_cpu(si4703_read_regs[i]);
		}
	}
	return 0;
}

static int si4703_read_all_registers(struct i2c_client *c, u16* buf)
{
	int count = i2c_master_recv(c, (u8 *) buf, 32);
	if (count < 0)
		return -EBUSY;
	else {
		int i;
		for (i = 0; i < 16; i++) {
			buf[i] =
			    be16_to_cpu(buf[i]);
		}
	}
	return 0;
}

/* write n registers. Registers to be written are transformed from
 * cpu endianness to big endian first
 * Pre: global variable si4703_write_regs was manipulated first
 * Post: global variable si4703_write_reg [0,n) is written to i2c chip
 * note: global variable has cpu endian register values
 */
static int si4703_write_registers(struct i2c_client *c, const u8 n)
{
	u16 lreg[16];
	u16 *pSi4703_write_regs = (u16 *) & si4703_write_regs;
	int i;
	for (i = 0; i < n; i++) {
		lreg[i] = cpu_to_be16(pSi4703_write_regs[i]);
	}
	i2c_master_send(c, (u8 *) lreg, 2 * n);
	return 0;
}

/* Copies shadow registers (read to write) */
/* Rearranges the register order for writing */
static int si4703_sync_registers(void)
{
	int i = 0, j = 0, k = 0;
	for (i = 10, j = 2; k < 16; i = (++i) % 16, j = (++j) % 16, k++) {
		si4703_write_regs[j] = si4703_read_regs[i];
	}
	return 0;
}

static int si4703_enable_rds(struct i2c_client *c)
{
	PK_ENTRY;
	if (si4703_read_registers(c,16))
		return -1;
	si4703_sync_registers();
	{
		struct si4703_write_type* pregs =
			(struct si4703_write_type*) &si4703_write_regs;

		pregs->powercfg |= 0x0800;  /* set RDS verbose mode */
		pregs->sysconfig1 |= 0xD004; /* set RDS enable, set RDSI enable, set STCI enable, set GPIO2 == RDS/STC Interrupt */

		if (si4703_write_registers(c,5)) return -1;
	}
	return 0;
}

static int si4703_disable_rds(struct i2c_client* c)
{
	PK_ENTRY;
	if (si4703_read_registers(c,16))
		return -1;
	si4703_sync_registers();
	{
		struct si4703_write_type* pregs =
			(struct si4703_write_type*) &si4703_write_regs;

		pregs->sysconfig1 &= ~0x1000; /* unset RDS enable */
		/* pregs->sysconfig1 &= ~0x8000; */ /* unset RDSI enable */
		if (si4703_write_registers(c,3)) return -1;
	}
	return 0;
}

static int si4703_set_band(struct i2c_client *c, __u16 band)
{
	PK_ENTRY;
	if (si4703_read_registers(c, 16))
		return -1;
	si4703_sync_registers();
	{
		struct si4703_write_type *pregs =
		    (struct si4703_write_type *) &si4703_write_regs;
		u16 spacing;
		u16 deemphasis;
		switch (band) {
		case 0U:
			{
				/* Europe, 87.5 - 108.0 MHz, 100 kHz grid */
				band =  0x0000;
				spacing = 0x0010;
				deemphasis = 0x0100;
			}
			break;
		case 1U:
			{
				/* USA, 87.5 - 108.0 MHz, 200 kHz grid */
				band =  0x0000;
				spacing = 0x0000;
				deemphasis = 0x0000;
			}
			break;
		case 2U:
			{
				/* Japan, 76.0 - 90.0 MHz, 100 kHz grid */
				band = 0x0080;
				spacing = 0x0010;
				deemphasis = 0x0100;
			}
			break;
		default:
			/* Limbo */
			return -1;
		}
		pregs->sysconfig2 =
		    (pregs->sysconfig2 & ~0x00B0) | band | spacing;
		pregs->sysconfig1 =
		    (pregs->sysconfig1 & ~0x0100) | deemphasis;
		if (si4703_write_registers(c, 4))
			return -1;
	}
	return 0;
}

static int si4703_set_volume(struct i2c_client *c, u16 volume)
{
	PK_ENTRY;
	if (si4703_read_registers(c, 16))
		return -1;
	si4703_sync_registers();
	{
		struct si4703_write_type *pregs =
		    (struct si4703_write_type *) &si4703_write_regs;
		volume &= 0x000F;
		pregs->sysconfig2 = (pregs->sysconfig2 & ~0x000F) | volume;
		if (si4703_write_registers(c, 4))
			return -1;
	}
	return 0;
}


static int si4703_get_freq(struct i2c_client *c, __u16 * freq,
			   __u16 * rssi)
{
	PK_ENTRY;

	down(&gdev->sem);
	if (current_freq >= 7600 && current_freq <= 10800 ) // we have a valid freq range
	{
		*freq = current_freq;
		*rssi = current_rssi;
		up(&gdev->sem);
		return 0;
	}
	else {
		up(&gdev->sem);

		if (si4703_read_registers(c, 16))
			return -1;
		{
			struct si4703_read_type *pregs =
				(struct si4703_read_type *) si4703_read_regs;
			u8 band = (u8) ((pregs->sysconfig2 >> 6) & 0x03);
			u8 spacing = (u8) ((pregs->sysconfig2 >> 4) & 0x03);
			u16 offset = 0;
			u16 delta = 0;
			u16 maxfreq = 0;
			u16 channel = (u16) (pregs->readchan & 0x0003ff);

			switch (band) {
				case 0x0:
					offset = 8750;
					maxfreq = 10800;
					break;
				case 0x1:
					offset = 7600;
					maxfreq = 10800;
					break;
				case 0x2:
					offset = 7600;
					maxfreq = 9000;
					break;
				default:
					return -1;
			}

			switch (spacing) {
				case 0x0:
					delta = 20;
					break;
				case 0x1:
					delta = 10;
					break;
				case 0x2:
					delta = 5;
					break;
				default:
					return -1;
			}

			*freq = channel * delta + offset;
			*rssi = pregs->statusrssi & 0x00FF;
		}
	}
	return 0;
}

static int si4703_set_freq(struct i2c_client *c, __u16 freq)
{
	PK_ENTRY;

	if (si4703_read_registers(c, 16))
		return -1;
	si4703_sync_registers();

	{

		struct si4703_write_type *pwregs =
		    (struct si4703_write_type *) si4703_write_regs;

		u8 band = (u8) ((pwregs->sysconfig2 >> 6) & 0x03);
		u8 spacing = (u8) ((pwregs->sysconfig2 >> 4) & 0x03);
		u16 offset = 0;
		u16 delta = 0;
		u16 maxfreq = 0;
		u16 channel = 0;

		switch (band) {
		case 0x0:
			offset = 8750;
			maxfreq = 10800;
			break;
		case 0x1:
			offset = 7600;
			maxfreq = 10800;
			break;
		case 0x2:
			offset = 7600;
			maxfreq = 9000;
			break;
		default:
			return -1;
		}

		switch (spacing) {
		case 0x0:
			delta = 20;
			break;
		case 0x1:
			delta = 10;
			break;
		case 0x2:
			delta = 5;
			break;
		default:
			return -1;
		}

		if (freq > maxfreq)
			return -1;
		if (freq < offset)
			return -1;

		channel = (freq - offset) / delta;
		pwregs->channel =
		    (pwregs->channel & 0xFC00) | (channel & 0x3FF);
		if (si4703_write_registers(c, 2))
			return -1;
		pwregs->channel |= 0x8000;
		if (si4703_write_registers(c, 2))
			return -1;
	}

	return 0;
}

static int si4703_powerup(struct i2c_client *c)
{
	PK_DBG("Before powerresume in powerup!\n");

	if (si4703_powerresume(c))
		goto resume_error;
	/* is this the right chip? */

	PK_DBG("After powerresume\n");
	PK_DBG("XOSEN in register test1 is enabled\n");

	{
		struct si4703_read_type *pregs =
		    (struct si4703_read_type *) &si4703_read_regs;
		u16 n = 0U;
		if (pregs->deviceid != 0x1242)
			goto invalid_chip;
		n = (pregs->chipid >> 6) & 0xF;

		PK_DBG("devid = 0x%02X!\n", n);

		switch (n) {
		case 1U:
			s_device_id = 4702U;
			break;
		case 8U:
			PK_DBG("BAD REFCLK!\n");
		case 9U:
			s_device_id = 4703U;
			break;
		default:
			goto invalid_chip;
			break;
		}
	}
	if (si4703_enable_rds(c)) return -1;
	return 0;

      resume_error:
	PK_DBG("Power resume failed!\n");
	return -1;

      invalid_chip:
	PK_DBG("Invalid Chip!\n");
	s_device_id = 0xFFFF;
	return 0;
}

static int si4703_powerdown(struct i2c_client *c)
{
	if (si4703_powersuspend(c))
		return -1;
	return 0;
}

static int si4703_seek(struct i2c_client *c,
		       struct fmr_seek_param_type seek_param)
{
	PK_ENTRY;

	if ( current_freq < seek_param.freq ) /* we want to seek from other point of origin */
	{
		if (si4703_set_freq(c, (__u16) seek_param.freq))
			return -1;
		msleep(5);
	}

	if (si4703_read_registers(c, 16))
		return -1;
	si4703_sync_registers();
	{
		struct si4703_write_type *pregs =
		    (struct si4703_write_type *) &si4703_write_regs;

		if (seek_param.up)
			pregs->powercfg |= 0x0200;   /* seek direction = up */
		else
			pregs->powercfg &= ~0x0200;  /* seek direction = down */
		if (seek_param.wrap)
			pregs->powercfg &= ~0x0400;  /* wrap around fm band bounds */
		else
			pregs->powercfg |= 0x0400;  /* stop at fm band bounds */
		pregs->sysconfig2 |=  ((u16) (seek_param.threshold << 8) & 0xFF00);

		pregs->sysconfig3 |= 0x00A0 ; /* set Seek SNR to middle */

		if (si4703_write_registers(c, 5))
			return -1;

		msleep(10);

		pregs->powercfg |= 0x0100;	    /* set the seek flag to start the seek operation */
		if (si4703_write_registers(c, 1))
			return -1;
	}
	return 0;
}

static int si4703_fifo_alloc(struct si4703dev* dev, int size)
{
	dev->fifo = kfifo_alloc(size, GFP_KERNEL, &dev->sp_lock);
	if (dev->fifo != NULL ) return 0;
	return -1;
}

static void si4703_fifo_free(struct si4703dev* dev)
{
	kfifo_free(dev->fifo);
}

static void si4703_fifo_purge(struct si4703dev* dev)
{
	kfifo_reset(dev->fifo);
}

static int si4703_fifo_len(struct si4703dev* dev)
{
	return kfifo_len(dev->fifo);
}

static int si4703_fifo_put(struct si4703dev* dev, unsigned char* data,  int len )
{
	return kfifo_put(dev->fifo, data, len );
}

static int si4703_fifo_get(struct si4703dev* dev, unsigned char* data,  int len )
{
	return kfifo_get(dev->fifo, data, len);
}

static int si4703_open(struct inode *nodep, struct file *filep)
{
	struct si4703dev *dev =
	    container_of(nodep->i_cdev, struct si4703dev, cdev);

	PK_ENTRY;

	/* Only one client may open the device */
	if (!atomic_dec_and_test(&si4703_available)) {
		atomic_inc(&si4703_available);
		return -EBUSY;
	}

	filep->private_data = dev;	/* for easy reference */
#warning CAGLIARI HACK! Powercycle the silabs chip due to crossdepency with dock driver!
	si4703_powerup(dev->client);
	msleep(50);
// remove until here

	si4703_enable_rds(dev->client);

	return 0;
}

static int si4703_release(struct inode *nodep, struct file *filep)
{
	PK_ENTRY;
	struct si4703dev *dev = (struct si4703dev*) filep->private_data;

	if (filep && filep->private_data) {
		filep->private_data = NULL;
	}

	// si4703_disable_rds(dev->client);

	si4703_fifo_purge(dev);

	/* make device available for others to open */
	atomic_inc(&si4703_available);

	return 0;
}

static int
si4703_ioctl(struct inode *nodep, struct file *filep,
	     unsigned int cmd, unsigned long arg)
{
	struct si4703dev *dev = filep ? filep->private_data : NULL;

	/* PK_DBG("----------------------%s----------- cmd=0x%x\n",
	       __func__, cmd); */
	if (_IOC_TYPE(cmd) != FMRECEIVER_DRIVER_MAGIC) {
		PK_ERR("Wrong IOC type! Failing command\n");
		return -ENOTTY;
	}
	switch (cmd) {
	case TMCIOC_RDS_START:
		{
			if (si4703_enable_rds(dev->client))
				return -EFAULT;
		}
		break;
	case TMCIOC_RDS_STOP:
		{
			if (si4703_disable_rds(dev->client))
				return -EFAULT;
			si4703_fifo_purge(dev);
		}
		break;
	case TMCIOC_G_FREQUENCY:
		{
			__u16 freq;
			__u16 rssi;
			if (si4703_get_freq(dev->client, &freq, &rssi))
				return -EFAULT;
			if (put_user(freq, (__u16 __user *) arg))
				return -EFAULT;
		}
		break;
	case TMCIOC_S_FREQUENCY:
		{
			__u16 freq;
			if (get_user(freq, (__u16 __user *) arg))
				return -EFAULT;
			if (si4703_set_freq(dev->client, freq))
				return -EFAULT;
		}
		break;
	case TMCIOC_S_BAND:
		{
			__u16 band;
			if (get_user(band, (__u16 __user *) arg))
				return -EFAULT;
			if (si4703_set_band(dev->client, band))
				return -EFAULT;
		}
		break;
	case TMCIOC_S_VOLUME:
		{
			__u16 volume;
			if (get_user(volume, (__u16 __user *) arg))
				return -EFAULT;
			if (si4703_set_volume(dev->client, volume))
				return -EFAULT;
		}
		break;
	case TMCIOC_SEEK:
		{
			struct fmr_seek_param_type param;
			if (copy_from_user
			    (&param, (void __user *) arg,
			     sizeof(struct fmr_seek_param_type)))
				return -EFAULT;
			if (si4703_seek(dev->client, param))
				return -EFAULT;
		}
		break;
	case TMCIOC_QUERY_DEVICE_ID:
		{
			__u16 devid = s_device_id;
			if (put_user(devid, (__u16 __user *) arg))
				return -EFAULT;
		}
		break;
	case TMCIOC_Q_RSSI:
		{
			__u16 rssi;
			__u16 freq;
			if (si4703_get_freq(dev->client, &freq, &rssi))
				return -EFAULT;
			if (put_user(rssi, (__u16 __user *) arg))
				return -EFAULT;
		}
		break;
	default:
		{
			/* other commands not supported by this device */
			return -ENOTTY;
		}
		break;
	}
	return 0;
}

/* we are inmplementing poll to support the select/poll/epoll system calls
 * This is done to allow the userspace app to check whether it is possible
 * to read non-blocking */
static unsigned int
si4703_poll(struct file *filep, struct poll_table_struct *pq)
{
	struct si4703dev *dev = filep->private_data;
	unsigned int mask = 0 ;

	PK_ENTRY;

	down(&dev->sem);
	if (data_avail)
		mask |= POLLIN | POLLRDNORM ; /* readable */
	else  {
		poll_wait(filep, &dev->wq, pq);
		if (data_avail)
			mask |= POLLIN | POLLRDNORM ; /* readable */
	}
	up(&dev->sem);

	return mask;
}

/* read blocks when there is not data, as it should */
static ssize_t si4703_read(struct file *filep, char __user* buf, size_t count, loff_t* pos )
{
	struct si4703dev* dev = filep->private_data;

	struct fmr_rds_type rdsdata;

	count = sizeof(struct fmr_rds_type);

	if ( down_interruptible(&dev->sem)) return -ERESTARTSYS;

	while ( !data_avail )
	{
		up(&dev->sem );

		if (filep->f_flags & O_NONBLOCK )
			return -EAGAIN;

		if (wait_event_interruptible(dev->wq, data_avail ))
				return -ERESTARTSYS;

		if (down_interruptible(&dev->sem))
			return -ERESTARTSYS;
	}
	/* ok RDS data is available */
	/* unless, in the mean time, a successful seek purged the fifo. */
	/* then we need to restart the read */
	if ( sizeof(struct fmr_rds_type) != si4703_fifo_get(dev, (u8*) &rdsdata, sizeof(struct fmr_rds_type)))
	{
		up(&dev->sem);
		return -EAGAIN;
	}

	if (copy_to_user(buf, &rdsdata, count)) {
		up(&dev->sem);
		return -EFAULT;
	}

	// data_avail = 0 ; /* data copied so we can wait for new data */
	data_avail = (si4703_fifo_len(gdev) > 0 ) ;
	up(&dev->sem);

	if ( rdsdata.freq == 0 ) PK_ERR ("BUG!! RDSDATA.FREQ == 0! This is not supposed to happen!\n");

	return count;
}

struct file_operations si4703_fops = {
	.owner = THIS_MODULE,
	.open = si4703_open,
	.release = si4703_release,
	.ioctl = si4703_ioctl,
	.poll = si4703_poll,
	.read = si4703_read,
};


static int
si4703_detect_client(struct i2c_adapter *adapter, int address, int kind)
{
	struct si4703dev* dev;
	struct i2c_client *c;
	int rc = 0;

	PK_ENTRY;

	c = kmalloc(sizeof *c, GFP_KERNEL);
	if (!c) {
		rc = -ENOMEM;
		goto no_client;
	}
	memcpy(c, &client_template, sizeof *c);
	c->adapter = adapter;

	strcpy(c->name, /*"Silicon Labs 4703" */ SI4703_DEVNAME);

	dev = kmalloc(sizeof *dev, GFP_KERNEL);

	if (!dev) {
		rc = -ENOMEM;
		goto no_dev;
	}
	memset(dev, 0, sizeof *dev);

	gdev = dev; /* maintain a global pinter to the dev struct */

	i2c_set_clientdata(c, gdev);

	i2c_attach_client(c);

	i2c_use_client(c);

	dev->client = c;

	/* what kind of device do we have here? */
	PK_DBG("Detecting device on adapter: %s, at I2C address: 0x%02X\n",
	       adapter->name, address);

	init_waitqueue_head(&dev->wq);  /* initialize waitqueue for readers */
	init_MUTEX(&dev->sem);		/* initialize the semaphore */
	spin_lock_init(&dev->sp_lock); /* init spinlock */

	rc = si4703_powerup(c);	/* after powerup, the shadow registers are set and synced */

	if ( rc < 0 )
		goto no_cdev;


	/* Now setup our character device to the fm */
	cdev_init(&gdev->cdev, &si4703_fops);
	dev->cdev.owner = THIS_MODULE;
	dev->cdev.ops = &si4703_fops;
	if ((rc = cdev_add(&dev->cdev, tmcRX_devno, 1)))
		goto no_cdev;

	PK_DBG("%s: %p\n", __func__, c);

	if ( si4703_fifo_alloc(dev, 55 * sizeof(struct fmr_rds_type))) goto no_cdev;

	if ( tmcRX_irq > 0 )
	{
		int result = request_irq(tmcRX_irq,
				si4703_interrupthandler,
				SA_INTERRUPT,
				"si4703",
				(void*) dev );
		if (result ) {
			PK_ERR("Cannot register interrupt handler for irq %d.\n", tmcRX_irq);
		}
		else {
			PK_DBG("Interrupt handler registered for irq %d.\n", tmcRX_irq);
		}
	}

	if ( si4703_enable_rds(dev->client))
	{
		PK_ERR("Could not enable RDS!\n");
	}
	/* interrupts on Silabs device are enabled for RDS */

	return rc;

      no_cdev:
	if ( tmcRX_irq > 0 )
	{
		 free_irq( tmcRX_irq, (void*) dev );
	}
	i2c_detach_client(c);
      no_dev:
	si4703_fifo_free(dev);
	kfree(c);
	PK_DBG("! ---- No device! ----\n");
      no_client:
	PK_DBG("! ---- No client! ----\n");
	PK_ERR("%s: ERROR %d!\n", __func__, rc);
	return rc;
}

/* ----------------------------------------------------------------------- */

static int si4703_attach(struct i2c_adapter *adap)
{
	int ret = 0;
	PK_ENTRY;

#warning XH: This should be handled in the dock driver for cagliari.
	if (tmcRX_dock_i2c_en_pin) {
		IOP_Activate(tmcRX_dock_i2c_en_pin);
		udelay(500);
	}

	if (tmcRX_sen_pin) {
		IOP_Deactivate(tmcRX_sen_pin);
		udelay(500);
	}

	IOP_Activate(tmcRX_reset_pin);	/* First attach power then release reset  */
	udelay(500);

	IOP_Deactivate(tmcRX_reset_pin);
	udelay(500);

	ret = i2c_probe(adap, &addr_data, si4703_detect_client);
	PK_DBG("id=%d, name=%s, ret=%d \n", adap->id,
	       adap->algo->name, ret);

	return ret;
}

static int si4703_detach(struct i2c_client *c)
{
	struct si4703dev *fm = i2c_get_clientdata(c);

	PK_ENTRY;
	PK_DBG("%s: %p\n", __func__, c);

	i2c_release_client(c);
	i2c_detach_client(c);

	kfree(fm);
	kfree(c);

	return 0;
}

/* ----------------------------------------------------------------------- */

static int __init si4703_init(void)
{
	int result;
	  if(IO_GetTMCReceiverType() != GOTMC_SI4703)
  {
    PK_DBG("wrong type: %d vs %d\n", IO_GetTMCReceiverType(), GOTMC_SI4703);
    return -1;        
  }


result = driver_register(&driver.driver);
	PK_DBG
	    ("------------------%s:result = 0x%02X---------------\n",
	     __func__, result);

	return result;
}

static void __exit si4703_exit(void)
{
	PK_ENTRY;

	if ( tmcRX_irq > 0 )
	{
		 free_irq( tmcRX_irq, (void*) gdev );
	}

	driver_unregister(&driver.driver);
}

MODULE_AUTHOR("Xander Hover <Xander.Hover@tomtom.com>");
MODULE_DESCRIPTION
    ("Driver for I2C connected Silicon Labs 4703 FM/RDS/TMC Receiver");
module_init(si4703_init);
module_exit(si4703_exit);
