/* si4705.c
*
* Control driver for Silicon Lab 4705 chip.
*
* Copyright (C) 2006 TomTom BV <http://www.tomtom.com/>
* Authors: Xander Hover <Xander.Hover@tomtom.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*/

#include <linux/i2c.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <asm/uaccess.h>
#include <linux/fmreceiver.h>
#include <linux/si4705.h>
#include <linux/device.h>
#include <barcelona/gopins.h>
#include <linux/delay.h>
#include <asm/semaphore.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <asm/irq.h>
#include <asm/arch/irqs.h>
#include <linux/wait.h> /* needed to inmplement blocking read and wait queues */
#include <linux/poll.h> /* needed for poll_wait() call */
#include <linux/kfifo.h> /* needed for the kfifo ringbuffer */

//#define HEXLOG 1

#define PFX "SI4705: "
#define PK_HEX(fmt, arg...)    printk(fmt, ##arg);
#define PK_DBG(fmt, arg...)    printk(KERN_DEBUG PFX "%s: " fmt, __FUNCTION__ ,##arg)
#define PK_ERR(fmt, arg...)    printk(KERN_ERR PFX "%s: " fmt, __FUNCTION__ ,##arg)
#define PK_ENTRY	       PK_DBG("----------%s----------\n", __func__)

#ifndef __bigendian
#    define __bigendian
#endif				/* __bigendian */

#define TUNER_DELAY 2

struct si4705_property_type
{
  u16 property;
  u16 value;
}
__attribute__ ((__packed__));

/* command types are used to xfer over i2c using the i2c_master_send/recv functions */
struct si4705_set_property_cmd_type
{
  u16 __bigendian cmd;
  u16 __bigendian property;
  u16 __bigendian value;
}
__attribute__ ((__packed__));

struct si4705_get_property_cmd_type
{
  u16 __bigendian cmd;
  u16 __bigendian property;
}
__attribute__ ((__packed__));

struct si4705_tune_cmd_type
{
  u16 __bigendian tunecmd;
  u16 __bigendian tunefreq;
  u8 antenna;
}
__attribute__ ((__packed__));

/* SI4705 defined commands */
#define CMD_POWER_UP(a,b)		{ (0x01), (a), (b) }	/* a,b:: u8 */
#define CMD_GET_REV			{ (0x10) }
#define CMD_POWER_DOWN       		{ (0x11) }
#define CMD_SET_PROPERTY(a, b)  	{ cpu_to_be16(0x1200), cpu_to_be16(a), cpu_to_be16(b) }	/* a,b:: u16 */
#define CMD_GET_PROPERTY(a)		{ cpu_to_be16(0x1300), cpu_to_be16(a) }	/* a:: u16 (one of PROP_XXXX) */
#define CMD_GET_INT_STATUS		{ (0x14) }
#define CMD_FM_TUNE_FREQ(a,b)		{ cpu_to_be16(0x2000), cpu_to_be16(a), (b) }	/* a:: u16, b:: u8 */
#define CMD_FM_SEEK_START(a)		{ (0x21), (a) }	/* a:: u8 */
#define CMD_FM_TUNE_STATUS(a)		{ (0x22), (a) }	/* a:: u8 */
#define CMD_FM_RSQ_STATUS(a)		{ (0x23), (a) }	/* a:: u8 */
#define CMD_FM_RDS_STATUS(a)		{ (0x24), (a) }	/* a:: u8, si4705/06/21/31/35/37/39/49 only  */
#define CMD_FM_AGC_STATUS		{ (0x27) }
#define CMD_FM_AGC_OVERRIDE(a,b)	{ (0x28), (a), (b) }	/* a,b:: u8 */
#define CMD_GPIO_CTL(a)			{ (0x80), (a) }	/* a:: u8 */
#define CMD_GPIO_SET(a)			{ (0x81), (a) }	/* a:: u8 */

#define SET_PROPERTY_CMD		(0x1200)
#define GET_PROPERTY_CMD		(0x1300)

/* SI4705 defined properties in cpu endianness */
#define PROP_GPIO_IEN 				(0x0001)
#define PROP_DIGITAL_OUTPUT_FORMAT		(0x0102)
#define PROP_DIGITAL_OUTPUT_SAMPLE_RATE		(0x0104)
#define PROP_REFCLK_FREQ			(0x0201)
#define PROP_REFCLK_PRESCALE			(0x0202)

#define PROP_FM_DEEMPHASIS			(0x1100)
#define PROP_FM_BLEND_STEREO_THRESHOLD		(0x1105)
#define PROP_FM_BLEND_MONO_THRESHOLD		(0x1106)
#define PROP_FM_ANTENNA_INPUT			(0x1107)
#define PROP_FM_MAX_TUNE_ERROR			(0x1108)

#define PROP_FM_RSQ_INT_SOURCE			(0x1200)
#define PROP_FM_RSQ_SNR_HI_THRESHOLD		(0x1201)
#define PROP_FM_RSQ_SNR_LO_THRESHOLD		(0x1202)
#define PROP_FM_RSQ_RSSI_HI_THRESHOLD		(0x1203)
#define PROP_FM_RSQ_RSSI_LO_THRESHOLD		(0x1204)
#define PROP_FM_RSQ_MULTIPATH_HI_THRESHOLD	(0x1205)
#define PROP_FM_RSQ_MULTIPATH_LO_THRESHOLD	(0x1206)
#define PROP_RSQ_BLEND_THRESHOLD		(0x1207)

#define PROP_FM_SOFT_MUTE_RATE			(0x1300)
#define PROP_FM_SOFT_MUTE_MAX_ATTENUATION	(0x1302)
#define PROP_FM_SOFT_MUTE_SNR_THRESHOLD		(0x1303)

#define PROP_SEEK_BAND_BOTTOM			(0x1400)
#define PROP_SEEK_BAND_TOP			(0x1401)
#define PROP_SEEK_FREQ_SPACING			(0x1402)
#define PROP_SEEK_TUNE_SNR_THRESHOLD		(0x1403)

#define PROP_RDS_INT_SOURCE			(0x1500)
#define PROP_RDS_INT_FIFO_COUNT			(0x1501)
#define PROP_RDS_CONFIG				(0x1502)

#define PROP_RX_VOLUME				(0x4000)
#define PROP_RX_HARD_MUTE			(0x4001)

#define PROP_END				(0xFFFF)
#define PROP_LIST_END	                         PROP_END

/* macros below are used to initialize properties for the si4705_{get,set}_properties() functions */
#define S_PROP_GPIO_IEN(d) 			 { (0x0001), (d) }	/* d:: u16 */
#define S_PROP_DIGITAL_OUTPUT_FORMAT(d)		 { (0x0102), (d) }	/* d:: u16, si4705/21/31/35/37/39 only */
#define S_PROP_DIGITAL_OUTPUT_SAMPLE_RATE(d)	 { (0x0104), (d) }	/* d:: u16, si4705/21/31/35/37/39 only */
#define S_PROP_REFCLK_FREQ(d)			 { (0x0201), (d) }	/* d:: u16 */
#define S_PROP_REFCLK_PRESCALE(d)		 { (0x0202), (d) }	/* d:: u16 */

#define S_PROP_FM_DEEMPHASIS(d)			 { (0x1100), (d) }	/* d:: u16, not in si4706/49 */
#define S_PROP_FM_BLEND_STEREO_THRESHOLD(d)	 { (0x1105), (d) }	/* d:: u16 */
#define S_PROP_FM_BLEND_MONO_THRESHOLD(d)	 { (0x1106), (d) }	/* d:: u16 */
#define S_PROP_FM_ANTENNA_INPUT(d)		 { (0x1107), (d) }	/* d:: u16 */
#define S_PROP_FM_MAX_TUNE_ERROR(d)		 { (0x1108), (d) }	/* d:: u16 */

#define S_PROP_FM_RSQ_INT_SOURCE(d)		 { (0x1200), (d) }	/* d:: u16 */
#define S_PROP_FM_RSQ_SNR_HI_THRESHOLD(d)	 { (0x1201), (d) }	/* d:: u16 */
#define S_PROP_FM_RSQ_SNR_LO_THRESHOLD(d)	 { (0x1202), (d) }	/* d:: u16 */
#define S_PROP_FM_RSQ_RSSI_HI_THRESHOLD(d)	 { (0x1203), (d) }	/* d:: u16 */
#define S_PROP_FM_RSQ_RSSI_LO_THRESHOLD(d)	 { (0x1204), (d) }	/* d:: u16 */
#define S_PROP_FM_RSQ_MULTIPATH_HI_THRESHOLD(d)	 { (0x1205), (d) }	/* d:: u16, si4749 only */
#define S_PROP_FM_RSQ_MULTIPATH_LO_THRESHOLD(d)	 { (0x1206), (d) }	/* d:: u16, si4749 only */
#define S_PROP_RSQ_BLEND_THRESHOLD(d)		 { (0x1207), (d) }	/* d:: u16 */

#define S_PROP_FM_SOFT_MUTE_RATE(d)		 { (0x1300), (d) }	/* d:: u16 */
#define S_PROP_FM_SOFT_MUTE_MAX_ATTENUATION(d)	 { (0x1302), (d) }	/* d:: u16, not in si4706/49 */
#define S_PROP_FM_SOFT_MUTE_SNR_THRESHOLD(d)	 { (0x1303), (d) }	/* d:: u16, not in si4706/49 */

#define S_PROP_SEEK_BAND_BOTTOM(d)		 { (0x1400), (d) }	/* d:: u16 */
#define S_PROP_SEEK_BAND_TOP(d)			 { (0x1401), (d) }	/* d:: u16 */
#define S_PROP_SEEK_FREQ_SPACING(d)		 { (0x1402), (d) }	/* d:: u16 */
#define S_PROP_SEEK_TUNE_SNR_THRESHOLD(d)	 { (0x1403), (d) }	/* d:: u16 */
#define S_PROP_SEEK_TUNE_RSSI_THRESHOLD(d)	 { (0x1404), (d) }	/* d:: u16 */

#define S_PROP_RDS_INT_SOURCE(d)		 { (0x1500), (d) }	/* d:: u16 */
#define S_PROP_RDS_INT_FIFO_COUNT(d)		 { (0x1501), (d) }	/* d:: u16, si4705/06/21/31/35/37/39/49 only */
#define S_PROP_RDS_CONFIG(d)			 { (0x1502), (d) }	/* d:: u16, si4705/06/21/31/35/37/39/49 only */

#define S_PROP_RX_VOLUME(d)			 { (0x4000), (d) }	/* d:: u16, not in si4706/49 */
#define S_PROP_RX_HARD_MUTE(d)			 { (0x4001), (d) }	/* d:: u16, not in si4706/49 */

#define S_PROP_LIST_END				 { (0xFFFF), (0xFFFF) }	/* List terminator, not in any device */

/* FM receiver state definitions */
#define FM_OFF 		0U
#define FM_SUSPEND 	1U
#define FM_ON  		2U

/* FM receiver status bitsettings */
#define BSTAT_CTS	(0x80)
#define BSTAT_ERR	(0x40)
#define BSTAT_RSQINT	(0x08)
#define BSTAT_RDSINT	(0x04)
#define BSTAT_STCINT	(0x01)

#define WSTAT_CTS	(0x8000)
#define WSTAT_ERR	(0x4000)
#define WSTAT_RSQINT	(0x0800)
#define WSTAT_RDSINT	(0x0400)
#define WSTAT_STCINT	(0x0100)

/* FM receiver POWER_UP Argument bitsettings */
#define POWERUP_CTSIEN  (0x80)
#define POWERUP_GPO2OEN (0x40)
#define POWERUP_PATCH 	(0x20)
#define POWERUP_XOSCEN	(0x10)
#define POWERUP_FUNC_FMRX (0x00)
#define POWERUP_FUNC_QLID (0x0F)
#define POWERUP_OPMODE_ANA_OUT_SI4749 (0x00)
#define POWERUP_OPMODE_ANA_OUT_SI4705 (0x05)
#define POWERUP_OPMODE_DIG_OUT (0xB0)

/* FM receiver FM_SEEK_START argument bitsettings */
#define FM_SEEK_START_SEEKUP (0x08)
#define FM_SEEK_START_WRAP   (0x04)

/* FM receiver FM_TUNE_STATUS argument bitsettings */
#define FM_TUNE_STATUS_CANCEL (0x02)
#define FM_TUNE_STATUS_INTACK (0x01)

/* FM receiver FM_RSQ_STATUS argument bitsettings */
#define FM_RSQ_STATUS_INTACK (0x01)

/* FM receiver FM_RDS_STATUS argument bitsettings */
#define FM_RDS_STATUS_MTFIFO (0x02)
#define FM_RDS_STATUS_INTACK (0x01)

/* FM receiver FM_AGC_OVERRIDE argument bitsettings */
#define FM_AGC_OVERRIDE_RFAGCDIS (0x01)
#define FM_AGC_OVERRIDE_LNA_GAIN_INDEX (0x00)

/* FM receiver GPIO_CTL argument bitsettings */
#define GPIO_CTL_GPO3OEN (0x08)
#define GPIO_CTL_GPO2OEN (0x04)
#define GPIO_CTL_GPO1OEN (0x02)

/* FM receiver GPIO_SET argument settings */
#define GPIO_SET_GPO3LEVEL (0x08)
#define GPIO_SET_GPO2LEVEL (0x04)
#define GPIO_SET_GPO1LEVEL (0x02)


struct si4705dev
{
  wait_queue_head_t wq; /* wait queue for read */
  struct cdev cdev;
  struct i2c_client *client;
  struct semaphore sem; /* it may be possible that this is not needed */
  spinlock_t sp_lock;
  struct kfifo* fifo;
  u16 lo_limit;
  u16 hi_limit;
  u16 grid;
};

/* there shall only be one user of the device */
static atomic_t si4705_available = ATOMIC_INIT(1);
static gopin_t tmcRX_reset_pin;	/* reset is active low */
static gopin_t tmcRX_en_rds_clk; /* enable clock, active low */
static dev_t tmcRX_devno;
static unsigned int tmcRX_irq; /* the interrupt number of the device */
static unsigned int s_device_id = 0xFFFF;

/* Forward declarations */
//static struct i2c_client client_template;

/* Spec says SI4705 sits on address 0x11 or 0x63 */
static unsigned short normal_i2c[] =
  { SI4705_I2C_SLAVE_ADDR, I2C_CLIENT_END };

I2C_CLIENT_INSMOD;

static int si4705_attach(struct i2c_adapter *adapter);
static int si4705_detach(struct i2c_client *client);
static int si4705_i2c_probe(struct device *dev);

#ifdef CONFIG_PM
static int si4705_i2c_suspend(struct device *dev, pm_message_t state,
                              u32 level);
static int si4705_i2c_resume(struct device *dev, u32 level);
#else
#    define si4710_i2c_suspend  (NULL)
#    define si4710_i2c_resume   (NULL)
#endif				/* CONFIG_PM     */

static int si4705_powerup(struct i2c_client *c);
static int si4705_powerdown(struct i2c_client *c);

static struct i2c_driver driver =
  {
    .id = I2C_DRIVERID_SI4705,
    .owner = THIS_MODULE,
    .name = "si4705_i2c",
    .flags = I2C_DF_NOTIFY,
    .attach_adapter = si4705_attach,
    .detach_client = si4705_detach,
    .driver = {
                .probe = si4705_i2c_probe,
                .suspend = si4705_i2c_suspend,
                .resume = si4705_i2c_resume,
                .name = SI4705_DEVNAME,
                .owner = THIS_MODULE,
                .bus = &platform_bus_type,
              },
  };

static struct i2c_client client_template =
  {
    .name = "si4705_i2c_client",
    .flags = I2C_CLIENT_ALLOW_USE,
    .usage_count = 0,
    .driver = &driver,
    .addr = SI4705_I2C_SLAVE_ADDR,
  };

static int si4705_set_properties(struct i2c_client *client, struct si4705_property_type
                                 *properties);

/* prototypes for tuner commands */
static int si4705_get_freq(struct i2c_client *c, __u16 * freq,
                           __u16 * rssi, __u16 * snr);
// static int si4705_set_freq(struct i2c_client *c, __u16 freq);
static int si4705_set_volume(struct i2c_client *c, __u16 vol);
static int si4705_set_band(struct i2c_client *c, __u16 band);

static int si4705_seek(struct i2c_client *c,
                       struct fmr_seek_param_type seek_param);

#define SI4705_FIFO_SIZE (55 * sizeof(struct fmr_rds_type))
static int si4705_fifo_alloc(struct si4705dev* dev, int size);
static void si4705_fifo_free(struct si4705dev* dev);
static void si4705_fifo_purge(struct si4705dev* dev);
static int si4705_fifo_len(struct si4705dev* dev);
static int si4705_fifo_put(struct si4705dev* dev, unsigned char* data, int len );
static int si4705_fifo_get(struct si4705dev* dev, unsigned char* data, int len );

static u16 current_freq = 0;
static u16 current_rssi = 0;

static struct si4705dev* gdev; /* global pointer to dereference irq later */

static irqreturn_t si4705_interrupthandler(int irq, void* dev_id, struct pt_regs* regs );
static void si4705_workqueue_handler(void* arg );

DECLARE_WORK(rds_4705_data_grabber, si4705_workqueue_handler, (void*) NULL);

static void hexlog(u8 *name, int rx_tx, u8 *ptr, int amount)
{
#ifdef HEXLOG
  int counter;
  PK_HEX("%s %s (%db):", name, rx_tx?"RX":"TX", amount);
  for(counter=0;counter<amount;counter++)
  {
    PK_HEX("%02x ", ptr[counter]);
  }
  PK_HEX("\n");
#endif
}

static int si4705_sendandwait(u8 *tx_data_ptr, int tx_size, u8 *rx_data_ptr, int rx_max_size)
{
  int rc;

  if(!(tx_data_ptr && rx_data_ptr && tx_size && rx_max_size))
  {
    PK_ERR("wrong input params\n");
    return -1;
  }

  rc = i2c_master_send(gdev->client, tx_data_ptr, tx_size);
  hexlog("", 0, tx_data_ptr, tx_size);

  if(rc>0)
  {
    do
    {
      rc = i2c_master_recv(gdev->client, rx_data_ptr, rx_max_size);
      hexlog("", 1, rx_data_ptr, rx_max_size);
    }
    while ( (rc>0) && ((rx_data_ptr[0] & 0x80) != 0x80));
  }

  return rc;
}

static int si4705_tx_getrev(void)
{
  int retval;
  u8 cmd[] = CMD_GET_REV;
  u8 reply[8];

  retval = si4705_sendandwait(
             cmd, sizeof(cmd), reply, sizeof(reply));

  if(retval>0)
  {
    s_device_id = 4700 + reply[1];
  }

  return retval;
}

static int si4705_tx_powerup(int rdsint, int fmrx)
{
  int retval;
  
  u8 mask = ((fmrx?POWERUP_FUNC_FMRX:POWERUP_FUNC_QLID) | (rdsint?POWERUP_GPO2OEN:0));
  
  u8 cmd_si4705[] = CMD_POWER_UP(mask, POWERUP_OPMODE_ANA_OUT_SI4705);
  u8 cmd_si4749[] = CMD_POWER_UP(mask, POWERUP_OPMODE_ANA_OUT_SI4749);
   
  u8 reply[1];

  if(IO_GetTMCReceiverType() == GOTMC_SI4705) 
  {
     retval = si4705_sendandwait(&cmd_si4705[0], sizeof(cmd_si4705), &reply[0], sizeof(reply));
  }
  else
  {
     retval = si4705_sendandwait(&cmd_si4749[0], sizeof(cmd_si4749), &reply[0], sizeof(reply));
  }
  return retval;
}

static int si4705_tx_setfreq(u16 freq)
{
  int retval;
  u16 cmd[] = CMD_FM_TUNE_FREQ(freq, 0U);
  u8 reply[8];

  si4705_fifo_purge(gdev);
  retval = si4705_sendandwait(
             (u8*)cmd, sizeof(cmd), reply, sizeof(reply));

  return retval;
}

static int si4705_start_seek(int up, int wrap)
{
  int retval;
  u8 mask = (up?0x08:0) | (wrap?0x04:0);
  u8 cmd[] = CMD_FM_SEEK_START(mask);
  u8 reply[1];

  si4705_fifo_purge(gdev);
  retval = si4705_sendandwait(
             (u8*)cmd, sizeof(cmd), reply, sizeof(reply));

  return retval;
}

static int si4705_tx_intstatus(u8 *status)
{
  int retval;
  u8 cmd[] = CMD_GET_INT_STATUS;

  memset(status, 0x0, sizeof(*status));
  retval = si4705_sendandwait(cmd, sizeof(cmd), status, sizeof(*status));

  return retval;
}


static int si4705_tx_tune_status(int ack, void *data_ptr, int size)
{
  int retval;
  u8 mask = (u8)(ack?0x01:0);
  u8 cmd[] = CMD_FM_TUNE_STATUS(mask);

  memset(data_ptr, 0x0, size);
  retval = si4705_sendandwait(cmd, sizeof(cmd), data_ptr, size);

  return retval;
}

static int si4705_tx_rds_status(int ack, void *data_ptr, int size)
{
  int retval;
  u8 mask = (u8)(ack?0x01:0);
  u8 cmd[] = CMD_FM_RDS_STATUS(mask);

  memset(data_ptr, 0x0, size);
  retval = si4705_sendandwait(cmd, sizeof(cmd), data_ptr, size);

  return retval;
}

static void si4705_toggle_int(int int_stc, int int_rds)
{
  u16 mask = (int_stc?0x0101:0x0000)|(int_rds?0x0404:0x0000);

  /* set some default properties */
  struct si4705_property_type start_properties[] =
    {
      S_PROP_GPIO_IEN(mask),
      S_PROP_LIST_END,	        /* end of list */
    };

  si4705_set_properties(gdev->client, start_properties);
  mdelay(TUNER_DELAY);
}

static void si4705_workqueue_handler(void* arg )
{
  int push2fifo = 0;
  int purge_fifo = 0;
  u8 status = 0;
  struct fmr_rds_type rdsoutdata;

  PK_ENTRY;
  
  si4705_tx_intstatus(&status);

  /* we're only interested in RDSINT and STCINT */
  if( ((status & 0x05) != 0) && ((status & 0x40) != 0x40))
  {
    /* is it a STC interrupt ? */
    if((status & 0x01) == 0x01)
    {
      /* get tune stat, acknowledge interrupt */
      u8 buf[8];
      u16 *wbuf = (u16 *) buf;

      si4705_tx_tune_status(1, buf, sizeof(buf));

      rdsoutdata.freq = current_freq = be16_to_cpu(wbuf[1]);
      rdsoutdata.sf_bandlimit = ((buf[1] & 0x80) == 0x80);
      rdsoutdata.rssi = current_rssi = buf[4];
      rdsoutdata.blocka = 0x00FF;
      rdsoutdata.blockb = 0x0000;
      rdsoutdata.blockc = 0x0000;
      rdsoutdata.blockd = 0x0000;
      push2fifo = 1;
      purge_fifo = 1;
    }
    else
      /* is it a RDS interrupt ?*/
      if((status & 0x04) == 0x04)
      {
        u8 buf[13];
        u16 *wbuf = (u16 *) buf;

        si4705_tx_rds_status(1, &buf[0], sizeof(buf));

        if(
          /* no recent sync changes */
          ((buf[1] & 0x06) == 0x00) &&
          /* rds synced */
          ((buf[2] & 0x01) == 0x01) &&
          /* there is something in the rds fifo */
          (buf[3] > 0) &&
          /* no rds errors */
          (buf[12] == 0)
        )
        {
          rdsoutdata.freq = current_freq;
          rdsoutdata.sf_bandlimit = 0;
          rdsoutdata.rssi = current_rssi;
          rdsoutdata.blocka =  be16_to_cpu(wbuf[2]);
          rdsoutdata.blockb =  be16_to_cpu(wbuf[3]);
          rdsoutdata.blockc =  be16_to_cpu(wbuf[4]);
          rdsoutdata.blockd =  be16_to_cpu(wbuf[5]);
          push2fifo = 1;
        }
      }
  }
  else
  {
    PK_DBG("work scheduled but irq status is 0x%x\n", status);
  }

  if(push2fifo)
  {
    down(&gdev->sem);

    if(purge_fifo)
    {
      si4705_fifo_purge(gdev);
    }

    if((SI4705_FIFO_SIZE - si4705_fifo_len(gdev)) >= sizeof(struct fmr_rds_type))
    {
      si4705_fifo_put(gdev, (u8*) &rdsoutdata, sizeof(struct fmr_rds_type));
    }

    wake_up(&gdev->wq);
    up(&gdev->sem);
  }
}

static int si4705_get_freq(struct i2c_client *c, __u16 * freq,
                           __u16 * rssi, __u16 * snr)
{
  u8 buf[8];
  u16 *wbuf = (u16 *) buf;

  //  PK_ENTRY;
  si4705_tx_tune_status(0, buf, sizeof(buf));

  *freq = be16_to_cpu(wbuf[1]);
  *rssi = buf[4];
  *snr = buf[5];

  PK_DBG
  ("FM_TUNE_STAT: Freq = %d, RSSI = %d, SNR = %d, MULT = %d, ReadAntennaCap = %d\n",
   *freq, *rssi, *snr, buf[6], buf[7]);
  return 0;
}

// static int si4705_set_freq(struct i2c_client *c, __u16 freq)
// {
//   /* tune to freq */
//   struct si4705_tune_cmd_type tune_cmd = CMD_FM_TUNE_FREQ(freq, 0U);
//   int rc;
//   PK_ENTRY;
//   rc = i2c_master_send(c, (u8 *) & tune_cmd, sizeof(struct si4705_tune_cmd_type));
//   hexlog("si4705_set_freq", 0, (u8*)&tune_cmd, rc);
//
//   return (rc>0)?0:-1;
// }

static int si4705_set_volume(struct i2c_client *c, __u16 vol)
{
  struct si4705_set_property_cmd_type vol_cmd =
          CMD_SET_PROPERTY(PROP_RX_VOLUME, (4 * vol));
  //  PK_ENTRY;
  if (i2c_master_send
      (gdev->client, (u8 *) & vol_cmd,
       sizeof(struct si4705_set_property_cmd_type)))
    return 0;
  else
    return 1;
}

#define band_prop(b,t,s) \
struct si4705_property_type band_prop_array[] = \
{ \
S_PROP_SEEK_BAND_BOTTOM(b), \
S_PROP_SEEK_BAND_TOP(t), \
S_PROP_SEEK_FREQ_SPACING(s), \
S_PROP_LIST_END, \
} \

static int si4705_set_band(struct i2c_client *c, __u16 band)

{
  //  PK_ENTRY;
  {
    u16 b, t, s;
    switch (band)
    {
    default:
    case 0:
      /* Europe, 87.5 - 108.0 MHz, 100 kHz grid */
      b = 8750U;
      t = 10800U;
      s = 10U;
      break;
    case 1:
      /* USA, 87.5 - 108.0 MHz, 200 kHz grid */
      b = 8750U;
      t = 10800U;
      s = 20U;
      break;
    case 2:
      /* Japan, 76.0 - 90.0 MHz, 100 kHz grid */
      b = 7600U;
      t = 9000U;
      s = 10U;
      break;
    }

    {
      band_prop(b, t, s);
      gdev->lo_limit = b;
      gdev->hi_limit = t;
      gdev->grid = s;
      if (si4705_set_properties(c, band_prop_array))
        return -1;
    }
  }
  return 0;
}

static int si4705_seek(struct i2c_client *c,
                       struct fmr_seek_param_type seek_param)
{
  PK_ENTRY;
  if(seek_param.freq < gdev->hi_limit)
  {
    si4705_fifo_purge(gdev);
    if ( current_freq != seek_param.freq ) /* we want to seek from other point of origin */
    {
      u8 buf[8];
      u16 *wbuf = (u16 *) buf;
      PK_DBG("%d != %d --> correcting start freq\n", current_freq, seek_param.freq);
      si4705_toggle_int(0,0);
      si4705_tx_setfreq(seek_param.freq);
      si4705_tx_tune_status(1, buf, sizeof(buf));
      current_freq = be16_to_cpu(wbuf[1]);
      current_rssi = buf[4];
      si4705_toggle_int(1,0);
    }
    si4705_start_seek(seek_param.up, seek_param.wrap);
  }
  else
  {
    struct fmr_rds_type rdsoutdata;
    rdsoutdata.freq = seek_param.freq;
    rdsoutdata.sf_bandlimit = 1;
    rdsoutdata.rssi = 0;
    rdsoutdata.blocka = 0x00FF;
    rdsoutdata.blockb = 0x0000;
    rdsoutdata.blockc = 0x0000;
    rdsoutdata.blockd = 0x0000;
    if((SI4705_FIFO_SIZE - si4705_fifo_len(gdev)) >= sizeof(struct fmr_rds_type))
    {
      si4705_fifo_put(gdev, (u8*) &rdsoutdata, sizeof(struct fmr_rds_type));
    }
  }
  return 0;
}

/* Note: properties are stored in cpu endianness */
static int si4705_set_properties(struct i2c_client *client, struct si4705_property_type
                                 *properties)
{
  int count = 0;
  int rc;

  /* Set all entries in the array. */
  while (properties[count].property != PROP_LIST_END)
  {
    u16 setprop[3] =
      CMD_SET_PROPERTY(properties[count].property,
                       properties[count].value);
    u8 stat = 0x00;
    int retry = 0;
    rc = i2c_master_send(gdev->client, (u8 *) setprop,
                         sizeof(setprop) / sizeof(u8));
    hexlog("CMD_SET_PROPERTY", 0, (u8*)setprop, rc);
    if (rc < 0)
    {
      PK_ERR("Oops! i2c_master_send has failed!\n");
      return rc;
    }
    msleep(10);
    for (retry = 0; retry < 5; retry++)
    {
      rc = i2c_master_recv(client, &stat, sizeof(u8));
      if ((stat & 0x80) == 0x80)
        break;	/* command accepted */
    }
    count++;
    hexlog("CMD_SET_PROPERTY", 1, &stat, rc);
  }
  PK_DBG(" %d properties are set\n", count);
  return 0;
}

static int si4705_i2c_probe(struct device *dev)
{
  struct fm_receiver_info *pdata = dev->platform_data;
  struct platform_device* pdev = to_platform_device(dev);

  PK_ENTRY;

  tmcRX_devno = pdata->device_nr;

  tmcRX_reset_pin = pdata->fm_reset_pin;
  tmcRX_en_rds_clk = pdata->fm_en_rds_clk;
  tmcRX_irq = (unsigned int) platform_get_irq(pdev, 0);

  PK_DBG("SI4705 uses irq %d.\n", tmcRX_irq);

  return i2c_add_driver(&driver);
}

#ifdef CONFIG_PM
static int
si4705_i2c_suspend(struct device *dev, pm_message_t state, u32 level)
{
  PK_DBG("----------------------%s----------------level=%d\n",
         __func__, level);

  si4705_powerdown(gdev->client);
  return 0;
}

static int si4705_i2c_resume(struct device *dev, u32 level)
{
  struct i2c_client *c = to_i2c_client(dev);
  int rc = 0;

  PK_DBG("----------------------%s----------------level=%d\n",
         __func__, level);

  if (level != RESUME_POWER_ON)
    return 0;

  if (c == NULL)
  {
    PK_ERR("Can't find I2C driver!\n");
    rc = -ENODEV;
  }
  else
  {
    si4705_powerup(c);
  }

  return 0;
}

#endif

static int si4705_fifo_alloc(struct si4705dev* dev, int size)
{
  dev->fifo = kfifo_alloc(size, GFP_KERNEL, &dev->sp_lock);
  if (dev->fifo != NULL ) return 0;
  return -1;
}

static void si4705_fifo_free(struct si4705dev* dev)
{
  kfifo_free(dev->fifo);
}

static void si4705_fifo_purge(struct si4705dev* dev)
{
  kfifo_reset(dev->fifo);
}

static int si4705_fifo_len(struct si4705dev* dev)
{
  return kfifo_len(dev->fifo);
}

static int si4705_fifo_put(struct si4705dev* dev, unsigned char* data,  int len )
{
  return kfifo_put(dev->fifo, data, len );
}

static int si4705_fifo_get(struct si4705dev* dev, unsigned char* data,  int len )
{
  return kfifo_get(dev->fifo, data, len);
}

static int si4705_open(struct inode *nodep, struct file *filep)
{
  struct si4705dev *dev =
          container_of(nodep->i_cdev, struct si4705dev, cdev);
  int rc = -ENOTTY;

  PK_ENTRY;

  /* Only one client may open the device */
  if (!atomic_dec_and_test(&si4705_available))
{
    atomic_inc(&si4705_available);
    return -EBUSY;
  }

  filep->private_data = dev;	/* for easy reference */

  rc = si4705_powerup(dev->client);

  return (rc>0)?0:-1;
}

static int si4705_release(struct inode *nodep, struct file *filep)
{
  PK_ENTRY;
  struct si4705dev *dev = (struct si4705dev*) filep->private_data;

  if (filep && filep->private_data)
  {
    filep->private_data = NULL;
  }

  si4705_fifo_purge(dev);

  /* make device available for others to open */
  atomic_inc(&si4705_available);
  return 0;
}

static int
si4705_ioctl(struct inode *nodep, struct file *filep,
             unsigned int cmd, unsigned long arg)
{
  struct si4705dev *dev = filep ? filep->private_data : NULL;

  PK_DBG("----------------------%s----------- cmd=0x%x\n",
         __func__, cmd);
  if (_IOC_TYPE(cmd) != FMRECEIVER_DRIVER_MAGIC)
  {
    PK_ERR("Wrong IOC type! Failing command\n");
    return -ENOTTY;
  }
  switch (cmd)
  {
  case TMCIOC_RDS_START:
      {
        si4705_toggle_int(1,1);
      }
      break;
  case TMCIOC_RDS_STOP:
    {
      si4705_toggle_int(1,0);
    }
    break;
  case TMCIOC_G_FREQUENCY:
    {
      __u16 freq = 0;
      __u16 rssi, snr;
      if (si4705_get_freq
          (dev->client, &freq, &rssi, &snr))
        return -EFAULT;
      if (put_user(freq, (__u16 __user *) arg))
        return -EFAULT;
    }
    break;
  case TMCIOC_S_FREQUENCY:
    {
      __u16 freq = 0;
      if (get_user(freq, (__u16 __user *) arg))
        return -EFAULT;
      //       if (si4705_set_freq(dev->client, freq))
      if (si4705_tx_setfreq(freq) <= 0)
        return -EFAULT;
    }
    break;
  case TMCIOC_S_BAND:
    {
      __u16 band = 0;
      if (get_user(band, (__u16 __user *) arg))
        return -EFAULT;
      if (si4705_set_band(dev->client, band))
        return -EFAULT;
    }
    break;
  case TMCIOC_S_VOLUME:
    {
      __u16 vol = 0;
      if (get_user(vol, (__u16 __user *) arg))
        return -EFAULT;
      if (si4705_set_volume(dev->client, vol))
        return -EFAULT;
    }
    break;
  case TMCIOC_SEEK:
    {
      struct fmr_seek_param_type param;
      if (copy_from_user
          (&param, (void __user *) arg,
           sizeof(struct fmr_seek_param_type)))
        return -EFAULT;
      if (si4705_seek(dev->client, param))
        return -EFAULT;
    }
    break;
  case TMCIOC_QUERY_DEVICE_ID:
    {
      /* s_device_id is static global */
      __u16 id = s_device_id;
      if (put_user(id, (__u16 __user *) arg))
        return -EFAULT;
    }
    break;
  case TMCIOC_Q_RSSI:
    {
      __u16 rssi = 0;
      __u16 freq, snr;
      if (si4705_get_freq
          (dev->client, &freq, &rssi, &snr))
        return -EFAULT;
      if (put_user(rssi, (__u16 __user *) arg))
        return -EFAULT;
    }
    break;
  default:
    printk("unknown ioctl\n");
    return -ENOTTY;
  }
  return 0;
}

static unsigned int
si4705_poll(struct file *filep, struct poll_table_struct *pq)
{
  struct si4705dev *dev = filep->private_data;
  unsigned int mask = 0 ;

  //  PK_ENTRY;

  down(&dev->sem);
  if ((si4705_fifo_len(gdev) > 0 ))
    mask |= POLLIN | POLLRDNORM ; /* readable */
  else
  {
    poll_wait(filep, &dev->wq, pq);
    if ((si4705_fifo_len(gdev) > 0 ))
      mask |= POLLIN | POLLRDNORM ; /* readable */
  }
  up(&dev->sem);

  return mask;
}

/* read blocks when there is not data, as it should */
static ssize_t si4705_read(struct file *filep, char __user* buf, size_t count, loff_t* pos )
{
  struct si4705dev* dev = filep->private_data;

  struct fmr_rds_type rdsdata;

  count = sizeof(struct fmr_rds_type);

  if ( down_interruptible(&dev->sem)) return -ERESTARTSYS;

  while ( !(si4705_fifo_len(gdev) > 0 ) )
  {
    up(&dev->sem );

    if (filep->f_flags & O_NONBLOCK )
      return -EAGAIN;

    if (wait_event_interruptible(dev->wq, (si4705_fifo_len(gdev) > 0 ) ))
      return -ERESTARTSYS;

    if (down_interruptible(&dev->sem))
      return -ERESTARTSYS;
  }
  /* ok RDS data is available */
  /* unless, in the mean time, a successful seek purged the fifo. */
  /* then we need to restart the read */
  if ( sizeof(struct fmr_rds_type) != si4705_fifo_get(dev, (u8*) &rdsdata, sizeof(struct fmr_rds_type)))
  {
    up(&dev->sem);
    return -EAGAIN;
  }

  if (copy_to_user(buf, &rdsdata, count))
  {
    up(&dev->sem);
    return -EFAULT;
  }

  up(&dev->sem);

  if (rdsdata.freq == 0) 
  {
    //PK_ERR ("BUG!! RDSDATA.FREQ == 0! This is not supposed to happen!\n");
    return -ERESTARTSYS;
  }

  return count;
}

struct file_operations si4705_fops =
  {
    .owner = THIS_MODULE,
    .read = si4705_read,
    .open = si4705_open,
    .release = si4705_release,
    .ioctl = si4705_ioctl,
    .poll = si4705_poll,
  };

static int si4705_powerdown(struct i2c_client *c)
{
  /* rds rst */
  IOP_Deactivate(tmcRX_en_rds_clk);
  IOP_Activate(tmcRX_reset_pin);
  msleep(100);
  return 0;
}

static int si4705_powerup(struct i2c_client *c)
{
  int rc = -ENOTTY;

  PK_ENTRY;

  /* set iopins */
  /* rds rst */
  IOP_Activate(tmcRX_reset_pin);
  msleep(100);
  IOP_Deactivate(tmcRX_reset_pin);
  msleep(100);
  IOP_Activate(tmcRX_en_rds_clk);

  rc = si4705_tx_powerup(1, 1);

  if(rc>0)
    rc = si4705_tx_getrev();

  if(rc>0)
  {
    /* set some default properties */
    struct si4705_property_type start_properties[] =
      {
        S_PROP_GPIO_IEN(0x0505),        /* RDS IRQ and STC IRQ, repeat both, even if already set */
        S_PROP_REFCLK_FREQ(0x8000),
        S_PROP_REFCLK_PRESCALE(1),
        S_PROP_FM_ANTENNA_INPUT(0),
        S_PROP_FM_MAX_TUNE_ERROR(0x0028),
        S_PROP_SEEK_TUNE_SNR_THRESHOLD(0x000A),
        S_PROP_SEEK_TUNE_RSSI_THRESHOLD(0x000E),
        S_PROP_LIST_END,	        /* end of list */
      };

    si4705_set_properties(gdev->client, start_properties);
    mdelay(TUNER_DELAY);
  }

  //   if(rc>0)
  //     rc = si4705_tx_setfreq(8570);

  if(rc>0)
  {
    /* set some default properties */
    struct si4705_property_type rds_properties[] =
      {
        S_PROP_RDS_INT_SOURCE(0x0001),      // RDSRECV
        S_PROP_RDS_INT_FIFO_COUNT(0x0001),  // 1 element fifo
        S_PROP_RDS_CONFIG(0x0001),          // enable rds, no errors allowed
        S_PROP_LIST_END,	        /* end of list */
      };

    si4705_set_properties(gdev->client, rds_properties);
    mdelay(TUNER_DELAY);
  }

  PK_DBG("si4705_powerup returns %d\n", rc);

  return rc;
}

static int
si4705_detect_client(struct i2c_adapter *adapter, int address, int kind)
{
  struct si4705dev *dev;
  struct i2c_client *c;
  int rc = 0;
  PK_ENTRY;

  c = kmalloc(sizeof *c, GFP_KERNEL);
  if (!c)
  {
    rc = -ENOMEM;
    goto no_client;
  }

  memcpy(c, &client_template, sizeof *c);
  c->adapter = adapter;
  c->addr = address;
  strcpy(c->name, /*"Silicon Labs 4705" */ SI4705_DEVNAME);

  dev = kmalloc(sizeof *dev, GFP_KERNEL);
  if (!dev)
  {
    rc = -ENOMEM;
    goto no_dev;
  }
  memset(dev, 0, sizeof *dev);

  gdev = dev; /* maintain a global pinter to the dev struct */

  i2c_set_clientdata(c, dev);
  i2c_attach_client(c);
  i2c_use_client(c);

  dev->client = c;

  /* is this a 4705 kind of device? */
  PK_DBG("Detecting device on adapter: %s, at i2c address: 0x%02X\n",
         adapter->name, address);

  init_waitqueue_head(&dev->wq);  /* initialize waitqueue for readers */
  init_MUTEX(&dev->sem);		/* initialize the semaphore */
  spin_lock_init(&dev->sp_lock); /* init spinlock */

  rc = si4705_powerup(c);

  if (rc < 0)
    goto no_cdev;

  /* Now setup our character device to the fm */
  cdev_init(&dev->cdev, &si4705_fops);
  dev->cdev.owner = THIS_MODULE;
  dev->cdev.ops = &si4705_fops;
  if ((rc = cdev_add(&dev->cdev, tmcRX_devno, 1)))
    goto no_cdev;

  PK_DBG("%s: %p\n", __func__, c);

  if ( si4705_fifo_alloc(dev, SI4705_FIFO_SIZE)) goto no_cdev;

  if ( tmcRX_irq > 0 )
  {
    int result = request_irq(tmcRX_irq,
                             si4705_interrupthandler,
                             SA_INTERRUPT,
                             "si4705",
                             (void*) dev );
    if (result )
    {
      PK_DBG("Cannot register interrupt handler for irq %d.\n", tmcRX_irq);
    }
    else
    {
      PK_DBG("Interrupt handler registered for irq %d.\n", tmcRX_irq);
    }
  }
  else
  {
    PK_DBG("no interrupt handler to register\n");
  }

  return 0;

no_cdev:
  if ( tmcRX_irq > 0 )
  {
    free_irq( tmcRX_irq, (void*) dev );
  }
  i2c_detach_client(c);
no_dev:
  si4705_fifo_free(dev);
  kfree(c);
  PK_DBG("! ---- No device! ----\n");
no_client:
  PK_DBG("! ---- No client! ----\n");
  PK_ERR("%s: ERROR %d!\n", __func__, rc);
  return rc;
}

/* ----------------------------------------------------------------------- */

static int si4705_attach(struct i2c_adapter *adap)
{
  int ret = 0;
  PK_ENTRY;

  /* needed for si4705 busmode selection */
  IOP_Activate(tmcRX_reset_pin);
  udelay(500);

  IOP_Deactivate(tmcRX_reset_pin);
  udelay(500);

  ret = i2c_probe(adap, &addr_data, si4705_detect_client);
  PK_DBG("id=%d, name=%s, ret=%d \n", adap->id,
         adap->algo->name, ret);

  return ret;
}

static int si4705_detach(struct i2c_client *c)
{
  struct si4705dev *fm = i2c_get_clientdata(c);

  PK_ENTRY;
  PK_DBG("%s: %p\n", __func__, c);

  i2c_release_client(c);
  i2c_detach_client(c);

  kfree(fm);
  kfree(c);

  return 0;
}

static int __init si4705_init(void)
{
  int result;
  if((IO_GetTMCReceiverType() != GOTMC_SI4705) &&
     (IO_GetTMCReceiverType() != GOTMC_SI4749))
  {
    PK_DBG("wrong type: %d vs %d/%d\n", IO_GetTMCReceiverType(), GOTMC_SI4705, GOTMC_SI4749);
    return -1;        
  }

  result = driver_register(&driver.driver);
  PK_DBG
  ("------------------%s:result = 0x%02X---------------\n",
   __func__, result);

  return result;
}

static void __exit si4705_exit(void)
{
  PK_ENTRY;

  if ( tmcRX_irq > 0 )
  {
    free_irq( tmcRX_irq, (void*) gdev );
  }

  driver_unregister(&driver.driver);
}

MODULE_AUTHOR("Xander Hover <Xander.Hover@tomtom.com>");
MODULE_DESCRIPTION
("Driver for I2C connected Silicon Labs 4705/06/49 FM/RDS/TMC Receiver");
module_init(si4705_init);
module_exit(si4705_exit);

static irqreturn_t si4705_interrupthandler(int irq, void* dev_id, struct pt_regs* regs )
                {
                  schedule_work(&rds_4705_data_grabber);
                  return IRQ_HANDLED;
                }
