/* 
 *  adv7180 - adv7180 video decoder driver
 *
 * Copyright (c) 2007 TomTom B.V
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. NO WARRANTY OF ANY KIND is expressed or implied.
 
 * Author: Onno Hovers, <onno.hovers@tomtom.com>
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <barcelona/gopins.h>
#include <barcelona/gotype.h>

MODULE_DESCRIPTION("Analog Devices ADV7180 video decoder driver");
MODULE_AUTHOR("Onno Hovers");
MODULE_LICENSE("GPL");

#include <linux/i2c.h>
#include <linux/i2c-dev.h>

#define I2C_NAME(s) (s)->name

#include <linux/videodev2.h>

#include "camera.h"

static int debug = 0;
module_param(debug, int, 0);
MODULE_PARM_DESC(debug, "Debug level (0-1)");

#define dprintk(num, format, args...) \
	do { \
		if (debug >= num) \
			printk(format, ##args); \
	} while (0)

#define ADV7180_MAX_INPUT	9	/* 6 Composite, 3 S-Video, 2 YPrPb */
#define ADV7180_MAX_OUTPUT	0	/* its a decoder only */

#define	I2C_ADV7180		0x40	/* or 0x42 */

#define ADV7180_NR_REG		0xFF

static inline int adv7180_read (struct i2c_client *client, u8 reg)
{
	return i2c_smbus_read_byte_data(client, reg);
}

static inline int adv7180_write (struct i2c_client *client, u8 reg, u8 value)
{
	return i2c_smbus_write_byte_data(client, reg, value);
}

struct adv7180 {
	int mode;
	int input;
	int enable;
	int bright;
	int contrast;
	int hue;
	int saturation;

	wait_queue_head_t wq;
};


/* VIDEO MODES supported by ADV7180 */
struct adv7180_vmode
{
	v4l2_std_id  id;
	char        *name;
	u8           input_control;
};

#define VMODES 9

static const struct adv7180_vmode vmodes[VMODES] =
{
	{ V4L2_STD_ALL,       "auto",          0x00 },
	{ V4L2_STD_PAL,       "PAL B/G/H/I/D", 0x80 },
	{ V4L2_STD_PAL_N,     "PAL N",         0x90 },
	{ V4L2_STD_PAL_M,     "PAL M",         0xA0 },
	{ V4L2_STD_PAL_Nc,    "PAL Nc",        0xC0 },
	{ V4L2_STD_PAL_60,    "PAL 60",        0x60 },
	{ V4L2_STD_NTSC_M_JP, "NTSC J",        0x40 },
	{ V4L2_STD_NTSC_M,    "NTSC M",        0x50 },
	{ V4L2_STD_SECAM,     "SECAM",         0xE0 }
};

/* INPUTS supported by ADV7180 */
struct adv7180_vinput
{
	char        *name;
	u8           input_control;
};

#define VINPUTS 5

static const struct adv7180_vinput vinputs[VINPUTS] =
{
	{ "Composite 1", 0x0 },
	{ "Composite 2", 0x1 },
	{ "Composite 3", 0x2 },
	{ "S-Video",     0x6 },
	{ "YPrPb",       0x9 }, 
};

static int adv7180_setup (struct i2c_client *client)
{
	struct adv7180 *decoder = i2c_get_clientdata(client);
	uint8_t         ic;
	
	ic = vinputs[decoder->input].input_control | vmodes[decoder->mode].input_control;
		
	dprintk(1, KERN_INFO "%s: setup ic=0x%02x\n", I2C_NAME(client), ic);
	
	adv7180_write(client, 0x00, ic); /* select input and video mode */
	if(decoder->input < 9)
		adv7180_write(client, 0x04, 0x57); /* enable SFL */
	if(decoder->input < 6)
		adv7180_write(client, 0x17, 0x41); /* enable SH1 */
	adv7180_write(client, 0x3D, 0xA2); /* MWE enable manual window, color kill threshold to 2 */
	adv7180_write(client, 0x3E, 0x6A); /* BLM optimalization */
	adv7180_write(client, 0x3F, 0xA0); /* BGB optimalization */	
	adv7180_write(client, 0x35, 0x00); /* Adjust HSYNC begin */
	adv7180_write(client, 0x36, 0x7D); /* Adjust HSYNC end */
	adv7180_write(client, 0xE6, 0x10); /* Adjust VSYNC end for NTSC */
	adv7180_write(client, 0xE9, 0x15); /* Adjust VSYNC end for PAL */
	if((decoder->input >= 6) && (decoder->input < 10))
		adv7180_write(client, 0x58, 0x04); /* Mandatory for S-Video */
	adv7180_write(client, 0x0E, 0x80); /* Hidden space */
	adv7180_write(client, 0x55, 0x81); /* ADC configuration */
	adv7180_write(client, 0x01, 0x04); /* Square Pixel support ??? */
	adv7180_write(client, 0x0E, 0x00); /* User space */

	/* 
	 * The previous undocumented commands do not always work in hidden space. 
	 * In that case make sure that Video Selection has the default value 
	 */
	adv7180_write(client, 0x01, 0xC8); 
	
	adv7180_write(client, 0x08, decoder->contrast); /* Contrast: 0 to 255 */
	adv7180_write(client, 0x0A, decoder->bright);   /* Brightness: -128 to 127 */
	adv7180_write(client, 0xE3, decoder->saturation); /* gain on the Cb channel: 0 to 255 */
	adv7180_write(client, 0xE4, decoder->saturation); /* gain on the Cr channel: 0 to 255 */
	adv7180_write(client, 0x0B, decoder->hue - 128); /* Hue: 0 to 255*/
	
	return 0;
}


static int adv7180_get_status (struct i2c_client *client)
{
	int res     = 0;
	int power   = adv7180_read(client, 0x0f);
	int status1 = adv7180_read(client, 0x10);
	int status3 = adv7180_read(client, 0x13);

	dprintk(1, KERN_INFO "%s: status3=0x%02x\n", I2C_NAME(client),  status3);
	
	if (power & 0x20) /* PWRDWN */
		res |= V4L2_IN_ST_NO_POWER;
	
	if (status3 & 0x10) /* not FREE_RUN_ACT */
		res |= V4L2_IN_ST_NO_SIGNAL;
		
	if (!(status3 & 0x01)) /* not INST_HLOCK  */
		res |= V4L2_IN_ST_NO_H_LOCK;
	
	if (status1 & 0x80) /* COL_KILL  */
		res |= V4L2_IN_ST_COLOR_KILL;

	return res;
}

static int adv7180_wait_for_signal(struct i2c_client *client)
{
	DEFINE_WAIT(wait);
	struct adv7180 *decoder = i2c_get_clientdata(client);
	int i;
	
	/* Dublin monitors the video stream so there is no need to wait */
	if (IO_GetModelId() != GOTYPE_DUBLIN) {
		/* poll status for 1 second or until we have a signal */
		for (i=0; i< 10; i++) {
			if((adv7180_get_status(client) & 
			   (V4L2_IN_ST_NO_POWER | V4L2_IN_ST_NO_SIGNAL | V4L2_IN_ST_NO_H_LOCK))==0)
				return 0;
			
			prepare_to_wait(&decoder->wq, &wait, TASK_UNINTERRUPTIBLE);
			schedule_timeout((HZ+9)/10);
			finish_wait(&decoder->wq, &wait);
		}
	}
	
	if((adv7180_get_status(client) & 
		(V4L2_IN_ST_NO_POWER | V4L2_IN_ST_NO_SIGNAL | V4L2_IN_ST_NO_H_LOCK))==0)
		return 0;
	
	return -EIO;
}

static int adv7180_enum_std(struct i2c_client *client, struct v4l2_standard *std)
{
	int err;
	unsigned int i;

	i = std->index;
	if (i >= VMODES)
		return -EINVAL;
	err = v4l2_video_std_construct(std, vmodes[i].id, vmodes[i].name);
	std->index = i;
	
	return err;
}

static int adv7180_query_std(struct i2c_client *client, v4l2_std_id *std)
{
	struct adv7180 *decoder = i2c_get_clientdata(client);
	
	if(decoder->mode != 0) {	
		*std = vmodes[decoder->mode].id;
	} else {
		int status1;
		
		// we do not wait for a signal here, so we can be polled
		if ((adv7180_get_status(client) &  (V4L2_IN_ST_NO_POWER | V4L2_IN_ST_NO_SIGNAL | V4L2_IN_ST_NO_H_LOCK))==0)
		{
			status1 = adv7180_read(client, 0x10);
					
			switch (status1 & 0x70) {
			case 0x00: /* NTSC M/J */
			case 0x10: /* NTSC 4.43 */				
				*std = V4L2_STD_NTSC;
				break;
			case 0x20: /* PAL M */
				*std = V4L2_STD_PAL_M;
				break;
			case 0x30: /* PAL 60 */
				*std = V4L2_STD_PAL_60;
				break;
			case 0x40: /* PAL B/G/H/I/D */
				*std = V4L2_STD_PAL;
				break;
			case 0x60: /* PAL Combination N */
				*std = V4L2_STD_PAL_Nc;
				break;
			case 0x50: /* SECAM */
			case 0x70: /* SECAM 525 */
				*std = V4L2_STD_SECAM;
				break;
			}
		}
		else
		{
			// no signal
			*std = 0;
		}
	}
	return 0;
}

static int adv7180_get_std(struct i2c_client *client, v4l2_std_id *std)
{
	struct adv7180 *decoder = i2c_get_clientdata(client);
	
	*std = vmodes[decoder->mode].id;
	return 0;
}

static int adv7180_set_std(struct i2c_client *client, v4l2_std_id *std)	
{
	struct adv7180 *decoder = i2c_get_clientdata(client);
	int mode=0;
	int i;
	
	for(i = 1; i < VMODES; i++) {
		if((vmodes[i].id & *std) != 0)
			mode = i;
	}
	
	if((mode == 0) && ((*std & V4L2_STD_ALL) == 0))
		return -EINVAL;
	
	decoder->mode = mode;
	adv7180_setup(client);
	adv7180_wait_for_signal(client);
	return 0;
}

static int adv7180_enum_input(struct i2c_client *client, struct v4l2_input *in)
{
	struct adv7180 *decoder = i2c_get_clientdata(client);
	unsigned int i = in->index;
	
	if (i >= VINPUTS)
		return -EINVAL;
	
	memset(in, 0, sizeof(struct v4l2_input));
	in->index = i;
	strcpy(in->name, vinputs[i].name);
	in->type   = V4L2_INPUT_TYPE_CAMERA;
	in->std    = V4L2_STD_ALL;
	
	if(i == decoder->input) 
		in->status = adv7180_get_status(client);
	
	return 0;
}

static int adv7180_get_input(struct i2c_client *client, int *input)
{
	struct adv7180 *decoder = i2c_get_clientdata(client);
	
	*input = decoder->input;
	return 0;
}

static int adv7180_set_input(struct i2c_client *client, int *input)
{
	struct adv7180 *decoder = i2c_get_clientdata(client);
	
	if((*input<0) || (*input > VINPUTS))
		return -EINVAL;
	
	decoder->input = *input;
	
	adv7180_setup(client);
	return 0;
}	

static int adv7180_query_ctrl(struct i2c_client *client, struct v4l2_queryctrl *qctl)
{
	switch (qctl->id) {
	case V4L2_CID_BRIGHTNESS:
		strcpy(qctl->name, "Brightness");
		qctl->minimum =  -128;
		qctl->maximum =   127;
		qctl->default_value =  0;
		break;
	case V4L2_CID_CONTRAST:
		strcpy(qctl->name, "Contrast");
		qctl->minimum = 0;
		qctl->maximum = 255;
		qctl->default_value =  128;
		break;
	case V4L2_CID_SATURATION:
		strcpy(qctl->name, "Saturation");
		qctl->minimum = 0;
		qctl->maximum = 255;
		qctl->default_value =  128;
		break;
	case V4L2_CID_HUE:
		strcpy(qctl->name, "Hue");
		qctl->minimum =  0;
		qctl->maximum =  255;
		qctl->default_value = 128;
		break;
	default:
		return -EINVAL;
	}
	
	qctl->step = 1;
	qctl->type = V4L2_CTRL_TYPE_INTEGER;
	qctl->flags = 0;
	
	return 0;
}


static int adv7180_get_ctrl(struct i2c_client *client, struct v4l2_control *ctl)
{
	switch (ctl->id) {
	case V4L2_CID_CONTRAST:
		ctl->value = adv7180_read(client, 0x08);
		break;
	case V4L2_CID_BRIGHTNESS:
		ctl->value = (signed char) adv7180_read(client, 0x0A);
		break;
	case V4L2_CID_SATURATION:
		ctl->value = (adv7180_read(client, 0xE3) + adv7180_read(client, 0xE4))/2;
		break;
	case V4L2_CID_HUE:
		ctl->value = ((signed char) adv7180_read(client, 0x0B)) + 128;
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static int adv7180_set_ctrl(struct i2c_client *client, struct v4l2_control *ctl)
{
	struct adv7180 *decoder = i2c_get_clientdata(client);
	
	switch (ctl->id) {
	case V4L2_CID_CONTRAST:
		decoder->contrast = ctl->value;
		adv7180_write(client, 0x08, ctl->value);
		break;
	case V4L2_CID_BRIGHTNESS:
		decoder->bright = ctl->value;
		adv7180_write(client, 0x0A, ctl->value);
		break;
	case V4L2_CID_SATURATION:
		decoder->saturation = ctl->value;
		adv7180_write(client, 0xE3, ctl->value); /* gain on the Cb channel */
		adv7180_write(client, 0xE4, ctl->value); /* gain on the Cr channel */
		break;
	case V4L2_CID_HUE:
		decoder->hue = ctl->value;
		adv7180_write(client, 0x0B, ctl->value - 128); /* Hue: 0 to 255*/
		break;
	default:
		return -EINVAL;
	}
	return 0;
}


#ifdef USE_WITH_VDTEST
/*
If you want to use vdtest enable the USE_WITH_VDTEST define in the file:
- linux-s3c24xx/include/media/s3c-camif.h (included via camera.h)
*/
#warning "For release builds undefine USE_WITH_VDTEST in linux-s3c24xx/include/media/s3c-camif.h\n"

static int adv7180_rr(struct i2c_client *client, int  *res)
{
	int _res = *res;
	int idx=_res &255;
	int val=0;
	char c='c';
	
	val = adv7180_read(client, idx);
	if (_res & 0x8000)
	{
		val <<= 8;
		val |= adv7180_read(client, idx+1);
		c='w';
	}
//	printk(KERN_ERR "_res=%x.  [%02X].%c->%02X.\n",_res,idx,c,val);
	*res = val;
	return 0;
}

static int adv7180_rw(struct i2c_client *client, int *res)
{
	int arg=*res;
	int idx=arg &255;
	int val=arg>>16;
	int advres=0;
	
//	printk(KERN_ERR "%02X->[%02X].\n",idx,val&255);
	adv7180_write(client, idx, val &255);
	if (arg & 0x8000)
	{
		printk(KERN_ERR "%02X->[%02X].\n",idx+1,(val>>8)&255);
		adv7180_write(client, idx+1, (val>>8) &255);
	}
	
	return 0;
}
#endif


static int adv7180_get_defres(struct i2c_client *client, struct v4l2_rect  *res)
{
	/* default is 0 (NTSC) */
	int status1 = 0;
	
	adv7180_wait_for_signal(client);
	
	/* we do not wait for a signal here, so we can be polled */
	if ((adv7180_get_status(client) &  (V4L2_IN_ST_NO_POWER | V4L2_IN_ST_NO_SIGNAL | V4L2_IN_ST_NO_H_LOCK))==0)
	{
		status1 = adv7180_read(client, 0x10);
	}
	
	
	switch (status1 & 0x70) {
	case 0x00: /* NTSC M/J */
	case 0x10: /* NTSC 4.43 */
	case 0x70: /* SECAM 525 */	
		/* 720 x (480/2) */
		res->left = 0;
		res->top  = 0;
		res->width = 720;
		res->height = 240;
		break;
	default:
		/* 720 x (576/2) */
		res->left = 0;
		res->top  = 0;
		res->width = 720;
		res->height = 288;
		break;
	}
	return 0;
}

static int adv7180_set_defres(struct i2c_client *client, struct v4l2_rect  *arg)
{
	return -EINVAL;
}

static int adv7180_suspend(struct i2c_client *client, void  *arg)
{
	adv7180_write(client, 0x0F, 0x24); // According to adv1780 datasheet, this has the same effect as nPWRDWN-line
	IO_Activate(CAM_DPWDN);

	return 0;
}

static int adv7180_resume(struct i2c_client *client, void  *arg)
{
	IO_Deactivate(CAM_DPWDN);
	adv7180_write(client, 0x0F, 0x80);
	mdelay(25); /* data sheet says power-up time is 20ms, let's be on the safe side*/
	adv7180_setup(client);
	
	return 0;
}

static int adv7180_dumpregs(struct i2c_client *client, void  *arg)
{
	printk(KERN_DEBUG "ADV7180 ICTL =0x%02X VSEL =0x%02X OCTL =0x%02X EOCTL=0x%02X\n",
		adv7180_read(client, 0x00), adv7180_read(client, 0x01), adv7180_read(client, 0x03), adv7180_read(client, 0x04));
	printk(KERN_DEBUG "ADV7180 ADEN =0x%02X CONT =0x%02X BRIG =0x%02X HUE  =0x%02X\n",
		adv7180_read(client, 0x07), adv7180_read(client, 0x08), adv7180_read(client, 0x0A), adv7180_read(client, 0x0B));
	printk(KERN_DEBUG "ADV7180 DEFY =0x%02X DEFC =0x%02X ADIC =0x%02X POWER=0x%02X\n",
		adv7180_read(client, 0x0C), adv7180_read(client, 0x0D), adv7180_read(client, 0x0E), adv7180_read(client, 0x0F));
	printk(KERN_DEBUG "ADV7180 STAT1=0x%02X IDENT=0x%02X STAT2=0x%02X STAT3=0x%02X\n",
		adv7180_read(client, 0x10), adv7180_read(client, 0x11), adv7180_read(client, 0x12), adv7180_read(client, 0x13));
	return 0;
}

static int adv7180_get_wss_data(struct i2c_client *client, int *arg)
{
	int status1, vdp_status;
	
	status1 = adv7180_read(client, 0x10); 
			
	adv7180_write(client, 0x0E, 0x20);		/* access VDP register space */
	vdp_status = adv7180_read(client, 0x78);	/* get VDP_STATUS */
	if (vdp_status & 0x8) {				/* CGMS/WSS detected */
		if (( status1 & 70) < 0x20) {		/* NTSC */
			*arg = adv7180_read(client, 0x7F) | (adv7180_read(client, 0x7E)<<8) | ((adv7180_read(client, 0x7D) & 0x0f)<<16);
		} else	{				/* PAL */
			*arg = adv7180_read(client, 0x7F) | ((adv7180_read(client, 0x7E) & 0x1f)<<8);
		}
		adv7180_write(client, 0x78, 0x55);
		adv7180_write(client, 0x0E, 0x00);	/* access normal register space */
		return 0;
	} else {
		adv7180_write(client, 0x78, 0x55);
		adv7180_write(client, 0x0E, 0x00);	/* access normal register space */
		return -EIO;
	}
}

static int adv7180_command (struct i2c_client *client,
                            unsigned int       cmd,
                            void              *arg)
{
	switch (cmd) {
	/* video modes */
	case VIDIOC_ENUMSTD:
		return adv7180_enum_std(client, arg);
	case VIDIOC_QUERYSTD:
		return adv7180_query_std(client, arg);
	case VIDIOC_G_STD:
		return adv7180_get_std(client, arg);
	case VIDIOC_S_STD:
		return adv7180_set_std(client, arg);
	/* inputs */
	case VIDIOC_ENUMINPUT:
		return adv7180_enum_input(client, arg);
	case VIDIOC_G_INPUT:
		return adv7180_get_input(client, arg);
	case VIDIOC_S_INPUT:
		return adv7180_set_input(client, arg);
	/* settings */
	case VIDIOC_QUERYCTRL:
		return adv7180_query_ctrl(client, arg);
	case VIDIOC_G_CTRL:
		return adv7180_get_ctrl(client, arg);
	case VIDIOC_S_CTRL:
		return adv7180_set_ctrl(client, arg);
	/* default resolution */
	case VIDIOC_PRIV_GETDEFRES:
		return adv7180_get_defres(client, arg);
	case VIDIOC_PRIV_SETDEFRES:
		return adv7180_set_defres(client, arg);
	/* suspend/resume */
	case VIDIOC_PRIV_SLEEP:
	case VIDIOC_PRIV_SUSPEND:
		return adv7180_suspend(client, arg);
	case VIDIOC_PRIV_WAKE:
	case VIDIOC_PRIV_RESUME:
		return adv7180_resume(client, arg);
	/* debugging */
	case VIDIOC_PRIV_DUMPREGS:
		return adv7180_dumpregs(client, arg);
	/* screen ratio */
	case VIDIOC_PRIV_GET_WSS_DATA:
		return adv7180_get_wss_data(client, arg);

#ifdef USE_WITH_VDTEST
/*
If you want to use vdtest enable the USE_WITH_VDTEST define in the file:
- linux-s3c24xx/include/media/s3c-camif.h (included via camera.h)
*/
#warning "For release builds undefine USE_WITH_VDTEST in linux-s3c24xx/include/media/s3c-camif.h\n"

	case VIDIOC_PRIV_CRR:
		return adv7180_rr(client, arg);
	case VIDIOC_PRIV_CRW:
		return adv7180_rw(client, arg);
#endif
	
	default:
		dprintk(1, KERN_INFO "unknown adv7180_command??(%d)\n",
			cmd);
		return -EINVAL;
	}
	return 0;
}

static unsigned short normal_i2c[] = {
	I2C_ADV7180 >> 1,
	(I2C_ADV7180 >> 1) + 1,
	I2C_CLIENT_END
};

static unsigned short ignore = I2C_CLIENT_END;

static struct i2c_client_address_data addr_data = {
	.normal_i2c		= normal_i2c,
	.probe			= &ignore,
	.ignore			= &ignore,
	.force			= &ignore,
};

static struct i2c_driver i2c_driver_adv7180;

static int adv7180_detect_client (struct i2c_adapter *adapter,
		       int                 address,
		       int                 kind)
{
	struct i2c_client *client;
	struct adv7180 *decoder;
	int rv;
  int ident;

	/* power up device */
	IO_Deactivate(CAM_DPWDN);
	/* data sheet says power-up time is 20ms, let's be on the safe side*/
	mdelay(25);

	dprintk(1,
		KERN_INFO
		"adv7180.c: detecting adv7180 client on address 0x%x\n",
		address << 1);

	/* Check if the adapter supports the needed features */
	if (!i2c_check_functionality
	    (adapter,
	     I2C_FUNC_SMBUS_READ_BYTE | I2C_FUNC_SMBUS_WRITE_BYTE_DATA))
		return 0;

	client = kmalloc(sizeof(struct i2c_client), GFP_KERNEL);
	if (client == 0)
		return -ENOMEM;
	memset(client, 0, sizeof(struct i2c_client));
	client->addr = address;
	client->adapter = adapter;
	client->driver = &i2c_driver_adv7180;
	client->flags = I2C_CLIENT_ALLOW_USE;
	strlcpy(I2C_NAME(client), "adv7180", sizeof(I2C_NAME(client)));

	decoder = kmalloc(sizeof(struct adv7180), GFP_KERNEL);
	if (decoder == 0) {
		kfree(client);
		return -ENOMEM;
	}
	memset(decoder, 0, sizeof(struct adv7180));
	decoder->mode   = 0;
	decoder->input  = 0;
	decoder->enable = 1;
	decoder->bright = 0;
	decoder->contrast = 128;
	decoder->saturation = 128;	
	decoder->hue = 128;

	init_waitqueue_head(&decoder->wq);
	i2c_set_clientdata(client, decoder);

	rv = i2c_attach_client(client);
	if (rv) {
		kfree(client);
		kfree(decoder);
		return rv;
	}
		
	/* Driver has been implemented for chips with identity 0x1B or 0x1C. 
	Newer identities will be allowed, but a warning is printed on the terminal */
	ident = adv7180_read(client, 0x11);
	if (ident >= 0x1B)  {
		adv7180_setup(client);
		dprintk(1, KERN_INFO "%s_attach: ADV7180\n",I2C_NAME(client));
	}
	if ((ident != 0x1B) && (ident != 0x1C)) {
		printk(KERN_INFO "%s_attach: ADV7180 with unsupported ident number: 0x%x\n",I2C_NAME(client), ident);
	}
		
	return 0;
}

static int
adv7180_attach_adapter (struct i2c_adapter *adapter)
{
	dprintk(1,
		KERN_INFO
		"adv7180.c: starting probe for adapter %s (0x%x)\n",
		I2C_NAME(adapter), adapter->id);
	return i2c_probe(adapter, &addr_data, &adv7180_detect_client);
}

static int
adv7180_detach_client (struct i2c_client *client)
{
	struct adv7180 *decoder = i2c_get_clientdata(client);
	int err;

	err = i2c_detach_client(client);
	if (err) {
		return err;
	}

	kfree(decoder);
	kfree(client);

	return 0;
}

static struct i2c_driver i2c_driver_adv7180 = {
	.owner = THIS_MODULE,
	.name = "adv7180",

	.id = I2C_DRIVERID_ADV7180,
	.flags = I2C_DF_NOTIFY,

	.attach_adapter = adv7180_attach_adapter,
	.detach_client = adv7180_detach_client,
	.command = adv7180_command,
};

static int __init
adv7180_init (void)
{
	return i2c_add_driver(&i2c_driver_adv7180);
}

static void __exit
adv7180_exit (void)
{
	i2c_del_driver(&i2c_driver_adv7180);
}

module_init(adv7180_init);
module_exit(adv7180_exit);
