/* Video For Linux 2 Interface for camera driver.
 * (C) 2006 TomTom BV.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *	
 * Author: Ithamar R. Adema, <ithamar.adema@tomtom.com>
 */

#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/dma-mapping.h>
#include <asm/arch/regs-camif.h>
#include <asm/arch/regs-lcd.h>
#include <asm/io.h>
#include <media/ovcamchip.h>
#include <linux/page-flags.h>
#include <linux/pagemap.h>
#include <linux/delay.h>
#include <asm/page.h>
#include <asm/system.h>
#include <barcelona/gopins.h>
#include <barcelona/gotype.h>

#ifdef CONFIG_CPU_FREQ
#include <linux/notifier.h>
#include <linux/cpufreq.h>
#include <asm/arch/regs-clock.h>
#endif
#include "camera.h"

#include <barcelona/video_overlay.h>

extern void unmap_pages( struct s3c_cam *cam );
extern void disable_cam_irq( struct s3c_cam *cam );
extern void enable_cam_irq( struct s3c_cam *cam );

/* Local Prototypes */
int cam_stream_off(struct s3c_cam *cam);


/*Get framerate from video mode*/
static int
framerate_from_stdid(v4l2_std_id * std)
{
	int framerate=50;
	
	if ((*std) & (V4L2_STD_NTSC_M | V4L2_STD_NTSC_M_JP | V4L2_STD_PAL_60))
	{
		framerate=60;
	}

	return framerate;
}


/* Set overlay resolution using Video Aspect Ratio calculations */
static int
cam_set_overlay_aspect_ratio(struct s3c_cam *cam, int force)
{
	int aspect;
	int left, top, xres, yres, bypp;

	if (!video_overlay)
		return -EINVAL;
	
	aspect = cam->aspect_ratio;

	video_overlay->get_resolution(&xres, &yres, &bypp);

	/* sensible defaults */
	left = 0;
	top  = 0;

	switch (aspect) {
	case V4L2_PRIV_ASPECT_RATIO_4_3:
		if ((xres == 480) &&  (yres == 272)) {
			left = 60;
			xres = 360;
		}
		break;
	case V4L2_PRIV_ASPECT_RATIO_14_9:
		if ((xres == 480) &&  (yres == 272)) {
			left = 30;
			xres = 420;
		} else if ((xres == 320) &&  (yres == 240)) {
			top = 17;
			yres = 206;
		}
		break;
	case V4L2_PRIV_ASPECT_RATIO_16_9:
		if ((xres == 320) &&  (yres == 240)) {
			top = 30;
			yres = 180;
		}
		break;
	default:
		/* nothing */
		break;
	}
	
	if ((aspect != V4L2_PRIV_ASPECT_RATIO_FILL) || (force))
	{
		cam->currfmt.type			= V4L2_BUF_TYPE_VIDEO_CAPTURE;
		cam->currfmt.fmt.pix.width              = xres;
		cam->currfmt.fmt.pix.height	        = yres;
		cam->currfmt.fmt.pix.pixelformat	= V4L2_PIX_FMT_RGB565;
		cam->currfmt.fmt.pix.field		= V4L2_FIELD_NONE;
		cam->currfmt.fmt.pix.bytesperline	= xres*bypp;
		cam->currfmt.fmt.pix.sizeimage		= cam->currfmt.fmt.pix.bytesperline * yres;
		cam->currfmt.fmt.pix.colorspace		= V4L2_COLORSPACE_SMPTE170M;
		
		cam->currwin.w.left                     = 0;
		cam->currwin.w.top                      = 0;
		cam->currwin.w.width                    = xres;
		cam->currwin.w.height                   = yres;
		
		cam->currwin.field                      = V4L2_FIELD_NONE;
		cam->currwin.clips                      = NULL;
		cam->currwin.clipcount                  = 0;
		cam->currwin.bitmap                     = NULL;
	}
	
	video_overlay->set_window(cam->currwin.w.left, cam->currwin.w.top, cam->currwin.w.left + left,
				cam->currwin.w.top + top, cam->currwin.w.width, cam->currwin.w.height, xres);
	
	/* zoom */
	if (cam->client && cam->client->driver->command)
		cam->client->driver->command( cam->client, VIDIOC_PRIV_GETDEFRES, &cam->defres );
	
	switch (aspect) {
	case V4L2_PRIV_ASPECT_RATIO_4_3_ZOOM:
		if ((xres == 480) &&  (yres == 272)) {
			cam->croprect.width  = cam->defres.width;
			cam->croprect.left   = cam->defres.left;
			cam->croprect.height = (9 * cam->defres.height)/12;
			cam->croprect.top    = (3 * cam->defres.height)/24;
		}
		break;
	case V4L2_PRIV_ASPECT_RATIO_14_9_ZOOM:
		/* not implemented */
		cam->croprect  = cam->defres;

		break;
	case V4L2_PRIV_ASPECT_RATIO_16_9_ZOOM:
		if ((xres == 320) &&  (yres == 240)) {
			cam->croprect.width  = (3 * cam->defres.width)/4;
			cam->croprect.left   = cam->defres.width/8;
			cam->croprect.height = cam->defres.height;
			cam->croprect.top    = cam->defres.top;
		}
		break;
	default:
		if ((aspect != V4L2_PRIV_ASPECT_RATIO_FILL) || (force))
		{
			cam->croprect  = cam->defres;
		}
		break;
	}
	
	return 0;
}

/* --- Functions handling the Video For Linux 2 ioctl() interface --- */
static int
cam_query_capabilities(struct s3c_cam *cam, struct v4l2_capability *arg)
{
	struct v4l2_capability cap;
	
	memset(&cap, 0, sizeof(cap));
	strcpy(cap.driver, CAMERA_DRIVER_NAME);
	strlcpy(cap.card, cam->vdev->name, sizeof(cap.card));
	strcpy(cap.bus_info, CAMERA_DRIVER_NAME);
	cap.bus_info[0] = 0;
	cap.version = CAMERA_VERSION_CODE;
	cap.capabilities = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_READWRITE;


	if (copy_to_user(arg, &cap, sizeof(cap)))
		return -EFAULT;
	
	return 0;
}

static int
cam_get_control(struct s3c_cam *cam, struct v4l2_control *arg)
{
	struct v4l2_control ctrl;
	int rc=0;
	
	if(copy_from_user(&ctrl, arg, sizeof(ctrl)))
		return -EFAULT;

	switch( ctrl.id )
	{
		case V4L2_CID_HFLIP :
			ctrl.value=(cam->vCICOTRGFMT & S3C2413_CICOTRGFMT_FLIPMD_XMIRROR) ? 1 : 0;
			break;

		case V4L2_CID_VFLIP :
			ctrl.value=(cam->vCICOTRGFMT & S3C2413_CICOTRGFMT_FLIPMD_YMIRROR) ? 1 : 0;
			break;

		case OVCAMCHIP_CID_FREQ :
			/* Workaround for framecorruption. To solve this problem we use LASTIRQ. However, this will */
			/* cause the actual framerate to be half of what we program. For this we need to modify what*/
			/* is specified. */
                        if (cam->client && cam->client->driver->command)
                                rc = cam->client->driver->command(cam->client, VIDIOC_G_CTRL, &ctrl);
			else
				return -EINVAL;

			/* Correct the value found. */
			if( cam->slots_in_use == 1 )
			{
				/* Workaround is in use. Modify what is set. */
				ctrl.value/=2;
			}
			break;

                default :
                        /* As we don't define other controls ourselfs, just pass it onto the ovcamchip module */
                        /* and see if the sensor module can handle it */
                        if (cam->client && cam->client->driver->command)
                                rc = cam->client->driver->command(cam->client, VIDIOC_G_CTRL, &ctrl);
			else
				return -EINVAL;
                        break;
        }
	
	if (copy_to_user(arg, &ctrl, sizeof(ctrl)))
		return -EFAULT;

	/* Done. */
	return rc;
}

static int
cam_set_control(struct s3c_cam *cam, struct v4l2_control *arg)
{
	struct v4l2_control ctrl;
	
	if (copy_from_user(&ctrl, arg, sizeof(ctrl)))
		return -EFAULT;

	switch( ctrl.id )
	{
		case V4L2_CID_HFLIP :
			if( ctrl.value ) cam->vCICOTRGFMT|=S3C2413_CICOTRGFMT_FLIPMD_XMIRROR;
			else cam->vCICOTRGFMT&=~(S3C2413_CICOTRGFMT_FLIPMD_XMIRROR);
			writel( cam->vCICOTRGFMT, S3C2413_CICOTRGFMT );
			break;

		case V4L2_CID_VFLIP :
			if( ctrl.value ) cam->vCICOTRGFMT|=S3C2413_CICOTRGFMT_FLIPMD_YMIRROR;
			else cam->vCICOTRGFMT&=~(S3C2413_CICOTRGFMT_FLIPMD_YMIRROR);
			writel( cam->vCICOTRGFMT, S3C2413_CICOTRGFMT );
			break;

		case OVCAMCHIP_CID_FREQ :
			/* Workaround for framecorruption. To solve this problem we use LASTIRQ. However, this will */
			/* cause the actual framerate to be half of what we program. For this we need to modify what*/
			/* is specified. */
			if( cam->slots_in_use == 1 )
			{
				/* Workaround is in use. Modify what is set. */
				ctrl.value*=2;
			}

		default :
		        /* As we don't define other controls ourselfs, just pass it onto the ovcamchip module */
			/* and see if the sensor module can handle it */
			if (cam->client && cam->client->driver->command)
				return cam->client->driver->command(cam->client, VIDIOC_S_CTRL, &ctrl);
			else
				return -EINVAL;
			break;
	}
	
	/* Done. */
	return 0;
}

static int
cam_enumerate_formats(struct s3c_cam *cam, struct v4l2_fmtdesc *arg)
{
	struct v4l2_fmtdesc f;	
	int index;
	
	if (copy_from_user(&f, arg, sizeof(f)))
		return -EFAULT;
	
	if (f.type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;
	
	index = f.index;
	
	memset(&f,0,sizeof(struct v4l2_fmtdesc));
	f.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	f.index = index;
	f.flags = 0;


	switch(index)
	{
		case 0:	/* V4L2_PIX_FMT_YUV420 - Most memory-friendly format */
			f.pixelformat = V4L2_PIX_FMT_YUV420;
			strcpy(f.description, "YUV 4:2:0 (planar)");
			break;

		case 1: /* V4L2_PIX_FMT_RGB24 - Most efficient for preview */
			f.pixelformat = V4L2_PIX_FMT_RGB24;
			strcpy(f.description, "24 bits RGB");
			break;

		case 2:	/* V4L2_PIX_FMT_RGB565 - Less resource hungry, but more processor intensive */
			f.pixelformat = V4L2_PIX_FMT_RGB565;
			strcpy(f.description, "5-6-5 bits RGB");
			break;

		case 3: /* V4L2_PIX_FMT_YUV422P - Highest quality YUV */
			f.pixelformat=V4L2_PIX_FMT_YUV422P;
			strcpy(f.description, "YUV 4:2:2 (planar)" );
			break;

		case 4:	/* V4L2_PIX_FMT_YUYV - YUV422 (YCbYCr) non planar. */
			f.pixelformat=V4L2_PIX_FMT_YUYV;
			strcpy(f.description, "YUV 4:2:2" );
			break;

		default:
			return -EINVAL;
	}
	
	if (copy_to_user(arg, &f, sizeof(f)))
		return -EFAULT;
	
	return 0;
}

static int
cam_try_format_capture(struct s3c_cam *cam, struct v4l2_format *fmt)
{
	
	/* Ignore what the app selected, we don't do interlaced */
	fmt->fmt.pix.field = V4L2_FIELD_NONE;

	/* CAMIF wants a target width as a multiple of 8 */
	if (fmt->fmt.pix.width % 8)
		fmt->fmt.pix.width += 8 - (fmt->fmt.pix.width % 8);

	/* Default to max of camera interface if dimensions are out of range */
	if (fmt->fmt.pix.width == 0 || fmt->fmt.pix.width > CAMI_MAX_WIDTH)
		fmt->fmt.pix.width = CAMI_MAX_WIDTH;
        if (fmt->fmt.pix.height == 0 || fmt->fmt.pix.height > CAMI_MAX_HEIGHT)
		fmt->fmt.pix.width = CAMI_MAX_HEIGHT;

	/* Calculate size of image/scanline */
	switch(fmt->fmt.pix.pixelformat) {
		case V4L2_PIX_FMT_YUV420:
			/* YUV4:2:0 Ysize=widthxheight, Usize=widthxheight/4, Vsize=widthxheight/4 */
			fmt->fmt.pix.bytesperline = fmt->fmt.pix.width;		/* U and V bytesperline=width/2, height/2. */
			fmt->fmt.pix.sizeimage=fmt->fmt.pix.width * fmt->fmt.pix.height * 3/2;
			break;

		case V4L2_PIX_FMT_YUYV:
		case V4L2_PIX_FMT_YUV422P:
			/* YUV4:2:2 Ysize=widthxheight, Usize=widthxheight/2, Vsize=widthxheight/2 */
			/* Doesn't matter whether we output planar or non planar for the size. */
			fmt->fmt.pix.bytesperline = fmt->fmt.pix.width;		/* U and V bytesperline=width/2, height/1. */
			fmt->fmt.pix.sizeimage=fmt->fmt.pix.width * fmt->fmt.pix.height * 2;
			break;

		case V4L2_PIX_FMT_RGB24:
			fmt->fmt.pix.bytesperline = fmt->fmt.pix.width * sizeof(__u32);
			fmt->fmt.pix.sizeimage = fmt->fmt.pix.height * fmt->fmt.pix.bytesperline;
			break;

		case V4L2_PIX_FMT_RGB565:
			fmt->fmt.pix.bytesperline = fmt->fmt.pix.width * sizeof(__u16);
			fmt->fmt.pix.sizeimage = fmt->fmt.pix.height * fmt->fmt.pix.bytesperline;
			break;

		default:
			return -EINVAL;
	}


	return 0;
}


static int
cam_try_format_overlay(struct s3c_cam *cam, struct v4l2_format *fmt)
{
	int xres, yres, bypp;
	
	if (!video_overlay)
		return -EINVAL;
		
	video_overlay->get_resolution(&xres, &yres, &bypp);
	
	/* Default to max of framebuffer if dimensions are out of range */
	if((fmt->fmt.win.w.left < 0) || (fmt->fmt.win.w.left >= xres)  ||
	   (fmt->fmt.win.w.width < 0)  || (fmt->fmt.win.w.width  >= xres) || 
	   (fmt->fmt.win.w.left + fmt->fmt.win.w.width > xres )) {
		fmt->fmt.win.w.left = 0;
		fmt->fmt.win.w.width = xres;
	}
	
	if((fmt->fmt.win.w.top < 0) || (fmt->fmt.win.w.top >= yres)  ||
	   (fmt->fmt.win.w.height < 0)  || (fmt->fmt.win.w.height  >= yres) || 
	   (fmt->fmt.win.w.top + fmt->fmt.win.w.height > yres )) {
		fmt->fmt.win.w.top = 0;
		fmt->fmt.win.w.height = yres;
	}
	
	/* framebuffer wants coordinates as a multiple of 4 */
	fmt->fmt.win.w.left  = fmt->fmt.win.w.left & ~3;
	fmt->fmt.win.w.top   = fmt->fmt.win.w.top & ~3;
	fmt->fmt.win.w.width  = (fmt->fmt.win.w.width + 3) & ~3;
	fmt->fmt.win.w.height = (fmt->fmt.win.w.height + 3) & ~3;
	
		
	fmt->fmt.win.field = V4L2_FIELD_NONE;	
	fmt->fmt.win.chromakey &= 0xffffff;
	
	if((fmt->fmt.win.clips)||(fmt->fmt.win.clipcount)||(fmt->fmt.win.bitmap))
		return -EINVAL;

	return 0;
}

static int
cam_try_format(struct s3c_cam *cam, struct v4l2_format *arg)
{
	struct v4l2_format fmt;
	
	if (copy_from_user(&(fmt.type), arg, sizeof(enum v4l2_buf_type)))
		return -EFAULT;
	
	switch(fmt.type) {
	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
		if(copy_from_user(&(fmt.fmt.pix), &(arg->fmt.pix),
		                 sizeof(struct v4l2_pix_format)))
			return -EFAULT;
		
		return cam_try_format_capture(cam,&fmt);
	case V4L2_BUF_TYPE_VIDEO_OVERLAY:
		if(copy_from_user(&(fmt.fmt.win), &(arg->fmt.pix), 
		                  sizeof(struct v4l2_window)))
			return -EFAULT;
		
		return cam_try_format_overlay(cam,&fmt);
	default:
		return -EINVAL;
	}
	
	if (copy_to_user(arg, &fmt, sizeof(fmt)))
		return -EFAULT;
}

static int
cam_set_scaler( struct s3c_cam *cam )
{
	unsigned long flags;
	short winHorOfst2, winVerOfst2;
	int src_width, src_height, dst_width, dst_height, h_shift, v_shift;
	int PreHorRatio_xx, PreVerRatio_xx, PreDstHeight_xx, 
	    PreDstWidth_xx, MainVerRatio_xx, MainHorRatio_xx, SHfactor_xx;
	long val;
	
	/* Verify the size is correct. */
	if( ((cam->croprect.width+cam->croprect.left) > cam->defres.width) ||
	    ((cam->croprect.height+cam->croprect.top) > cam->defres.height) ) 
		return -EINVAL;

	/* Save FIQ and IRQ flags, then disable them. This since we don't want the interrupt routine to modify */
	/* them once we're busy. */
	local_irq_save(flags);
	local_fiq_disable( );

	/* First set the registers correctly for the cropping/offset. */
	if( (cam->croprect.top != 0) || (cam->croprect.left != 0) ||
	    (cam->croprect.width < cam->defres.width) ||
	    (cam->croprect.height < cam->defres.height) )
	{
		cam->vCIWDOFST = S3C2413_CIWDOFST_WINOFSEN |
			(cam->croprect.left << S3C2413_CIWDOFST_WINHOROFST_SHIFT) |
			(cam->croprect.top << S3C2413_CIWDOFST_WINVEROFST_SHIFT);
		writel( cam->vCIWDOFST, S3C2413_CIWDOFST );

		winHorOfst2 = cam->defres.width - (cam->croprect.left + cam->croprect.width);
		winVerOfst2 = cam->defres.height - (cam->croprect.top + cam->croprect.height);
		cam->vCIDOWSFT2 = (winHorOfst2 << S3C2413_CIDOWSFT2_WINHOROFST2_SHIFT) |
			(winVerOfst2 << S3C2413_CIDOWSFT2_WINVEROFST2_SHIFT);
		writel( cam->vCIDOWSFT2, S3C2413_CIDOWSFT2 );
	}	
	else
	{
		/* Disable cropping. */
		cam->vCIWDOFST&=~(S3C2413_CIWDOFST_WINOFSEN | S3C2413_CIWDOFST_WINVEROFST_MASK | S3C2413_CIWDOFST_WINHOROFST_MASK);
		writel( cam->vCIWDOFST, S3C2413_CIWDOFST );
	}

	/* Setup CAMIF scaling. Note that we enable the scaler regardless of if we actually scale. This is because otherwise */
	/* we cannot output YUV420 in YUV mode, since the conversion is done by the scaler codec. */
	src_width = cam->croprect.width;
	src_height = cam->croprect.height;
	dst_width = cam->currfmt.fmt.pix.width;
	dst_height = cam->currfmt.fmt.pix.height;
	
	if (src_width >= 64 * dst_width)	{ local_irq_restore( flags ); return -EINVAL; /* Out Of Horizontal Scale Range */ }
	else if (src_width >= 32 * dst_width)	{ PreHorRatio_xx = 32; h_shift = 5; }
	else if (src_width >= 16 * dst_width)	{ PreHorRatio_xx = 16; h_shift = 4; }
	else if (src_width >= 8 * dst_width)	{ PreHorRatio_xx = 8;  h_shift = 3; }
	else if (src_width >= 4 * dst_width)	{ PreHorRatio_xx = 4;  h_shift = 2; }
	else if (src_width >= 2 * dst_width)	{ PreHorRatio_xx = 2;  h_shift = 1; }
	else					{ PreHorRatio_xx = 1;  h_shift = 0; }

	PreDstWidth_xx  = src_width / PreHorRatio_xx;
	MainHorRatio_xx = (src_width << 8) / (dst_width << h_shift);

	if	(src_height >= 64 * dst_height)		{ local_irq_restore( flags ); return -EINVAL; /* Out Of Vertical Scale Range */ }
	else if (src_height >= 32 * dst_height)		{ PreVerRatio_xx = 32; v_shift = 5; }
	else if (src_height >= 16 * dst_height)		{ PreVerRatio_xx = 16; v_shift = 4; }
	else if (src_height >= 8  * dst_height)		{ PreVerRatio_xx = 8;  v_shift = 3; }
	else if (src_height >= 4  * dst_height)		{ PreVerRatio_xx = 4;  v_shift = 2; }
	else if (src_height >= 2  * dst_height)		{ PreVerRatio_xx = 2;  v_shift = 1; }
	else						{ PreVerRatio_xx = 1;  v_shift = 0; }

	PreDstHeight_xx = src_height / PreVerRatio_xx;
	MainVerRatio_xx = (src_height << 8) / (dst_height << v_shift);
	SHfactor_xx     = 10 - (h_shift + v_shift);

	/* Make sure the scaler can handle this zoom factor */
	if (dst_width > src_width && (src_width / PreHorRatio_xx) > 320) {
		local_irq_restore( flags );
		return -EINVAL;
	}
	
	/* If we get here, we've got valid settings, so program scaler */
	cam->vCICOSCPRERATIO = (PreHorRatio_xx << S3C2413_CICOSCPRERATIO_PREHORRATIO_SHIFT) |
		(PreVerRatio_xx << S3C2413_CICOSCPRERATIO_PREVERRATIO_SHIFT) |
		(SHfactor_xx << S3C2413_CICOSCPRERATIO_SHFACTOR_SHIFT);
	writel(cam->vCICOSCPRERATIO, S3C2413_CICOSCPRERATIO);

	cam->vCICOSCPREDST = (PreDstWidth_xx << S3C2413_CICOSCPREDST_PREDSTWIDTH_SHIFT) |
		(PreDstHeight_xx << S3C2413_CICOSCPREDST_PREDSTHEIGHT_SHIFT);
	writel(cam->vCICOSCPREDST, S3C2413_CICOSCPREDST);

	val =	(MainHorRatio_xx << S3C2413_CICOSCCTRL_MAINHORRATIO_SHIFT) |
		(MainVerRatio_xx << S3C2413_CICOSCCTRL_MAINVERRATIO_SHIFT) |
		S3C2413_CICOSCCTRL_COSCALERSTART;				/* XXX No docs on this flag?? */

	/* Check if we're scaling up (H and V) */
	if (dst_width >= src_width) val |= S3C2413_CICOSCCTRL_SCALEUPH;
	if (dst_height >= src_height) val |= S3C2413_CICOSCCTRL_SCALEUPV;

	/* Set the scaler bit if used. */
	cam->vCIIMGCPT |= S3C2413_CIIMGCPT_IMGCPTEN_COSC;

	/* Write the remaining values. */
	cam->vCICOSCCTRL = val;
	writel(val, S3C2413_CICOSCCTRL);
	writel( cam->vCIIMGCPT, S3C2413_CIIMGCPT ); 

	/* Restore the flags. */
	local_irq_restore( flags );	
	return 0;
}

static int
cam_set_window( struct s3c_cam *cam )
{
	cam_set_overlay_aspect_ratio(cam, 0);

	if(cam->overlay_flags & V4L2_FBUF_FLAG_CHROMAKEY) {
		video_overlay->set_color_key(1, cam->currwin.chromakey);
	} else {
		video_overlay->set_color_key(0, 0);
	}
	

	if(cam->overlay_flags & V4L2_FBUF_FLAG_GLOBAL_ALPHA) {
		video_overlay->set_global_alpha(cam->currwin.global_alpha);
	} else {
		video_overlay->set_global_alpha(0xff);
	}
	return 0;
}

static int
cam_get_format(struct s3c_cam *cam, struct v4l2_format *arg)
{
	struct v4l2_format fmt;
	
	if (copy_from_user(&(fmt.type), arg, sizeof(enum v4l2_buf_type)))
		return -EFAULT;
	
	switch (fmt.type) {
	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
		if(cam->overlay)
			return -EBUSY;
		
		fmt = cam->currfmt;
		if (copy_to_user(&(arg->fmt.pix), &(fmt.fmt.pix), 
		                 sizeof(enum v4l2_buf_type) + sizeof(struct v4l2_pix_format)))
			return -EFAULT;
		
		return 0;
	case V4L2_BUF_TYPE_VIDEO_OVERLAY:
		if(!(cam->overlay))
			return -EBUSY;
		
		fmt.fmt.win = cam->currwin;
		
		if (copy_to_user(&(arg->fmt.win), &(fmt.fmt.win),
		                 sizeof(enum v4l2_buf_type) + sizeof(struct v4l2_window)))
			return -EFAULT;
		
		return 0;
	default:
		return -EINVAL;
	}
	
	return 0;
}

static int
cam_set_format(struct s3c_cam *cam, struct v4l2_format *arg)
{
	struct v4l2_format fmt;
	int rc;
		
	if (copy_from_user(&(fmt.type), arg, sizeof(enum v4l2_buf_type)))
		return -EFAULT;
	
	switch(fmt.type) {
	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
		if( cam->overlay )
			return -EBUSY;
	
		if(copy_from_user(&(fmt.fmt.pix), &(arg->fmt.pix),
			          sizeof(struct v4l2_pix_format)))
			return -EFAULT;
		
		/* Try format, and exit if fails */
		rc = cam_try_format_capture(cam, &fmt);
		if (rc)
			return rc;

		/* Store selected format for future reference */
		cam->currfmt = fmt;

		/* Program the scaler/cropping function. */
		return cam_set_scaler( cam );
	case V4L2_BUF_TYPE_VIDEO_OVERLAY:
		if(! cam->overlay )
			return -EBUSY;
		
		if(copy_from_user(&(fmt.fmt.win), &(arg->fmt.win),
			          sizeof(struct v4l2_window)))
			return -EFAULT;
		
		/* Try format, and exit if fails */
		rc = cam_try_format_overlay(cam, &fmt);
		if (rc)
			return rc;
		
		/* Store selected format for future reference */
		cam->currwin = fmt.fmt.win;
		
		return cam_set_window( cam );
	default:
		return -EINVAL;
	}
	return 0;
}

int ProgramBurstSize( struct s3c_cam *cam )
{
	unsigned short int	nr_datawordsY=0;
	unsigned short int	nr_datawordsUV=0;
	unsigned short int	MainBurstY=16;
	unsigned short int	RemBurstY=0;
	unsigned short int	MainBurstUV=16;
	unsigned short int	RemBurstUV=0;
	int			done=0;
	int			is_interleaved=0;

	/* Determine the linesize. */
	switch( cam->currfmt.fmt.pix.pixelformat )
	{
		case V4L2_PIX_FMT_RGB24:
			/* 32 bits / pixel */
			is_interleaved=1;
			nr_datawordsY=cam->currfmt.fmt.pix.width;
			nr_datawordsUV=0;
			break;

		case V4L2_PIX_FMT_RGB565:
		case V4L2_PIX_FMT_RGB555:
		case V4L2_PIX_FMT_YUYV:
			/* 32 bits / 2 pixel */
			is_interleaved=1;
			nr_datawordsY=cam->currfmt.fmt.pix.width/2;
			nr_datawordsUV=0;
			break;

		case V4L2_PIX_FMT_YUV420:
		case V4L2_PIX_FMT_YUV422P:
			is_interleaved=0;
			/* 32 bits / 4 pixel (planar, count only Y plane) */
			nr_datawordsY=cam->currfmt.fmt.pix.width/4;

			/* 32 bits / 8 pixel (planar, count only U or V plane) */
			nr_datawordsUV=cam->currfmt.fmt.pix.width/8;
			break;

		default:
			/* Shouldn't get here. */
			return -1;
	}

	/* Now try out which burst we can use. Keep in account that Main burst ought to be */
	/* 4,8,16 and remained burst ought to be 2,4,8,16. */
	done=0;
	do
	{
		/* Calculate what the remained burst would be. */
		RemBurstY=nr_datawordsY % MainBurstY;
		if( !RemBurstY ) RemBurstY=MainBurstY;

		/* Check if it's a valid value. */
		switch( RemBurstY )
		{
			case 2 :
			case 4 :
			case 8 :
			case 16:
				/* Valid result found. */
				done=1;
				break;

			default:
				/* Result is invalid. Try again. */
				MainBurstY>>=1;
				done=0;
				break;
		}
	} while( (MainBurstY >= 4) && !done );

	/* Check if the result is valid. */
	if( (MainBurstY < 4) || (RemBurstY < 2) )
		return -2;

	/* If this is a planar format, calculate for the U/V planes. For interleaved, there is a fixed ratio. */
	if( !is_interleaved )
	{
		/* Now do the same for the U/V planes. */
		done=0;
		do
		{
			/* Calculate what the remained burst would be. */
			RemBurstUV=nr_datawordsUV % MainBurstUV;
			if( !RemBurstUV ) RemBurstUV=MainBurstUV;

			/* Check if it's a valid value. */
			switch( RemBurstUV )
			{
				case 2 :
				case 4 :
				case 8 :
				case 16:
					/* Valid result found. */
					done=1;
					break;

				default:
					/* Result is invalid. Try again. */
					MainBurstUV>>=1;
					done=0;
					break;
			}
		} while( (MainBurstUV >= 4) && !done );

		/* Check if the result is valid. */
		if( (MainBurstUV < 4) || (RemBurstUV < 2) )
			return -3;
	}

	/* When we get here, we got valid values. */
	if( is_interleaved )
	{
		cam->vCICOCTRL=((MainBurstY/2) << S3C2413_CICOCTRL_YBURST1_SHIFT) |
		               ((RemBurstY/2) << S3C2413_CICOCTRL_YBURST2_SHIFT) |
		               ((MainBurstY/4) << S3C2413_CICOCTRL_CBURST1_SHIFT) |
		               ((RemBurstY/4) << S3C2413_CICOCTRL_CBURST2_SHIFT) |
		               (cam->vCICOCTRL & S3C2413_CICOCTRL_LASTIRQEN);
	}
	else
	{
		cam->vCICOCTRL=(MainBurstY << S3C2413_CICOCTRL_YBURST1_SHIFT) |
		               (RemBurstY << S3C2413_CICOCTRL_YBURST2_SHIFT) |
		               (MainBurstUV << S3C2413_CICOCTRL_CBURST1_SHIFT) |
		               (RemBurstUV << S3C2413_CICOCTRL_CBURST2_SHIFT) |
		               (cam->vCICOCTRL & S3C2413_CICOCTRL_LASTIRQEN);
	}

	/* Done. Now write. */
	writel( cam->vCICOCTRL, S3C2413_CICOCTRL );
	return 0;
}

int cam_stream_on(struct s3c_cam *cam)
{
	unsigned long int	val;
	unsigned long int	count=0;
	unsigned long int	index=0;
	dma_addr_t              dma_ptr;
	
	if (video_overlay)
	{
		/* clear video memory (2 buffers) */
		memset(cam->overlay_kernel_ptr, 0, cam->overlay_size * 2);
	}
	
	/* allocate memory in case we are not overlaying to the framebuffer */
	if ((!cam->overlay) && ( cam->kernel_ptr == NULL)) {
		cam->kernel_ptr = dma_alloc_writecombine(&cam->pdev->dev, CAMI_BUFFER_SIZE, &cam->dma_ptr, GFP_KERNEL);	
		if( cam->kernel_ptr == NULL)
		{
			printk( KERN_ERR "s3c-camif: cannot allocate DMA buffer.\n" );
			return -ENOMEM;
		}
				
		/* Allocate for the pages array. Always allocate the maximum frames that can be used so we can never run into */
		/* The situation where we do not have enough memory available. For this reason we assume RGB24 (something which*/
		/* will likely not happen). Add one incase the division by PAGE_SIZE gives a fraction. */
		cam->userspace.page_array=(struct page **) kcalloc( (CAMI_MAX_WIDTH*CAMI_MAX_HEIGHT*sizeof( __u32 )/PAGE_SIZE) + 1,
									sizeof( struct page * ), GFP_KERNEL );
		if( cam->userspace.page_array == NULL )
		{
			printk( KERN_ERR "s3c-camif: allocation of the page array buffer failed.\n" );
			return -ENOMEM;
		}
		
		cam->userspace.nr_pages=0;
	}
		
	/* Lock. */
	spin_lock(&cam->lock);

	/* Set ITU601_656n bit. This is required prior to resetting the camera interface. */
	writel(cam->vCISRCFMT | S3C2413_CISRCFMT_FMT_BT601, S3C2413_CISRCFMT);
	
	/* First reset camera interface, since we might have changed mode or format. */
	cam->vCIGCTRL=readl( S3C2413_CIGCTRL );
	writel( cam->vCIGCTRL | S3C2413_CIGCTRL_SWRST, S3C2413_CIGCTRL );

	/* Wait for 2ms for the interface to reset. Note: We do not reset the camera */
	/* itself. It is still in the correct mode/settings. */
	mdelay( 2 );

	/* Enable HREF mask */
	switch(IO_GetVideoDecoder())
	{
	case GOVD_ADV7180:
		cam->vCIGCTRL |= S3C2413_CIGCTRL_HREFMASK;
		break;
	default:
		break;
	}

	writel( cam->vCIGCTRL & ~(S3C2413_CIGCTRL_SWRST), S3C2413_CIGCTRL ); 

	/* Reprogram the scaler, as this one will have lost its settings. */
	if( cam_set_scaler( cam ) != 0 )
	{
		spin_unlock( &cam->lock );
		return -EINVAL;
	}

	/* Switch off any possible test modes. Make sure VSYNC polarity inversion is enabled. */
	val = readl(S3C2413_CIGCTRL) & ~S3C2413_CIGCTRL_TEST_MASK;
	val |= S3C2413_CIGCTRL_TEST_NORMAL;
	writel(val, S3C2413_CIGCTRL);
	cam->vCIGCTRL = val;
	
	/* Clear any overflows */
	cam->vCIWDOFST|=S3C2413_CIWDOFST_CLROVCOFIY | S3C2413_CIWDOFST_CLROVCOFICB | S3C2413_CIWDOFST_CLROVCOFICR;
	writel( cam->vCIWDOFST, S3C2413_CIWDOFST );
	cam->vCIWDOFST&=~(S3C2413_CIWDOFST_CLROVCOFIY | S3C2413_CIWDOFST_CLROVCOFICB | S3C2413_CIWDOFST_CLROVCOFICR);
	writel( cam->vCIWDOFST, S3C2413_CIWDOFST );
	
	/* Set source format for codec */
	switch(IO_GetVideoDecoder())
	{
	case GOVD_OV9655:
		cam->vCISRCFMT = S3C2413_CISRCFMT_FMT_BT601 | /* OV sensor feeds us this */
			S3C2413_CISRCFMT_UVOFF_NONE | /* It has "proper' YUV, not the +128 offsetted kind */
			(cam->defres.width << S3C2413_CISRCFMT_SRCHSZ_SHIFT) |
			(cam->defres.height << S3C2413_CISRCFMT_SRCVSZ_SHIFT);
		break;
	case GOVD_ADV7180:
		cam->vCICOCPTSEQ = 0x55555555; /* Only capture odd frames */
		writel(cam->vCICOCPTSEQ, S3C2413_CICOCPTSEQ);
                
		cam->vCIIMGCPT &=  ~(S3C2413_CIIMGCPT_DMAMODE_COUNT | (0x1f << S3C2413_CIIMGCPT_DMAPTR_SHIFT));
		writel(cam->vCIIMGCPT, S3C2413_CIIMGCPT);
		
		cam->vCISRCFMT = S3C2413_CISRCFMT_FMT_BT656 | /* ADV sensor feeds us this */
			S3C2413_CISRCFMT_UVOFF_NONE | /* It has "proper' YUV, not the +128 offsetted kind */
			S3C2413_CISRCFMT_ORDER422_MODE3 | /* CbYCrY order */
			(cam->defres.width << S3C2413_CISRCFMT_SRCHSZ_SHIFT) |
			(cam->defres.height << S3C2413_CISRCFMT_SRCVSZ_SHIFT);	
	}
	writel(cam->vCISRCFMT, S3C2413_CISRCFMT);

	/* Set output format for codec. Ensure the FLIP bits are copied from previous settings, to ensure these */
	/* pass between consecutive calls of stream on/stream off. */
	val = S3C2413_CICOTRGFMT_IN422 | /* Input for codec is YUV 4:2:2 */
		(cam->currfmt.fmt.pix.width << S3C2413_CICOTRGFMT_TRGHSZ_SHIFT) |
		(cam->currfmt.fmt.pix.height << S3C2413_CICOTRGFMT_TRGVSZ_SHIFT) |
		(cam->vCICOTRGFMT & S3C2413_CICOTRGFMT_FLIPMD_MASK);

	switch(cam->currfmt.fmt.pix.pixelformat) {
		case V4L2_PIX_FMT_RGB565:	val |= S3C2413_CICOTRGFMT_OUT422 | S3C2413_CICOTRGFMT_INTERLEAVE;       break;
		case V4L2_PIX_FMT_RGB555:	val |= S3C2413_CICOTRGFMT_OUT422 | S3C2413_CICOTRGFMT_INTERLEAVE;       break;
		case V4L2_PIX_FMT_RGB24:	val |= S3C2413_CICOTRGFMT_OUT422 | S3C2413_CICOTRGFMT_INTERLEAVE;	break;
		case V4L2_PIX_FMT_YUV422P:	val |= S3C2413_CICOTRGFMT_OUT422;					break;
		case V4L2_PIX_FMT_YUV420:	val &= ~(S3C2413_CICOTRGFMT_OUT422);					break;
		case V4L2_PIX_FMT_YUYV:		val |= S3C2413_CICOTRGFMT_OUT422 | S3C2413_CICOTRGFMT_INTERLEAVE;	break;
        	default:			spin_unlock(&cam->lock); return -EINVAL;
	}
	cam->vCICOTRGFMT = val;
	writel(val, S3C2413_CICOTRGFMT);

	/* Setup CAMIF DMA */
	if( ProgramBurstSize( cam ) )
		return -EINVAL;

	cam->vCICOTAREA = cam->currfmt.fmt.pix.width * cam->currfmt.fmt.pix.height;
	writel(cam->vCICOTAREA, S3C2413_CICOTAREA );

	/* Select the correct output. Note: val is used as bitmask, and the ImgCptEn_Cosc bit is copied from the register. */
	val=cam->vCIIMGCPT & S3C2413_CIIMGCPT_IMGCPTEN_COSC;
	switch(cam->currfmt.fmt.pix.pixelformat) {
		case V4L2_PIX_FMT_RGB24:	val |= S3C2413_CIIMGCPT_RGB | S3C2413_CIIMGCPT_RGBFMT_24BIT; break;
                case V4L2_PIX_FMT_RGB565:	val |= S3C2413_CIIMGCPT_RGB | S3C2413_CIIMGCPT_RGBFMT_16BIT; break;
                case V4L2_PIX_FMT_RGB555:	val |= S3C2413_CIIMGCPT_RGB | S3C2413_CIIMGCPT_RGBFMT_16BIT; break;
		case V4L2_PIX_FMT_YUYV:
		case V4L2_PIX_FMT_YUV420:
		case V4L2_PIX_FMT_YUV422P:	val |= S3C2413_CIIMGCPT_YUV | S3C2413_CIIMGCPT_RGBFMT_16BIT; break;
		default:			spin_unlock( &cam->lock ); return -EINVAL;
	}
	cam->vCIIMGCPT=val;

	/* Determine how many buffers we can use for this resolution. */
	if(cam->overlay) {
		if (IO_GetModelId() == GOTYPE_DUBLIN) {
			cam->slots_in_use=1;	// Dublin does not do double buffering
		} else {
			cam->slots_in_use=2;
		}
	} else {
		cam->slots_in_use=(__u16) (CAMI_BUFFER_SIZE/cam->currfmt.fmt.pix.sizeimage);
	}
	
	switch( cam->slots_in_use )
	{
		case 0 :
			/* Should not happen, but just in case. */
			spin_unlock( &cam->lock );
			return -EINVAL;
		case 1 :
		case 2 :
		case 4 :
			/* No need to do anything. */
			break;

		case 3 :
			/* For sake of simplicity, we don't do 3 buffers, we do 2 then. */
			cam->slots_in_use=2;
			break;

		default :
			/* More than 4 buffers are possible. However, our maximum is 4. */
			cam->slots_in_use=4;
			break;
	}

	/* Start the stream. */
	if(cam->overlay)
		dma_ptr = cam->overlay_dma_ptr;
	else
		dma_ptr = cam->dma_ptr;
	
	for( count=0; count < 16; count+=4 )
	{
		/* Which buffer index? */
		index=((count/4) % cam->slots_in_use) * ((unsigned long int) cam->currfmt.fmt.pix.sizeimage);

		/* Write Y. */
		if(cam->overlay)
			writel( dma_ptr + index, S3C2413_CICOYSA1 + count );
		else
			writel( dma_ptr + index, S3C2413_CICOYSA1 + count );
		
		switch( cam->currfmt.fmt.pix.pixelformat )
		{
			case V4L2_PIX_FMT_YUV422P:
			{
				unsigned long int	uoffset=cam->currfmt.fmt.pix.width*cam->currfmt.fmt.pix.height;
				/* 1 byte Y, 4 bits U, 4 bits V */
				writel( dma_ptr + index + uoffset, S3C2413_CICOCBSA1 + count );
				writel( dma_ptr + index + ((uoffset*3)/2), S3C2413_CICOCRSA1 + count );
				break;
			}

			case V4L2_PIX_FMT_YUV420:
			{
				unsigned long int	uoffset=cam->currfmt.fmt.pix.width*cam->currfmt.fmt.pix.height;
				/* 1byte Y, 2 bits U, 2 bits V */
				writel( dma_ptr + index + uoffset, S3C2413_CICOCBSA1 + count );
				writel( dma_ptr + index + ((uoffset*5)/4), S3C2413_CICOCRSA1 + count );
				break;
			}

			case V4L2_PIX_FMT_RGB565:
			case V4L2_PIX_FMT_RGB555:
			case V4L2_PIX_FMT_RGB24:
			case V4L2_PIX_FMT_YUYV:
			{
				/* Not used registers. Set to 0. */
				writel( 0, S3C2413_CICOCBSA1 + count );
				writel( 0, S3C2413_CICOCRSA1 + count );
				break;
			}
		}
	}

	/* Flag that we're streaming. Make sure the stopped flag is cleared (just in case). */
	/* Note that we need to set these flags before the camera is started, other wise the*/
	/* camera might be stopped again when the interrupt fires before these flags are set*/
	cam->flags|=S3C_CAM_STREAMING;
	cam->flags|=S3C_CAM_STATECHANGE;

	/* Flag the LASTIRQ bit if we only have space for one buffer. This to ensure that enough time is left */
	/* to copy the frame. This is to workaround a problem that the camera interface starts streaming the  */
	/* data for the next frame while we are still copying the current frame. */
	if (( cam->slots_in_use == 1 ) && (!cam->overlay))
		writel( cam->vCICOCTRL | S3C2413_CICOCTRL_LASTIRQEN, S3C2413_CICOCTRL );

	/* Now start the camera. */
	val=cam->vCIIMGCPT | S3C2413_CIIMGCPT_IMGCPTEN | S3C2413_CIIMGCPT_DMAEN;
	
	writel( val,  S3C2413_CIIMGCPT );
	cam->vCIIMGCPT=val;

	/* Restore IRQs. */
        spin_unlock(&cam->lock);

	if ((IO_GetModelId() == GOTYPE_DUBLIN)
		&& (video_overlay)
		&& (video_overlay->set_input_synchronisation))
	{
		video_overlay->set_input_synchronisation(TFT_ADJUST_TO_INPUTSYNC);
	}
	
	
	/*Hand framerate to overlay device*/
	if ((IO_GetModelId() == GOTYPE_DUBLIN)
		&& (cam->client)
		&& (cam->client->driver)
		&& (cam->client->driver->command)
		&& (video_overlay)
		&& (video_overlay->set_framerate))
	{
		v4l2_std_id std=0;

		cam->client->driver->command( cam->client, VIDIOC_QUERYSTD, &std );

		video_overlay->set_framerate(framerate_from_stdid(&std));
	}

	return 0;
}


static int cam_dumpregs(struct s3c_cam *cam)
{
	if(video_overlay)
		video_overlay->dump_regs();
		
	printk(KERN_DEBUG "CISRCFMT  = 0x%08X\n", readl(S3C2413_CISRCFMT));
	printk(KERN_DEBUG "CIWDOFST  = 0x%08X\n", readl(S3C2413_CIWDOFST));
	printk(KERN_DEBUG "CIGCTRL   = 0x%08X\n", readl(S3C2413_CIGCTRL));
	printk(KERN_DEBUG "CIDOWSFT2 = 0x%08X\n", readl(S3C2413_CIDOWSFT2));
	printk(KERN_DEBUG "CICOYSA1  = 0x%08X\n", readl(S3C2413_CICOYSA1));
	printk(KERN_DEBUG "CICOYSA2  = 0x%08X\n", readl(S3C2413_CICOYSA2));
	printk(KERN_DEBUG "CICOYSA3  = 0x%08X\n", readl(S3C2413_CICOYSA3));
	printk(KERN_DEBUG "CICOYSA4  = 0x%08X\n", readl(S3C2413_CICOYSA4));
	printk(KERN_DEBUG "CICOCBSA1 = 0x%08X\n", readl(S3C2413_CICOCBSA1));
	printk(KERN_DEBUG "CICOCBSA2 = 0x%08X\n", readl(S3C2413_CICOCBSA2));
	printk(KERN_DEBUG "CICOCBSA3 = 0x%08X\n", readl(S3C2413_CICOCBSA3));
	printk(KERN_DEBUG "CICOCBSA4 = 0x%08X\n", readl(S3C2413_CICOCBSA4));
	printk(KERN_DEBUG "CICOCRSA1 = 0x%08X\n", readl(S3C2413_CICOCRSA1));
	printk(KERN_DEBUG "CICOCRSA2 = 0x%08X\n", readl(S3C2413_CICOCRSA2));
	printk(KERN_DEBUG "CICOCRSA3 = 0x%08X\n", readl(S3C2413_CICOCRSA3));
	printk(KERN_DEBUG "CICOCRSA4 = 0x%08X\n", readl(S3C2413_CICOCRSA4));
	printk(KERN_DEBUG "CICOTRGFMT= 0x%08X\n", readl(S3C2413_CICOTRGFMT));
	printk(KERN_DEBUG "CICOCTRL  = 0x%08X\n", readl(S3C2413_CICOCTRL));
	printk(KERN_DEBUG "CICOSCPRERATIO= 0x%08X\n", readl(S3C2413_CICOSCPRERATIO));
	printk(KERN_DEBUG "CICOSCPREDST= 0x%08X\n", readl(S3C2413_CICOSCPREDST));
	printk(KERN_DEBUG "CICOSCCTRL= 0x%08X\n", readl(S3C2413_CICOSCCTRL));
	printk(KERN_DEBUG "CICOTAREA = 0x%08X\n", readl(S3C2413_CICOTAREA));
	printk(KERN_DEBUG "CICOSTATUS= 0x%08X\n", readl(S3C2413_CICOSTATUS));
	printk(KERN_DEBUG "CIIMGCPT  = 0x%08X\n", readl(S3C2413_CIIMGCPT));
	printk(KERN_DEBUG "CICOCPTSEQ= 0x%08X\n", readl(S3C2413_CICOCPTSEQ));
	printk(KERN_DEBUG "CICOSCOS  = 0x%08X\n", readl(S3C2413_CICOSCOS));

	if (cam->client && cam->client->driver->command)
		cam->client->driver->command( cam->client, VIDIOC_PRIV_DUMPREGS, 0 );
	
	return 0;
}


/*forward synchronisation request to video_overlay device*/
static int cam_sync(struct s3c_cam *cam, int newstat)
{
	if ((video_overlay)&&(video_overlay->set_input_synchronisation))
	{
		video_overlay->set_input_synchronisation(newstat);
	}
	
	return 0;
}

int
cam_stream_off(struct s3c_cam *cam)
{
	/* Lock */
	spin_lock(&cam->lock);
	disable_cam_irq( cam );
	
	/*Turn off synchornisation */
	if ((IO_GetModelId() == GOTYPE_DUBLIN)
		&& (video_overlay)
		&& (video_overlay->set_input_synchronisation))
	{
		video_overlay->set_input_synchronisation(TFT_IGNORE_INPUTSYNC);
	}

	
	if (cam->flags & S3C_CAM_STREAMING) {
		/* Flag we stopped streaming. */
		cam->flags&=~(S3C_CAM_STREAMING);

		/* The camera interface will generate at least one more IRQ before it's stopped so wait for the interface */
		/* to stop. */
		while( !(cam->flags & S3C_CAM_STATECHANGE) )
		{
			/* Ensure IRQs can happen again. */
			enable_cam_irq( cam );
			spin_unlock( &cam->lock ); 

			/* Wait for the camera to be stopped. */
			if( wait_event_interruptible(cam->inq, (cam->flags & S3C_CAM_STATECHANGE) ) )
	                        return -ERESTARTSYS;
	
			/* Event happened or condition reached. Get the lock and disable IRQs. */
			spin_lock( &cam->lock );
			disable_cam_irq( cam );
		}
	
		/* Clear the flag. It's stopped. */
		cam->flags&=~(S3C_CAM_STATECHANGE);
	}
	
	/* free memory for if we were not using overlaying */	
	if( cam->userspace.nr_pages != 0 ) {
		unmap_pages( cam ); 
		kfree( cam->userspace.page_array );
		cam->userspace.nr_pages=0;
	}
	
	if (cam->kernel_ptr  != NULL) {
		dma_free_writecombine( &cam->pdev->dev, CAMI_BUFFER_SIZE, cam->kernel_ptr, cam->dma_ptr);
	}
	
	enable_cam_irq( cam );
        spin_unlock(&cam->lock);
	return 0;
}

static int
cam_crop_capabilities(struct s3c_cam *cam, struct v4l2_cropcap* arg)
{
	struct v4l2_cropcap cap;
	
	if (copy_from_user(&cap, arg, sizeof(cap)))
		return -EFAULT;
	
	if (cap.type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	/* Set the bounds; area we can capture from */
	cap.bounds.top = 0;
	cap.bounds.left = 0;
	cap.bounds.width = cam->defres.width;;
	cap.bounds.height = cam->defres.height;

	/* Default rectangle */
	cap.defrect = cam->croprect;

	/* set the pixelratio; as we're doing all, set 1:1 */
	cap.pixelaspect.numerator = 1; 
	cap.pixelaspect.denominator = 1;

	if (copy_to_user(arg, &cap, sizeof(cap)))
		return -EFAULT;
	
	return 0;
}

static int
cam_get_crop(struct s3c_cam *cam, struct v4l2_crop* arg)
{
	struct v4l2_crop	crop;

	if (copy_from_user(&crop, arg, sizeof(crop))) 
		return -EFAULT;
	
	if (crop.type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	crop.c = cam->croprect;

	if (copy_to_user(arg, &crop, sizeof(crop))) 
		return -EFAULT;
	
	return 0;
}

static int
cam_set_crop(struct s3c_cam *cam, struct v4l2_crop* arg)
{
	struct v4l2_crop	crop;
	__s32			tmp_res;
	__s32			offperc;

	if (copy_from_user(&crop, arg, sizeof(crop))) 
		return -EFAULT;
	
	if (crop.type != V4L2_BUF_TYPE_VIDEO_CAPTURE) 
		return -EINVAL;

	/* Ensure that the format specified in fmt has resolutions that are multiples */
	/* of 8. This is a requirement for the camera interface. */

	/* First try to get the width as close as possible. */
	if( crop.c.width % 8 )
	{
		/* First calculate how far the window is offset into the total image, we need to */
		/* keep this similar so that the zoomed in picture doesn't change too much. */
		offperc=1000*crop.c.left/(cam->defres.width - crop.c.width);

		/* Calculate the new width. */
		tmp_res=8*(crop.c.width/8);
		if( (crop.c.width - tmp_res) > 4 )
			tmp_res+=8;

		/* Calculate the new horizontal offset. */
		crop.c.left=offperc * (cam->defres.width - tmp_res) / 1000;
		crop.c.width=tmp_res;
	}

	/* Now try to get the height as close as possible. */
	if( crop.c.height % 8 )
	{
		/* First calculate how far the window is offset into the total image, we need to */
		/* keep this similar so that the zoomed in picture doesn't change too much. */
		offperc=1000*crop.c.top/(cam->defres.height - crop.c.height);

		/* Calculate the new height. */
		tmp_res=8*(crop.c.height/8);
		if( (crop.c.height - tmp_res) > 4 )
			tmp_res+=8;

		/* Calculate the new vertical offset. */
		crop.c.top=offperc * (cam->defres.height - tmp_res) / 1000;
		crop.c.height=tmp_res;
	}

	/* Now make sure the offsets are also aligned on 8 pixel boundary. */
	if( crop.c.left % 8 )
		crop.c.left=8*(crop.c.left/8);
		
	if( crop.c.top % 8 )
		crop.c.top=8*(crop.c.top/8);

	cam->croprect = crop.c;

	/* Program the scaler/cropping function. */
	return cam_set_scaler( cam );
}



#ifdef USE_WITH_VDTEST
/*
If you want to use vdtest enable the USE_WITH_VDTEST define in the file:
- linux-s3c24xx/include/media/s3c-camif.h (included via camera.h)
*/
#warning "For release builds undefine USE_WITH_VDTEST in linux-s3c24xx/include/media/s3c-camif.h\n"

int
cam_r( struct s3c_cam *cam, int *arg )
{
	int adr;
	int val;
	
	if (copy_from_user(&adr, arg, sizeof(adr))) 
		return -EFAULT;

	val = readl( S3C2413_CISRCFMT + adr );
	
	if(copy_to_user(arg, &val, sizeof(val)))
		return -EFAULT;
	
	return 0;
}


int cam_w( struct s3c_cam *cam, int * arg )
{
	int args[2];
	
	if (copy_from_user(&args, arg, sizeof(args))) 
		return -EFAULT;
	printk(KERN_ERR "%08X->[%08x].\n",args[1],args[0]);
	writel( args[1], S3C2413_CISRCFMT+args[0] );
	
	return 0;
}

static int
cam_rr( struct s3c_cam *cam, int *arg )
{
	int res;
	int rc = -EINVAL;
	
	if (copy_from_user(&res, arg, sizeof(res))) 
		return -EFAULT;

	if (cam->client && cam->client->driver->command)
		rc = cam->client->driver->command( cam->client, VIDIOC_PRIV_CRR, &res );
	
	if(copy_to_user(arg, &res, sizeof(res)))
		return -EFAULT;
	
	return rc;
}

int cam_rw( struct s3c_cam *cam, int * arg )
{
	int rc = -EINVAL;
	int res;
	
	if (copy_from_user(&res, arg, sizeof(res))) 
		return -EFAULT;

	if (cam->client && cam->client->driver->command)
		rc = cam->client->driver->command( cam->client, VIDIOC_PRIV_CRW, &res );
	
		return 0;
}
#endif



static int
cam_get_defres( struct s3c_cam *cam, struct v4l2_rect *arg )
{
	struct v4l2_rect res;
	int rc = -EINVAL;
	
	if (cam->client && cam->client->driver->command)
		rc = cam->client->driver->command( cam->client, VIDIOC_PRIV_GETDEFRES, &res );
	
	if(copy_to_user(arg, &res, sizeof(res)))
		return -EFAULT;
	
	return rc;
}

int cam_set_defres( struct s3c_cam *cam, struct v4l2_rect *arg )
{
	int rc = -EINVAL;
	struct v4l2_rect res;
	
	if(copy_from_user(&res, arg, sizeof(res)))
		return -EFAULT;
			
	/* Check if we are streaming. No changing of this while streaming. */
	if( cam->flags & S3C_CAM_STREAMING  )
		return -EBUSY;

	/* Offsets are always 0. */
	if( res.left || res.top )
		return -EINVAL;

	/* Initialize the mode also for the Camera itself (the I2C registers). */
	if (cam->client && cam->client->driver->command)
		rc = cam->client->driver->command( cam->client, VIDIOC_PRIV_SETDEFRES, &res );
	
	if(rc);
	{
		printk( "Unable to set default resolution for the camera!\n");
		return rc;
	}
	
	/* Set the default resolution. */
	memcpy( &(cam->defres), &res, sizeof( cam->defres ) );

	/* Now make sure the crop region is within boundaries. */
	cam->croprect.top=0;
	cam->croprect.left=0;
	cam->croprect.width=cam->defres.width;
	cam->croprect.height=cam->defres.height;
	cam->currfmt.fmt.pix.width=cam->defres.width;
	cam->currfmt.fmt.pix.height=cam->defres.height;
	if( cam_set_scaler( cam ) != 0 )
	{
		printk( "Unable to reset cropping settings.\n" );
		return -EINVAL;
	}
	
	return 0;
}

static int
cam_enum_std(struct s3c_cam *cam, struct v4l2_standard *arg)
{
	struct v4l2_standard std;
	int rc = -EINVAL;
	
	if (copy_from_user(&std, arg, sizeof(std)))
		return -EFAULT;
	
	if (cam->client && cam->client->driver->command)
		rc =  cam->client->driver->command(cam->client, VIDIOC_ENUMSTD, &std);

	if (copy_to_user(arg, &std, sizeof(std)))
		return -EFAULT;
	
	return rc;
}

static int
cam_query_std(struct s3c_cam *cam, v4l2_std_id *arg)
{
	v4l2_std_id std;
	int rc = -EINVAL;
	
	if (cam->client && cam->client->driver->command)
		rc =  cam->client->driver->command(cam->client, VIDIOC_QUERYSTD, &std);

	if (copy_to_user(arg, &std, sizeof(std)))
		return -EFAULT;
	
	return rc;
}

static int
cam_get_std(struct s3c_cam *cam, v4l2_std_id *arg)
{
	v4l2_std_id std;
	int rc = -EINVAL;
		
	if (cam->client && cam->client->driver->command)
		rc =  cam->client->driver->command(cam->client, VIDIOC_G_STD, &std);
	
	if (copy_to_user(arg, &std, sizeof(std)))
		return -EFAULT;
	
	return rc;
}

static int
cam_set_std(struct s3c_cam *cam, v4l2_std_id *arg)	
{
	v4l2_std_id std;
	int rc = -EINVAL;
		
	if (copy_from_user(&std, arg, sizeof(std)))
		return -EFAULT;

	if (cam->client && cam->client->driver->command) {
		rc =  cam->client->driver->command(cam->client, VIDIOC_S_STD, &std);
		
		/*Hand framerate to overlay device*/
		if ((IO_GetModelId() == GOTYPE_DUBLIN)
			&& (video_overlay)
			&& (video_overlay->set_framerate))
		{
			video_overlay->set_framerate(framerate_from_stdid(&std));
		}
		/* when the video standard is changed, the default resolution may also be changed */
		if(!rc)
			rc = cam->client->driver->command( cam->client, VIDIOC_PRIV_GETDEFRES, &cam->defres );
		
		if(!rc)
			cam_set_scaler(cam);
	}
	
	return rc;
}

static int
cam_enum_input(struct s3c_cam *cam, struct v4l2_input *arg)
{
	struct v4l2_input input;
	int rc = -EINVAL;
		
	if (copy_from_user(&input, arg, sizeof(input)))
		return -EFAULT;
	
	if (cam->client && cam->client->driver->command)
		rc =  cam->client->driver->command(cam->client, VIDIOC_ENUMINPUT, &input);

	if (copy_to_user(arg, &input, sizeof(input)))
		return -EFAULT;
	
	return rc;
}

static int
cam_get_input(struct s3c_cam *cam, int *arg)
{
	int input;
	int rc = -EINVAL;

        if (cam->client && cam->client->driver->command)
		rc =  cam->client->driver->command(cam->client, VIDIOC_G_INPUT, &input);

	if (copy_to_user(arg, &input, sizeof(input)))
		return -EFAULT;
	
	return rc;

}

static int
cam_set_input(struct s3c_cam *cam, int *arg)
{
	int input;
	int rc = -EINVAL;
		
	if (copy_from_user(&input, arg, sizeof(input)))
		return -EFAULT;

	/* As we don't define controls ourselfs, just pass it onto the ovcamchip module, so sensor module can handle it */
        if (cam->client && cam->client->driver->command) {
		
		rc =  cam->client->driver->command(cam->client, VIDIOC_S_INPUT, &input);
		
		/* when the input is changed, the default resolution may also be changed */
		if(!rc)
			rc = cam->client->driver->command( cam->client, VIDIOC_PRIV_GETDEFRES, &cam->defres );
		
		if(!rc)
			cam_set_scaler(cam);
	}
	return rc;
}

static int
cam_query_control(struct s3c_cam *cam, struct v4l2_queryctrl *arg)
{
	struct v4l2_queryctrl ctrl;
	int rc = -EINVAL;
	
	if (copy_from_user(&ctrl, arg, sizeof(ctrl)))
		return -EFAULT;

	/* As we don't define controls ourselfs, just pass it onto the ovcamchip module, so sensor module can handle it */
        if (cam->client && cam->client->driver->command)
		rc =  cam->client->driver->command(cam->client, VIDIOC_QUERYCTRL, &ctrl);

	if (copy_to_user(arg, &ctrl, sizeof(ctrl)))
		return -EFAULT;
	
	return rc;
}

static int
cam_get_framebuffer( struct s3c_cam *cam, struct v4l2_framebuffer *arg )
{
	struct v4l2_framebuffer fbuf;
	int    xres, yres, bpp;
	void *kptr;
	unsigned long dma, size;
	
	if (!video_overlay)
		return -EINVAL;
		
	video_overlay->get_resolution(&xres, &yres, &bpp);
	 
	/* this is actually a destructive overlay, but it does not destroy the primary framebuffer, 
	 * so let's assume it is not destructive
	 */
	fbuf.capability = V4L2_FBUF_CAP_EXTERNOVERLAY | \
			  V4L2_FBUF_CAP_CHROMAKEY | \
			  V4L2_FBUF_CAP_GLOBAL_ALPHA;
	
	fbuf.flags =  cam->overlay_flags;
	
	video_overlay->get_buffer(0, &kptr, &dma, &size);
	
	fbuf.base = (void *) dma;
	
	fbuf.fmt.width		= xres;
	fbuf.fmt.height	        = yres;
	fbuf.fmt.pixelformat	= V4L2_PIX_FMT_RGB565;
	fbuf.fmt.field		= V4L2_FIELD_NONE;
	fbuf.fmt.bytesperline	= xres*bpp;
	fbuf.fmt.sizeimage	= xres*yres*bpp;
	fbuf.fmt.colorspace	= V4L2_COLORSPACE_SMPTE170M;
	
	if(copy_to_user(arg, &fbuf, sizeof(fbuf)))
		return -EFAULT;
	
	return 0;
}

static int
cam_set_framebuffer( struct s3c_cam *cam, struct v4l2_framebuffer *arg )
{
	struct v4l2_framebuffer fbuf;
	
	if (!video_overlay)
		return -EINVAL;
	
	if(copy_from_user(&fbuf, arg, sizeof(fbuf)))
		return -EFAULT;
	
	cam->overlay_flags = fbuf.flags;
	
	return cam_set_window( cam );
}

static int
cam_overlay( struct s3c_cam *cam, int enable )
{
	if(enable)
	{
		void *kptr;
		unsigned long dma, size;

		if (!video_overlay)
			return -EINVAL;
			

		if (cam->flags & S3C_CAM_STREAMING  )
			return -EBUSY;

		cam->overlay = 1;

		video_overlay->get_buffer(0, &kptr, &dma, &size);
		
		cam->overlay_kernel_ptr = kptr;
		cam->overlay_dma_ptr    = dma;
		cam->overlay_size       = size;
		
		cam->overlay_flags                      = 0;
		cam->currwin.chromakey                  = 0;
		cam->currwin.global_alpha               = 0xff;

		return cam_set_overlay_aspect_ratio(cam, 1);
	}
	else
	{
		cam->overlay = 0;
	}
	
	return 0;
}

static int
cam_get_wss_data( struct s3c_cam *cam, int *arg )
{
	int wss_data=-1;
	int rc=0;

	/* Let the video decoder module can handle it */
        if (cam->client && cam->client->driver->command)
		rc =  cam->client->driver->command(cam->client, VIDIOC_PRIV_GET_WSS_DATA, &wss_data);
 
	if(copy_to_user(arg, &wss_data, sizeof(wss_data)))
		return -EFAULT;

	return rc;
}

static int
cam_set_aspect_ratio( struct s3c_cam *cam, int arg )
{
	int rc=0;

	switch(arg) {
	case V4L2_PRIV_ASPECT_RATIO_FILL:
	case V4L2_PRIV_ASPECT_RATIO_AUTO:
	case V4L2_PRIV_ASPECT_RATIO_4_3:
	case V4L2_PRIV_ASPECT_RATIO_14_9:
	case V4L2_PRIV_ASPECT_RATIO_16_9:
	case V4L2_PRIV_ASPECT_RATIO_4_3_ZOOM:
	case V4L2_PRIV_ASPECT_RATIO_14_9_ZOOM:
	case V4L2_PRIV_ASPECT_RATIO_16_9_ZOOM:
		cam->aspect_ratio = arg;
		break;
	default:
		rc = -EINVAL;
		break;
	}

	return rc;
}

static int
cam_start_video( struct s3c_cam *cam )
{
	// Dublin does not do double buffering, so we user one buffer for the video output
	// and the other for a black screen during the switching of video sources
	// NOTE: This only works for Dublin
	video_overlay->set_buffer(0);
	return 0;
}

static int
cam_stop_video( struct s3c_cam *cam )
{
	// NOTE: This only works for Dublin
	video_overlay->set_buffer(1);	
	return 0;
}

static int
cam_get_aspect_ratio( struct s3c_cam *cam, int *arg )
{
	int rc=0;

	if(copy_to_user(arg, &cam->aspect_ratio, sizeof(cam->aspect_ratio)))
		return -EFAULT;

	return rc;
}

int
s3c_cam_do_ioctl(struct inode *inode, struct file *file, 
		 unsigned int cmd, void *arg)

{
	struct video_device	*vdev;
	struct s3c_cam		*cam;

	vdev = video_devdata(file);
	if (vdev == NULL)
		return -EFAULT;
	cam = video_get_drvdata(vdev);
	if (cam == NULL)
		return -EFAULT;

	switch(cmd) {
		/* we support Video For Linux _2_ only! */
		case VIDIOC_QUERYCAP:	return cam_query_capabilities(cam,arg);
		/* cropping */
		case VIDIOC_CROPCAP:	return cam_crop_capabilities(cam,arg);
		case VIDIOC_G_CROP:	return cam_get_crop(cam,arg);
		case VIDIOC_S_CROP:	return cam_set_crop(cam,arg);
		/* video standards */
		case VIDIOC_ENUMSTD:	return cam_enum_std(cam, arg);
		case VIDIOC_QUERYSTD:	return cam_query_std(cam, arg);
		case VIDIOC_G_STD:	return cam_get_std(cam, arg);
		case VIDIOC_S_STD:	return cam_set_std(cam, arg);
		/* video inputs */
		case VIDIOC_ENUMINPUT:	return cam_enum_input(cam, arg);
		case VIDIOC_G_INPUT:	return cam_get_input(cam, arg);
		case VIDIOC_S_INPUT:	return cam_set_input(cam, arg);
		/* controls */
		case VIDIOC_QUERYCTRL:	return cam_query_control(cam,arg);
		case VIDIOC_G_CTRL:	return cam_get_control(cam,arg);
		case VIDIOC_S_CTRL:	return cam_set_control(cam,arg);
		/* formats */
		case VIDIOC_ENUM_FMT:	return cam_enumerate_formats(cam,arg);
		case VIDIOC_G_FMT:	return cam_get_format(cam,arg);
		case VIDIOC_TRY_FMT:	return cam_try_format(cam,arg);
		case VIDIOC_S_FMT:	return cam_set_format(cam,arg);
		/* streaming */
		case VIDIOC_STREAMON:	return cam_stream_on(cam);
		case VIDIOC_STREAMOFF:	return cam_stream_off(cam);
		/* overlays */
		case VIDIOC_OVERLAY:	return cam_overlay( cam, (int) arg );
		case VIDIOC_S_FBUF:	return cam_set_framebuffer(cam, arg);
		case VIDIOC_G_FBUF:	return cam_get_framebuffer(cam, arg);
		
		case VIDIOC_PRIV_SETDEFRES:
					return cam_set_defres( cam, arg );
		case VIDIOC_PRIV_GETDEFRES:
					return cam_get_defres( cam, arg );
		case VIDIOC_PRIV_DUMPREGS:
					return cam_dumpregs( cam );
		case VIDIOC_PRIV_SYNC:
					return cam_sync( cam, (int) arg );
		case VIDIOC_PRIV_START_VIDEO:
					return cam_start_video(cam);
		case VIDIOC_PRIV_STOP_VIDEO:
					return cam_stop_video(cam);

#ifdef USE_WITH_VDTEST
/*
If you want to use vdtest enable the USE_WITH_VDTEST define in the file:
- linux-s3c24xx/include/media/s3c-camif.h (included via camera.h)
*/
#warning "For release builds undefine USE_WITH_VDTEST in linux-s3c24xx/include/media/s3c-camif.h\n"

		case VIDIOC_PRIV_CR:
					return cam_r(cam, arg);

		case VIDIOC_PRIV_CW:
					return cam_w(cam, arg);

		case VIDIOC_PRIV_CRR:
					return cam_rr(cam, arg);

		case VIDIOC_PRIV_CRW:
					return cam_rw(cam, arg);
#endif

		/* aspect ratio */
		case VIDIOC_PRIV_GET_WSS_DATA:
					return cam_get_wss_data( cam, arg );
		case VIDIOC_PRIV_SET_ASPECT_RATIO:
					return cam_set_aspect_ratio( cam, (int) arg );
		case VIDIOC_PRIV_GET_ASPECT_RATIO:
					return cam_get_aspect_ratio( cam, arg );
		default:
			/* Not supported by the S3C2413 part of the driver. Pass on to I2C client. */
			if (cam->client && cam->client->driver->command)
				return cam->client->driver->command(cam->client, cmd, arg);
	}

	return -ENOSYS;
}

void
s3c_cam_setup_defaults(struct s3c_cam* cam)
{
	cam->flags				= 0;
	cam->currfmt.type			= V4L2_BUF_TYPE_VIDEO_CAPTURE;
	cam->currfmt.fmt.pix.pixelformat	= V4L2_PIX_FMT_RGB565;
	cam->currfmt.fmt.pix.field		= V4L2_FIELD_NONE;
	cam->currfmt.fmt.pix.bytesperline	= OVCC_MAX_WIDTH*2;
	cam->currfmt.fmt.pix.sizeimage		= cam->currfmt.fmt.pix.bytesperline * OVCC_MAX_HEIGHT;
	cam->currfmt.fmt.pix.colorspace		= V4L2_COLORSPACE_SMPTE170M;
	
	cam->overlay                            = 0;
	cam->overlay_flags                      = 0;
	cam->overlay_kernel_ptr                 = NULL;
	cam->overlay_dma_ptr                    = 0;
	cam->overlay_size                       = 0;

	cam->aspect_ratio                       = V4L2_PRIV_ASPECT_RATIO_FILL;

	memset(&cam->currwin, 0, sizeof(cam->currwin));
			
#ifdef CONFIG_CPU_FREQ
	cam->curr_divider=((readl( S3C2410_CLKDIVN ) & S3C2412_CLKDIVN_CAMCLKDIV_MASK) >> 16) + 1;
#endif

	/* Done. */
	return;
}

struct i2c_client*
s3c_cam_get_i2c_client(void)
{
	struct i2c_client* client=NULL;
	
	switch(IO_GetVideoDecoder())
	{
	case GOVD_OV9655:
		client = i2c_get_client(I2C_DRIVERID_OVCAMCHIP, 0, NULL);
		break;
	case GOVD_ADV7180:
		client = i2c_get_client(I2C_DRIVERID_ADV7180, 0, NULL);
		break;
	default:
		/* nothing */
		break;
	}
	
	return client;
}
