/*
 * linux/drivers/serial/dcc.c
 *
 * serial port emulation driver for the JTAG DCC Terminal.
 *
 * DCC(JTAG1) protocol version for JTAG ICE/ICD Debuggers:
 * Copyright (C) 2003, 2004, 2005 Hyok S. Choi (hyok.choi@xxxxxxxxxxx)
 * SAMSUNG ELECTRONICS Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Changelog:
 * Oct-2003 Hyok S. Choi Created
 * Feb-2004 Hyok S. Choi Updated for serial_core.c and 2.6 kernel
 * Mar-2005 Hyok S. Choi renamed from T32 to DCC
 * Apr-2006 Hyok S. Choi revised including the MAJOR number
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/tty_flip.h>
#include <linux/major.h>

#include <asm/io.h>
#include <asm/irq.h>

#include <linux/serial_core.h>

/*
 * if real irq interrupt used for receiving character,
 * uncomment below line and modify the DCC_IRQ to correct one.
 * Unless polling method id used.
 */
//#define DCC_IRQ_USED
//#define DCC_IRQ (INT_N_EXT0)

#ifndef DCC_IRQ_USED
static struct work_struct dcc_poll_task;
static void dcc_poll(void *data);
#endif

#define UART_NR 1 /* we have only one JTAG port */

#ifdef CONFIG_SERIAL_DCC_STDSERIAL
/* use ttyS for emulation of standard serial driver */
#define SERIAL_DCC_NAME "ttyS"
#define SERIAL_DCC_MINOR 64
#define SERIAL_DCC_MAJOR TTY_MAJOR
#else
/* use ttyJ0(128) */
#define SERIAL_DCC_NAME "ttyJ"
#define SERIAL_DCC_MINOR 186
#define SERIAL_DCC_MAJOR 204
#endif

/* DCC Status Bits */
#define DCC_STATUS_RX (1 << 0)
#define DCC_STATUS_TX (1 << 1)

/* primitive JTAG1 protocol utilities */
static inline volatile u32 __dcc_getstatus(void)
{
	u32 ret;
	
	asm __volatile__ ("mrc p14, 0, %0, c0, c0 @ read comms ctrl reg"
	: "=r" (ret));
	
	return ret;
}

static inline volatile char __dcc_getchar(void)
{
	char c;
	
	asm __volatile__ ("mrc p14, 0, %0, c1, c0 @ read comms data reg"
	: "=r" (c));
	
	return c;
}

static inline void __dcc_putchar(char c)
{
	asm __volatile__ ("mcr p14, 0, %0, c1, c0 @ write a char"
	: /* no output register */
	: "r" (c));
}
/* end of JTAG1 dependencies */

/* [PH] This is basically a time-out protection: if no JTAG Probe
   is connected, __dcc_getstatus() would wait forever, blocking
   dcc_putchar() and hence the kernel to boot. Now we add a simple
   time_out counter, which (if reached) turns of DCC output
   completely (skip_dcc = 1) until dcc_getchar() receives something,
   so we're sure somebody has typed a character on the DCC input using
   a JTAG probe, to use the shell. */
static volatile int skip_dcc = 0;

static void dcc_putchar(struct uart_port *port, int ch)
{
	int timeout = 0;
	
	if ( skip_dcc )
		return;
	
	while (__dcc_getstatus() & DCC_STATUS_TX) {
		cpu_relax();
		if ( ++timeout > (10 * 1024 * 1024) ) {
			skip_dcc = 1;	/* no Probe-DCC attached (yet...) */
			return;
		}
	}
	__dcc_putchar((char)(ch & 0xFF));
}

static inline void xmit_string(struct uart_port *port, char *p, int len)
{
	for ( ; len; len--, p++)
		dcc_putchar(port, *p);
}

static inline void dcc_transmit_buffer(struct uart_port *port)
{
	struct circ_buf *xmit = &port->info->xmit;
	int pendings = uart_circ_chars_pending(xmit);
	
	if (pendings + xmit->tail > UART_XMIT_SIZE) {
		xmit_string(port, &(xmit->buf[xmit->tail]),
		UART_XMIT_SIZE - xmit->tail);
		xmit_string(port, &(xmit->buf[0]), xmit->head);
	} else
		xmit_string(port, &(xmit->buf[xmit->tail]), pendings);
	
	xmit->tail = (xmit->tail + pendings) & (UART_XMIT_SIZE-1);
	port->icount.tx += pendings;
}

static inline void dcc_transmit_x_char(struct uart_port *port)
{
	dcc_putchar(port, port->x_char);
	port->icount.tx++;
	port->x_char = 0;
}

static void dcc_start_tx(struct uart_port *port, int start)
{
	dcc_transmit_buffer(port);
}

static void dcc_stop_tx(struct uart_port *port, int stop)
{
	/* dummy */
}

static inline void dcc_rx_chars(struct uart_port *port)
{
	unsigned char ch;
	struct tty_struct *tty = port->info->tty;
	
	/*
	* check input.
	* checking JTAG flag is better to resolve the status test.
	* incount is NOT used for JTAG1 protocol.
	*/
	
	if (__dcc_getstatus() & DCC_STATUS_RX) {
		/* for JTAG 1 protocol, incount is always 1. */
		ch = __dcc_getchar();
	
		/* [PH] */
		skip_dcc = 0;	/* We're in DCC business again! */
		if ( ch == '\0' )
			return;	/* To get rid of 'spurious' input... */
	
		if (tty) {
			tty_insert_flip_char(tty, ch, TTY_NORMAL);
			port->icount.rx++;
			tty_flip_buffer_push(tty);
		}
	}
}

static inline void dcc_tx_chars(struct uart_port *port)
{
	struct circ_buf *xmit = &port->info->xmit;
	
	if (port->x_char) {
		dcc_transmit_x_char(port);
		return; 
	}
	
	if (uart_circ_empty(xmit) || uart_tx_stopped(port))
		return;
	
	dcc_transmit_buffer(port);
	
	if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS)
		uart_write_wakeup(port);
}

#ifdef DCC_IRQ_USED /* real IRQ used */
static irqreturn_t dcc_int(int irq, void *dev_id, struct pt_regs *regs)
{
	struct uart_port *port = dev_id;
	int handled = 0;
	
	spin_lock(&port->lock);
	
	dcc_rx_chars(port);
	dcc_tx_chars(port);
	
	handled = 1;
	spin_unlock(&port->lock);
	
	return IRQ_RETVAL(handled);
}
#else /* emulation by scheduled work */
static void dcc_poll(void *data)
{
	struct uart_port *port = data;
	
	spin_lock(&port->lock);
	
	dcc_rx_chars(port);
	dcc_tx_chars(port);
	
	schedule_delayed_work(&dcc_poll_task, 1);
	
	spin_unlock(&port->lock);
}
#endif /* end of DCC_IRQ_USED */

static unsigned int dcc_tx_empty(struct uart_port *port)
{
	return TIOCSER_TEMT;
}

static unsigned int dcc_get_mctrl(struct uart_port *port)
{
	return TIOCM_CTS | TIOCM_DSR | TIOCM_CD;
}

static int dcc_startup(struct uart_port *port)
{
#ifdef DCC_IRQ_USED /* real IRQ used */
	int retval;
	
	/* Allocate the IRQ */
	retval = request_irq(port->irq, dcc_int, SA_INTERRUPT,
	"serial_dcc", port);
	if (retval)
		return retval;
#else /* emulation */
	/* shcedule the polling work */
	schedule_delayed_work(&dcc_poll_task, 1);
#endif

 return 0;
}

static void dcc_shutdown(struct uart_port *port)
{
#ifdef DCC_IRQ_USED /* real IRQ used */
	free_irq(port->irq, port);
#else
	spin_lock(&port->lock);
	cancel_rearming_delayed_work(&dcc_poll_task);
	spin_unlock(&port->lock);
#endif
}

static void dcc_set_termios(struct uart_port *port, struct termios *termios,
 struct termios *old)
{
#ifdef DCC_IRQ_USED
	unsigned long flags;
#endif
	unsigned int baud, quot;
	
	/*
	* We don't support parity, stop bits, or anything other
	* than 8 bits, so clear these termios flags.
	*/
	termios->c_cflag &= ~(CSIZE | CSTOPB | PARENB | PARODD | CREAD);
	termios->c_cflag |= CS8;
	
	/*
	* We don't appear to support any error conditions either.
	*/
	termios->c_iflag &= ~(INPCK | IGNPAR | IGNBRK | BRKINT);
	
	/*
	* Ask the core to calculate the divisor for us.
	*/
	baud = uart_get_baud_rate(port, termios, old, 0, port->uartclk/16); 
	quot = uart_get_divisor(port, baud);
	
#ifdef DCC_IRQ_USED
	spin_lock_irqsave(&port->lock, flags);
#endif

	uart_update_timeout(port, termios->c_cflag, baud);

#ifdef DCC_IRQ_USED
	spin_unlock_irqrestore(&port->lock, flags);
#endif
}

static const char * dcc_type(struct uart_port *port)
{
	return port->type == PORT_DCC_JTAG1 ? "DCC" : NULL;
}

static int dcc_request_port(struct uart_port *port)
{
	return 0;
}

/*
 * Configure/autoconfigure the port.
 */
static void dcc_config_port(struct uart_port *port, int flags)
{
	if (flags & UART_CONFIG_TYPE) {
		port->type = PORT_DCC_JTAG1;
		dcc_request_port(port);
	}
}

/*
 * verify the new serial_struct (for TIOCSSERIAL).
 */
static int dcc_verify_port(struct uart_port *port, struct serial_struct *ser)
{
	int ret = 0;
	if (ser->type != PORT_UNKNOWN && ser->type != PORT_DCC_JTAG1)
		ret = -EINVAL;
	if (ser->irq < 0 || ser->irq >= NR_IRQS)
		ret = -EINVAL;
	if (ser->baud_base < 9600)
		ret = -EINVAL;
	return ret;
}

/* dummy operation handlers for uart_ops */
static void dcc_dummy_ops(struct uart_port *port)
{
}
static void dcc_dummy_ops_ui(struct uart_port *port, unsigned int ui)
{
}
static void dcc_dummy_ops_i(struct uart_port *port, int i)
{
}

static struct uart_ops dcc_pops = {
	.tx_empty = dcc_tx_empty,
	.set_mctrl = dcc_dummy_ops_ui,
	.get_mctrl = dcc_get_mctrl,
	.stop_tx = dcc_stop_tx,
	.start_tx = dcc_start_tx,
	.stop_rx = dcc_dummy_ops,
	.enable_ms = dcc_dummy_ops,
	.break_ctl = dcc_dummy_ops_i,
	.startup = dcc_startup,
	.shutdown = dcc_shutdown,
	.set_termios = dcc_set_termios,
	.type = dcc_type,
	.release_port = dcc_dummy_ops,
	.request_port = dcc_request_port,
	.config_port = dcc_config_port,
	.verify_port = dcc_verify_port,
};

static struct uart_port dcc_port = {
	.membase = (char*)0x12345678, /* we need these garbages */
	.mapbase = 0x12345678, /* for serial_core.c */
	.iotype = UPIO_MEM, 
#ifdef DCC_IRQ_USED
	.irq = DCC_IRQ,
#else
	.irq = 0,
#endif
	.uartclk = 14745600,
	.fifosize = 0,
	.ops = &dcc_pops,
	.flags = UPF_BOOT_AUTOCONF,
	.line = 0,
};


#ifdef CONFIG_SERIAL_DCC_CONSOLE
static void dcc_console_write(struct console *co, const char *s, unsigned int count)
{
	int i;
	struct uart_port *port = &dcc_port;

	for (i = 0; i < count; i++) {
		dcc_putchar(port, s[i]);
		if (s[i] == '\n')
			dcc_putchar(port, '\r');
	}
}

static int __init dcc_console_setup(struct console *co, char *options)
{
	struct uart_port *port = &dcc_port;
	int baud = 9600;
	int bits = 8;
	int parity = 'n';
	int flow = 'n';
	
	if (options)
		uart_parse_options(options, &baud, &parity, &bits, &flow);
	
	return uart_set_options(port, co, baud, parity, bits, flow);
}

extern struct uart_driver dcc_reg;
static struct console dcc_console = {
	.name = SERIAL_DCC_NAME,
	.write = dcc_console_write,
	.device = uart_console_device,
	.setup = dcc_console_setup,
	.flags = CON_PRINTBUFFER,
	.index = -1,
	.data = &dcc_reg,
};

static int __init dcc_console_init(void)
{
	register_console(&dcc_console);
	return 0;
}
console_initcall(dcc_console_init);

#define DCC_CONSOLE &dcc_console
#else
#define DCC_CONSOLE NULL
#endif

static struct uart_driver dcc_reg = {
	.owner = THIS_MODULE,
	.driver_name = SERIAL_DCC_NAME,
	.dev_name = SERIAL_DCC_NAME,
	.major = SERIAL_DCC_MAJOR,
	.minor = SERIAL_DCC_MINOR,
	.nr = UART_NR,
	.cons = DCC_CONSOLE,
};

static int __init dcc_init(void)
{
	int ret;
	
	printk(KERN_INFO "DCC: JTAG1 Serial emulation driver driver $Revision: 1.1.2.1 $\n");
	
	ret = uart_register_driver(&dcc_reg);
	
	if (ret)
		return ret;
	
#ifndef DCC_IRQ_USED
	/* initialize the poll work */
	INIT_WORK(&dcc_poll_task, dcc_poll, &dcc_port);
#endif

	uart_add_one_port(&dcc_reg, &dcc_port);
	
	return 0;
}

__initcall(dcc_init);

MODULE_DESCRIPTION("DCC(JTAG1) JTAG debugger console emulation driver");
MODULE_AUTHOR("Hyok S. Choi <hyok.choi@xxxxxxxxxxx>");
MODULE_SUPPORTED_DEVICE(SERIAL_DCC_NAME);
MODULE_LICENSE("GPL");
