/* drivers/serial/tomtom-i2c-uart.c
 *
 * Implementation of several functions needed to let the 
 * standard 8250 serial driver initialize the external 
 * UART found on some TomTom GO boards.  
 *
 * Copyright (C) 2008 TomTom BV <http://www.tomtom.com/>
 * Authors:
 * Edgar Held <edgar.held@tomtom.com>
 * Robert Delien <robert.delien(at)tomtom.com>
 * Tom Urlings <tom.urlings(at)tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/config.h>
 
#if defined (CONFIG_BARCELONA_EXTERNAL_I2C_UART)
 
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/err.h>
#include <linux/i2c.h>
#include <linux/serial_reg.h>

#include <asm/io.h>
#include <asm/hardware/clock.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-dsc.h>
#include <asm/arch/regs-irq.h>


#include <barcelona/gopins.h>
#include <barcelona/gotype.h>
       
#include "tomtom-uart.h"

#define	PKG	"tomtom-i2c-uart"

#define	I2C_UART_CLK		"xuart_dublin"
#define I2C_16IS740             0x90 // TODO frbi: check for correct address
#define	SC16IS740_A_SHIFT	3

/* register address byte for sc16is7x0
 * taken from nxp semiconductors sc16is7x0 datasheet
 *
 * bit		name		function
 * 7		-		not used
 * 6:3		A[3:0]		uarts internal register select
 * 2:1		CH1,CH0		channel select: CH1=0 CH0=0
 * 				other values are reserved and should not be used
 * 0		-		not used
 */

#undef	READ_REG_TRACE
#undef	WRITE_REG_TRACE

#define	READ_REG_RETRY		10
#define	WRITE_REG_RETRY		10

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

static struct clk* xuart_clk;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

static unsigned short normal_i2c[] = {
	I2C_16IS740 >> 1,
	(I2C_16IS740 >> 1) + 1,
	I2C_CLIENT_END
};

static unsigned short ignore = I2C_CLIENT_END;

static struct i2c_client_address_data addr_data = {
	.normal_i2c		= normal_i2c,
	.probe			= &ignore,
	.ignore			= &ignore,
	.force			= &ignore,
};

static struct i2c_client* SC16IS740_client;

static struct i2c_driver i2c_driver_16IS740;

static int SC16IS740_detect_client (struct i2c_adapter *adapter,
                       int                 address,
                       int                 kind)
{
        int ret;
        
        /* Check if the adapter supports the needed features */
        ret = i2c_check_functionality(
                adapter,
                I2C_FUNC_SMBUS_READ_BYTE | I2C_FUNC_SMBUS_WRITE_BYTE_DATA);

        if (!ret) {
                printk( KERN_ERR 
                        "%s: funcionality check failed with error %d\n", 
                        __FUNCTION__, ret);
                return 0;
        }

        SC16IS740_client = kmalloc(sizeof(struct i2c_client), GFP_KERNEL);
        if (!SC16IS740_client) {
                printk( KERN_ERR 
                        "%s: out of memory!\n", 
                        __FUNCTION__);
                return -ENOMEM;
        }
        
        memset(SC16IS740_client, 0, sizeof(struct i2c_client));
        SC16IS740_client->addr = address;
        SC16IS740_client->adapter = adapter;
	SC16IS740_client->driver = &i2c_driver_16IS740;
        SC16IS740_client->flags = I2C_CLIENT_ALLOW_USE;
        strlcpy(SC16IS740_client->name, "SC16IS740IIC", sizeof(SC16IS740_client->name));

        /* write scratchpad register using smbus i2c write instead of
         * hw_tomtom_external_uart_i2c_writebyte, as the latter 
         * returns void */
        ret = i2c_smbus_write_byte_data(
                SC16IS740_client, 
                (UART_SCR << 3),        /* mimic hw_tomtom_external_uart_i2c_writebyte */
                0xaa);
        if (ret)
                printk( KERN_ERR
                        "%s: error %d writing scratchpad register\n", 
                        __FUNCTION__, 
                        ret);
        else
                printk( KERN_INFO
                        "%s: SPR='0x%X'\n", 
                        __FUNCTION__, 
                        i2c_smbus_read_byte_data(SC16IS740_client, (7 << SC16IS740_A_SHIFT)));

        ret = i2c_attach_client(SC16IS740_client);
        if (ret)
                printk( KERN_ERR 
                        "%s: attach i2c client failed with error %d\n", 
                        __FUNCTION__, 
                        ret);

        return (ret);
}

static int
SC16IS740_attach_adapter (struct i2c_adapter *adapter)
{
        int ret;

        ret = i2c_probe(adapter, &addr_data, &SC16IS740_detect_client);
        if (ret)
                printk(KERN_ERR "%s; i2c_probe returned '%d'\n", __FUNCTION__, ret);

        return (ret);
}


static int
SC16IS740_detach_client (struct i2c_client *client)
{
	int err = i2c_detach_client(client);
	if (err) {
		return err;
	}

	kfree(client);
	client = 0;

	return 0;
}

static struct i2c_driver i2c_driver_16IS740 = {
	.owner = THIS_MODULE,
	.name = "SC16IS740",

	.id = I2C_DRIVERID_SC16IS740,
	.flags = I2C_DF_NOTIFY,

	.attach_adapter = SC16IS740_attach_adapter,
	.detach_client = SC16IS740_detach_client,
};


static int __init
SC16IS740_init (void)
{
	printk(KERN_DEBUG "%s: %s\n", PKG, __FUNCTION__);
	return i2c_add_driver(&i2c_driver_16IS740);
}


static void 
SC16IS740_exit (void)
{
        i2c_del_driver(&i2c_driver_16IS740);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

void hw_tomtom_external_i2c_uart_reset( void )
{

	printk(KERN_DEBUG "%s: %s\n", PKG, __FUNCTION__);

	IO_Activate(UART_RESET);
	udelay(10);
	
	IO_Deactivate(UART_RESET);
	mdelay(50);
}
EXPORT_SYMBOL(hw_tomtom_external_i2c_uart_reset);

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

void hw_tomtom_external_i2c_uart_suspend( void )
{
	printk(KERN_DEBUG "%s: %s\n", PKG, __FUNCTION__);
	if (xuart_clk) {
		clk_disable(xuart_clk);
	}
}
EXPORT_SYMBOL(hw_tomtom_external_i2c_uart_suspend);

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

void hw_tomtom_external_i2c_uart_resume( void )
{
	printk(KERN_DEBUG "%s: %s\n", PKG, __FUNCTION__);
	if (xuart_clk) {
		clk_enable(xuart_clk);
		hw_tomtom_external_i2c_uart_reset();
	}
}
EXPORT_SYMBOL(hw_tomtom_external_i2c_uart_resume);

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

unsigned int hw_tomtom_external_i2c_uart_get_clock_rate ( void )
{
	unsigned int rate;
	
	printk(KERN_DEBUG "%s: %s\n", PKG, __FUNCTION__);
	
	if (!xuart_clk) {
		printk(KERN_ERR "%s: no xuart clock!\n", __FUNCTION__);
		return 0; // will be regarded by 8250.c as error and not register this port.
	}
	
	rate = (unsigned int)clk_get_rate(xuart_clk);
	printk(KERN_DEBUG "%s: xuart_clk @ %u Hz\n", __FUNCTION__, rate);
	return rate;
}
EXPORT_SYMBOL(hw_tomtom_external_i2c_uart_get_clock_rate);

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

void hw_tomtom_external_i2c_uart_init( void )
{
	int ret;
	
	printk(KERN_DEBUG "%s: %s\n", PKG, __FUNCTION__);
	
	if (IO_HaveSC16IS740Uart()) {
		printk(KERN_INFO "TomTom GO External I2C Uart, (C) 2008 TomTom BV\n");
		printk(KERN_DEBUG "%s: initializing\n", PKG);
		
		/* Configure */
		IO_SetInterruptOnActivated(UART_INTA);
		
		IO_SetFunction(UART_CSA);
		
		/* UART_PWRSAVE and UART_RESET are outputs, so no init needed.	*/
		
		/* now set the correct states   */
		IO_Deactivate(UART_PWRSAVE);
		/* we are not in Power save mode now */
		IO_Deactivate(UART_RESET);
		
		xuart_clk = clk_get(NULL, I2C_UART_CLK);
		
		if (IS_ERR(xuart_clk)) {
			/* Ooops, clock not found. */
			printk(KERN_ERR "%s unable to find clock '%s'\n", PKG, I2C_UART_CLK);
			xuart_clk = NULL;
		} else if ((ret = clk_use(xuart_clk))) {
			printk(KERN_ERR "%s: unable to grab clock '%s' (%d)\n", PKG, I2C_UART_CLK, ret);
			xuart_clk = NULL;
		} else {
			clk_enable(xuart_clk);
			/* TODO: we can decrease power consumption more by not
			 * enabling the clock until it is realy being used.
			 * Create a hook from serial8250_startup(struct uart_port *port) */
		}
		
		hw_tomtom_external_i2c_uart_reset();
		
		/* Register the 16IS740 as an I2C client and initialize it. */
		SC16IS740_init();
		printk(KERN_DEBUG "%s: initialized\n", PKG);
	}
}
EXPORT_SYMBOL(hw_tomtom_external_i2c_uart_init);

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

void hw_tomtom_external_i2c_uart_exit( void )
{
	/* Only start poking the PIO config after we made sure that we're correct hardware type. */
	if (IO_HaveExternalUart()) {
		printk(KERN_DEBUG "%s: entry\n", PKG);
		IO_Deactivate(UART_INTA);
		
		IO_Activate(UART_PWRSAVE); /* put IC in power save mode */
		if (xuart_clk) {
			clk_disable(xuart_clk);
			clk_unuse(xuart_clk);
			clk_put(xuart_clk);
			xuart_clk = NULL;
		}

		SC16IS740_exit();
		printk(KERN_DEBUG "%s: exited\n", PKG);
	}
}
EXPORT_SYMBOL(hw_tomtom_external_i2c_uart_exit);


// Internal read & write functions
int read_reg(unsigned char reg) {
	int ret;
	int tries = READ_REG_RETRY;
	
	if (!SC16IS740_client) {
		printk(KERN_ERR "%s: non-existent SC16IS740_client\n", __FUNCTION__);
		return -1;
	}
	
	/* i2c reads may fail, yet this fuction doesn't allow to return failures
	 * so the only thing we can do is retry a couple of times and log a kernel
	 * message if the failure is persistent. */
	do {
		ret = i2c_smbus_read_byte_data(SC16IS740_client, reg);
		--tries;
	} while ((ret < 0) && tries);

	if (ret < 0) {
		printk(KERN_ERR "%s: i2c_smbus_read_byte_data(offs='%d') returned '%d'\n",
			__FUNCTION__, (reg >> SC16IS740_A_SHIFT), ret);
	}
	
#if defined(READ_REG_TRACE)
	printk(KERN_DEBUG "%s: reg['%d']=='0x%2.2x' tries='%d'\n", __FUNCTION__,
		(reg >> SC16IS740_A_SHIFT), ret, (READ_REG_RETRY - tries));
#endif	/* READ_REG_TRACE */

	return ret;
}

void write_reg(unsigned char reg, unsigned char value) {
	int ret;
	int tries = WRITE_REG_RETRY;
	
	if (!SC16IS740_client) {
		printk(KERN_ERR "%s: attempt to write register 0x%2.2x with 0x%2.2x to non-existent client\n",
			__FUNCTION__, (reg >> SC16IS740_A_SHIFT), value);
		return;
	}

	/* I2C writes may fail, yet this fuction doesn't allow to return failures */
	/* so the only thing we can do is retry a couple of times and log a kernel */
	/* message if the failure is persistent. */
	do {
		ret = i2c_smbus_write_byte_data(SC16IS740_client, reg, value);
		tries --;
	} while ((ret < 0) && (tries));

	if (ret < 0) {
		printk(KERN_ERR "%s: attempt to write register 0x%2.2x with 0x%2.2x failed\n",
			__FUNCTION__, reg, value);
	}
	
#if defined(WRITE_REG_TRACE)
	printk(KERN_DEBUG "%s: reg['%d']='0x%2.2x' tries='%d'\n", __FUNCTION__,
		(reg >> SC16IS740_A_SHIFT), value, (WRITE_REG_RETRY - tries));
#endif  /* WRITE_REG_TRACE */
}

unsigned char hw_tomtom_external_uart_i2c_readbyte(int offset)
{
	return (unsigned char)(read_reg(offset << SC16IS740_A_SHIFT) & 0xff);
}
EXPORT_SYMBOL(hw_tomtom_external_uart_i2c_readbyte);

void hw_tomtom_external_uart_i2c_writebyte(int offset, unsigned char byte)
{
	write_reg((offset << SC16IS740_A_SHIFT), byte);
}
EXPORT_SYMBOL(hw_tomtom_external_uart_i2c_writebyte);

#endif //CONFIG_BARCELONA_EXTERNAL_I2C_UART

/* EOF */

