/* drivers/video/tomtom/a035qn02.c
 *
 * Specific functionality for the Samsung A035QN02 screen.
 *
 * ######## NOTE #############
 * Based on preliminary datasheet A035QN0203-001 rev 001 (15-02-2007)
 *
 * ###########################
 *
 * Copyright (C) 2005 TomTom BV <http://www.tomtom.com/>
 * Author: Jeroen Taverne <jeroen.taverne@tomtom.com>
 * Author: Martin Jackson <martin.jackson@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifdef __KERNEL__
#include <linux/delay.h>
#include <linux/cpufreq.h>
#include <barcelona/gopins.h>
#include <linux/err.h>
#include <asm/hardware/clock.h>
#endif /* __KERNEL */
#ifdef __BOOTLOADER__
#include "timer.h"
#include "gopins.h"
#endif /* __BOOTLOADER__ */
#include "lcdregs.h"
#include "screeninfo.h"
#include "framerate.h"
#include "a035qn02.h"

#define MVAL_USED		(0)
#define MVAL			(0)

#define VBPD_320240		((1-1)&0xff)
#define VFPD_320240		((3-1)&0xff)
#define VSPW_320240		((1-1)&0x3f)
#define HBPD_320240		((6-1)&0x7f)
#define HFPD_320240		((7-1)&0xff)
#define HSPW_320240		((2-1)&0xff)
#define DISABLE_DE		(1 << 6)
#define A035QN02_HBP		(0x21)
#define A035QN02_VBP		(0x8)

static void A035QN02_WriteByte(unsigned char data)
{
	int bit;

	for (bit=0;bit<8;bit++)
	{
		IO_Deactivate(LCD_SCL);
		if (data & 0x80) IO_Activate(LCD_SDI); else IO_Deactivate(LCD_SDI);
		IO_Activate(LCD_SCL);
		data <<= 1;
	}
}

static void A035QN02_Tx_Data(unsigned char Index,unsigned short REG_DATA)
{
	unsigned char DeviceID = 0x70;
	unsigned char lcd_Index = (unsigned char)(Index & 0xff);
	unsigned char lcd_data0 = (unsigned char)((REG_DATA >> 8)& 0xff);
	unsigned char lcd_data1 = (unsigned char)( REG_DATA & 0xff);

	IO_Activate(LCD_SCL);
	IO_Deactivate(LCD_CS);

	IO_Activate(LCD_CS);
	A035QN02_WriteByte(DeviceID|0x0);      //Send Device ID code
	A035QN02_WriteByte(0x00);	     //Write register address 8 bit
	A035QN02_WriteByte(lcd_Index);     //Write register address 8 bit
	IO_Deactivate(LCD_CS);
	mdelay(1);

	IO_Activate(LCD_CS);
	A035QN02_WriteByte(DeviceID|0x2);      //Send Device ID code
	A035QN02_WriteByte(lcd_data0);     //Write the data first 8 bit
	A035QN02_WriteByte(lcd_data1);     //Write the data second 8 bit
	IO_Deactivate(LCD_CS);
	mdelay(1);                 //need to delay 50us
}

void A035QN02_Init(void)
{
	unsigned long int	divider=(unsigned long int) get_clkval( &ttgfb_a035qn02_lcdspecs,
									ttgfb_a035qn02_lcdspecs.refresh_max,
									ttgfb_a035qn02_lcdspecs.curr_hclk );
	unsigned long int	hozval;
	unsigned long int	lineval;

	/* Get HOZVAL and LINEVAL. */
	get_hozlineval( &ttgfb_a035qn02_lcdspecs, &hozval, &lineval );

	/* Start of LCD Power On Sequence */
	IO_Deactivate(LCD_RESET);
	mdelay(20);
	IO_Activate(LCD_RESET);
	mdelay(10);
	IO_Deactivate(LCD_RESET);
	mdelay(10);

	// Start display controller
	rLCDCON1 &= ~1;
	rTCONSEL = 0;
	rLCDCON1 = (divider << 8) | (MVAL_USED << 7) | (((unsigned long int) ttgfb_a035qn02_lcdspecs.pnr_mode) << 5) |
		   (((unsigned long int) ttgfb_a035qn02_lcdspecs.bpp_mode) << 1) | (0 << 0);
	rLCDCON2 = (((unsigned long int) ttgfb_a035qn02_lcdspecs.screen.tft.vbpd) << 24) |
		   (lineval << 14) |
		   (((unsigned long int) ttgfb_a035qn02_lcdspecs.screen.tft.vfpd) << 6) |
		   (((unsigned long int) ttgfb_a035qn02_lcdspecs.screen.tft.vspw) << 0);
	rLCDCON3=(((unsigned long int) ttgfb_a035qn02_lcdspecs.screen.tft.hbpd) << 19) | (hozval << 8) |
		 (((unsigned long int) ttgfb_a035qn02_lcdspecs.screen.tft.hfpd) << 0);
	rLCDCON4 = (MVAL << 8) | (((unsigned long int) ttgfb_a035qn02_lcdspecs.screen.tft.hspw) << 0);
	rLCDCON5=(1<<11)|(1<<10)|(1<<9)|(1<<8)|(0<<6)|(1<<0);

	A035QN02_Tx_Data(0x01,0x2AEF);
	A035QN02_Tx_Data(0x03,0x920E);
	A035QN02_Tx_Data(0x0C,0x0002);
	A035QN02_Tx_Data(0x0D,0x000C);
	A035QN02_Tx_Data(0x0E,0x3100);
 	A035QN02_Tx_Data(0x10,0x00DC);	
	A035QN02_Tx_Data(0x1E,0x00A7);	
	A035QN02_Tx_Data(0x30,0x0304);
	A035QN02_Tx_Data(0x31,0x0507);
	A035QN02_Tx_Data(0x32,0x0405);
	A035QN02_Tx_Data(0x33,0x0007);
	A035QN02_Tx_Data(0x34,0x0507);
	A035QN02_Tx_Data(0x35,0x0004);
	A035QN02_Tx_Data(0x36,0x0605);
	A035QN02_Tx_Data(0x37,0x0103);
	A035QN02_Tx_Data(0x3A,0x000F);
	A035QN02_Tx_Data(0x3B,0x000F);
	
    rLCDCON1 |= 1;
    
    mdelay(164);  //tshut-on Typical is 164 ms
}

void A035QN02_Off(void)
{
	// Stop display controller
	rLCDCON1 = 0;

	IO_Activate(LCD_RESET);
	IO_Deactivate(LCD_SCL);
	IO_Deactivate(LCD_SDI);
	IO_Activate(LCD_CS);
}

#if defined CONFIG_CPU_FREQ && defined CONFIG_S3C24XX_DFS_CPUFREQ 
struct lcd_screen_info ttgfb_a035qn02_lcdspecs = {"SAM", "A035QN02", 70, 53, 320, 240, 25, 75, A035QN02_Init, A035QN02_Off, NULL, 
						   PNR_TFT, BPP_16BPP_TFT, 0, {{VSPW_320240, VBPD_320240, VFPD_320240,
						   HSPW_320240, HBPD_320240, HFPD_320240}}, lcd_def_freq_policy, 
						   lcd_def_freq_trans};
#else
struct lcd_screen_info ttgfb_a035qn02_lcdspecs = {"SAM", "A035QN02", 70, 53, 320, 240, 25, 75, A035QN02_Init, A035QN02_Off, NULL,
						   PNR_TFT, BPP_16BPP_TFT, 0, {{VSPW_320240, VBPD_320240, VFPD_320240,
						   HSPW_320240, HBPD_320240, HFPD_320240}}};
#endif

/* EOF */

