/* drivers/video/tomtom/s3c2443_tft.c
 *
 * TomTom GO S3C2443 TFT LCD Controller
 * based on ttgfb.c
 *
 * Copyright (C) 2005,2006,2007 TomTom BV <http://www.tomtom.com/>
 * Author: Onno Hovers <onno.hovers@tomtom.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifdef __KERNEL__
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/tomtomgofb.h>
#include <linux/dma-mapping.h>
#include <linux/cpufreq.h>
#include <linux/bootmem.h>

#include <asm/hardware/clock.h>
#include <asm/arch/regs-gpio.h>
#include <asm/uaccess.h>

#include <barcelona/gopins.h>
#include <barcelona/gotype.h>
#include <barcelona/debug.h>
#include <barcelona/ttgfb.h>
#endif /* __KERNEL__ */

#ifdef __BOOTLOADER__
#include "timer.h"
#include "gopins.h"
#include "gotype.h"
#include "option.h"
#endif /* __BOOTLOADER__ */

#include "lcd_controller.h"
#include "screeninfo.h"
#include "lcdregs.h"

#define BYTE_PER_PIXEL 2

static unsigned long resume_colorkey;
static unsigned long resume_alpha;
static int resume_colorkey_defined = 0;

#define _IGNORE_INPUTSYNC	0
#define _FOLLOW_INPUTSYNC	1

int last_set_framerate=0;
int last_requested_framerate=0;

/*DEBUG_GRAPHICS  bit masks      1:sync location flow       2:sync location    */
#define DEBUG_GRAPHICS 0

/* Enable when also horizontal settings should be changed for setting framerate: */
//#define S3C2443_TFT_TWEAK


int lines_delta=0;		/*deviatino on LINES can be changed each frame*/
int lines_step=0;		/*deviation on VPFD can be changed each frame*/
int lines_step_max=8;	/*maximum value for lines_step*/

unsigned int lineval=0; /*original LINEVAL*/
unsigned int vfpd=0; 	/*original VFPD */

unsigned int last_vsync_y=0;	/*y value of where TFT was when last time "sync" was called*/
int	vsync_adj_state=0;			/*state of sync-follow system.*/
int follow_y = 10;				/*sync follow system tries to synchronise with this line on TFT. */
								/*This line is always _later_ than when the real hardware pulse(irq) is received!!!*/
int corr_aggr=2;				/*When running of sync, this difference is devided by this value to correct*/

#if  (DEBUG_GRAPHICS)
unsigned int cpubase0=0;
unsigned int cpubase1=0;
unsigned int idxy=0;
int flags=0x6;
#endif


static int s3c2443_set_framerate(int framerate);



static void s3c2443_input_sync(void)
{
	int y=(rVIDCON1>>16)&0x3ff;		/*Get current TFT Y value*/
	
	switch (vsync_adj_state)
	{

	case _FOLLOW_INPUTSYNC:
		{
			int d = y-follow_y;		/*difference between current line and target line*/
			if (d!=0)
			{
				unsigned long tcon0=rVIDTCON0 & 0xffff00ff;
				int i;
				int	new_vfpd=0;
	
				i = corr_aggr-1;
				if (d<0) i = -corr_aggr+1;
				lines_step=((d+i)/corr_aggr);	/*correction value for adjustment*/
				/* if correction is to big then limit it */
				if (lines_step>lines_step_max) lines_step=lines_step_max;
				if (lines_step<-lines_step_max) lines_step=-lines_step_max;

				/* Use correction value*/
				new_vfpd=vfpd+lines_step;
				if (new_vfpd<1) new_vfpd=1;
				if (new_vfpd>254) new_vfpd=254;
				tcon0 |= (new_vfpd<<8);
				if (rVIDTCON0 != tcon0) rVIDTCON0 = tcon0;
	
			}
		}
		break;

	case _IGNORE_INPUTSYNC:
	default:
		// do nuffing.
		break;
	}
	

#if (DEBUG_GRAPHICS & 1)
	if ((cpubase0)&&(flags&2))
	{
		memset(cpubase0+idxy,0x55,y);
		memset(cpubase0+idxy+y,0xaa,310-y);
		if (lines_step>0)
		{
			memset(cpubase0+idxy+300,0xcc,lines_step);
		}
		else if (lines_step <0)
		{
			memset(cpubase0+idxy+300+lines_step,0x33,lines_step);
		}
		idxy+=960;
		if (idxy>(960*60)){idxy=0;}
	}
#endif
#if (DEBUG_GRAPHICS & 2)
	if ((cpubase0)&&(flags&4))
	{
		memset(cpubase0+960*last_vsync_y+900,0,8);
		memset(cpubase0+960*y+900,last_set_framerate+0x78,8);
	}
#endif	
	
	last_vsync_y = y ;
}



static void s3c2443_set_input_synchronisation(int newstate)
{

	printk(KERN_ERR "set_input_synchronisation(%d).\n",newstate);
	
	switch (newstate)
	{
	default:
	case TFT_IGNORE_INPUTSYNC:
		vsync_adj_state=_IGNORE_INPUTSYNC;
		/* make sure that there will be a proper screen visible*/
		rVIDCON0 &= (~2);
		if (lineval) rVIDTCON2 = (rVIDTCON2 & 0x7ff) | (lineval<<11);
		if (vfpd) rVIDTCON0 = (rVIDTCON0 & 0xffff00ff) | (vfpd<<8);			
		rVIDCON0 |=3;
		break;
	case TFT_ADJUST_TO_INPUTSYNC:
		if (last_requested_framerate!=last_set_framerate) s3c2443_set_framerate(last_requested_framerate);
		vsync_adj_state=_FOLLOW_INPUTSYNC;
		break;
			
	}
}

/*
framerate is calculated as: dotclock/(total image+boundary area)
That area is calculated as: (hspw+hfpd+hbpd+hozval)*(vspw,vbpd,vfpd,lineval)
Keeping the same display picture, the framerate can by changed by adjusting the hspw,hfpd,hbpd,vspw,vbpd and vfpd.
The vfpd part is the one that will be used for this. If framerate is still to low for that, then (when enabled
with S3C2443_TFT_TWEAK)  the values of hspw, hbpd and hfpd will be changed.
Possible addition if vfpd is not big enough to decrease the framerate:
  - use one of the v-ones that will not change the image position on the screen
  - increase liveval (but do not decrease to under what we want to view)
 */
static int s3c2443_set_framerate(int framerate)
{
	long clksel = 133000000;
	long clkdiv = ((rVIDCON0>>6)&0x3f)+1;
	int vspw,vbpd,_vfpd,_lineval,vtotal;
	int hspw,hfpd,hbpd,hozval,htotal;
	unsigned long vidcon=rVIDTCON0;
	
	last_requested_framerate = framerate;
	
	/*collect all register values*/
	vspw = 1+(vidcon&255);
	_vfpd = 1+((vidcon>>8)&255);
	vbpd = 1+((vidcon>>16)&255);
	vidcon=rVIDTCON1;
	hspw = 1+(vidcon&255);
	hfpd = 1+((vidcon>>8)&255);
	hbpd = 1+((vidcon>>16)&255);
	vidcon=rVIDTCON2;
	hozval = 1+(vidcon&0x7ff);
	_lineval = 1+(vidcon>>11);

	/* use shadow values in case we are following the sync. Because in that case, the register values will change all the time */
	if (lineval) _lineval=lineval;
	if (vfpd) _vfpd=vfpd;
	
	/*total line length*/
	htotal = hspw+hfpd+hbpd+hozval;
	
	/* calc new vtotal & vfpd  needed for requested framerate*/
    vtotal = clksel/(clkdiv*framerate*htotal);
	_vfpd = vtotal-vspw-vbpd-_lineval;

#ifdef 	S3C2443_TFT_TWEAK
	printk(KERN_ERR "Trying to set tft refresh rate by tweaking horizontal values.\n");
	/* change value till we get closest to needed vfpd*/
	while ((_vfpd<3)&&(hspw>4)&&(hbpd>4)&&(hfpd>2))
	{
		if (hfpd>hspw)
		{
			hfpd--;
		}
		else
		{
			if (hspw>hbpd)
			{
				hspw--;
			}
			else
			{
			    if (hbpd>4) hbpd--;
			}
		}
		
		htotal = hspw+hfpd+hbpd+hozval;
		vtotal = clksel/(clkdiv*framerate*htotal);
		_vfpd = vtotal-vspw-vbpd-_lineval;
	}
#endif	
	/*calc new htotal and hfpd*/
	htotal = clksel/(clkdiv*framerate*vtotal);
	hfpd = htotal-hspw-hbpd-hozval;

	if ((last_set_framerate != framerate)&&(vsync_adj_state))
	{
		
		if (_vfpd>0)
		{
			vfpd=_vfpd;
			rVIDTCON0 = (0 << 24) |
				(((u32) vbpd-1) << 16) |
				(((u32) _vfpd-1) << 8) |
				(((u32) vspw-1) << 0);
			rVIDTCON1 = (0 << 24) |
				(((u32) hbpd-1) << 16) |
				(((u32) hfpd-1) << 8) |
				(((u32) hspw-1) << 0);

				printk(KERN_ERR "Set tft refresh rate to %ldHz(%dHz requested). (clksel=%ld,dclk=%ld).\n",clksel/(htotal*vtotal*clkdiv),framerate,clksel,clksel/clkdiv);
				if (lineval) lineval=_lineval;
				if (vfpd) vfpd=_vfpd;
		}
		else
		{
			printk(KERN_ERR "Setting tft refresh rate to %dHz failed. (clksel=%ld,dclk=%ld).\n",framerate,clksel,clksel/clkdiv);
		}
		last_set_framerate=framerate;
	}	

	
	return(clksel/(htotal*vtotal*clkdiv));
	
}


static void s3c2443_tft_initialize(struct lcd_screen_info *screeninfo, struct lcd_additional_info *addinfo)
{
	unsigned long int new_divider = 2 * addinfo->divider + 1;
	
	rVIDCON0 =  (new_divider << 6) | (1 << 5) | (1 << 4);
	rVIDCON1 =  ((u32)addinfo->invvclk << 7)|((u32)addinfo->invvline << 6)|((u32)addinfo->invvframe << 5)|((u32)addinfo->invvden << 4); 
	vfpd = screeninfo->screen.tft.vfpd;
	rVIDTCON0 = (0 << 24) |
			(((u32) screeninfo->screen.tft.vbpd) << 16) |
			(((u32) screeninfo->screen.tft.vfpd) << 8) |
			(((u32) screeninfo->screen.tft.vspw) << 0);
	rVIDTCON1 = (0 << 24) |
			(((u32) screeninfo->screen.tft.hbpd) << 16) |
			(((u32) screeninfo->screen.tft.hfpd) << 8) |
			(((u32) screeninfo->screen.tft.hspw) << 0);
	lineval = (screeninfo->y_res -1);
	rVIDTCON2 = (lineval << 11) | ((screeninfo->x_res -1) << 0);
	/* window control */
	rWINCON0 =  ((u32)addinfo->bswp << 17) | ((u32)addinfo->hwswp << 16) | (0 << 9) |            (5 << 2) |           (1 << 0);
	rWINCON1 =  ((u32)addinfo->bswp << 17) | ((u32)addinfo->hwswp << 16) | (0 << 9) | (1 << 6) | (5 << 2) | (0 << 1) |(1 << 0);
	/* alpha control */
	rVIDOSD1C = 0xfff000;
	/* RGB 565, no dithering */
	rDITHMODE = (0 << 5) | (1 << 3) | (0 << 1) | (0 << 0);

	s3c2443_set_framerate(60);
	
  
  /* Wait for end-of-frame from LCD controller before switching over to TFT controller */
  if ((rLCDCON1 & 0x00000001) != 0)  /* If LCD controller is enabled */
  {
    while ((rLCDCON1 & 0x0FFC0000) != 0) /* Wait until linecount is 0 */
    {
    }
    rLCDCON1 &= 0xFFFFFFFE;  /* Disable LCD controller */
  }
  rMISCCR |= (1<<28);      /* Switch output to TFT controller */
  rVIDCON0 |= 3;           /* Enable TFT controller output */
}


static int s3c2443_tft_is_initialized(void) 
{
	return ((rMISCCR &  (1<<28)) != 0);
}

static void s3c2443_tft_update_base(int layer, struct lcd_screen_info *screeninfo, struct fb_info *fbi, struct ttgfb_window *winptr )
{
	unsigned long flags;
#ifdef __BOOTLOADER__	
	struct ttgfb_window win        = { 0, 0, screeninfo->x_res , screeninfo->y_res };
	unsigned long       start      = FRAMEBUFFER_ADDRESS;
	unsigned long       winstart   = FRAMEBUFFER_ADDRESS;
	unsigned long       winstop    = FRAMEBUFFER_ADDRESS + screeninfo->y_res * screeninfo->x_res * BYTE_PER_PIXEL; 
	unsigned long       winoffsize = 0;
#endif

#ifdef __KERNEL__
	struct ttgfb_window win        =  *winptr;
	unsigned long       start      = fbi->fix.smem_start + fbi->var.yoffset * win.layer_width  * BYTE_PER_PIXEL;
	unsigned long       winstart   = start    + (win.src_top    * win.layer_width  + win.src_left) * BYTE_PER_PIXEL; 
	unsigned long       winstop    = winstart + ((win.height-1) * win.layer_width + win.width) * BYTE_PER_PIXEL;
	unsigned long       winoffsize = win.layer_width - win.width;
#endif
	local_irq_save(flags);
	
	if ((rVIDCON0 & 3) == 3)
		while ((rVIDCON1 >> 16) == 0)
			/* nothing */;
	
	if(layer == 0) {
		rVIDOSD1A = win.dst_left  << 11 | win.dst_top;
		rVIDOSD1B = (win.dst_left + win.width - 1) << 11 | (win.dst_top + win.height - 1);
		
		rVIDW01ADD0 = winstart;
		rVIDW01ADD1 = winstop;
		rVIDW01ADD2 = ((winoffsize * BYTE_PER_PIXEL) <<13) | ((win.width * BYTE_PER_PIXEL) << 0);
#if (DEBUG_GRAPHICS	& 3)
		cpubase0=fbi->screen_base;
#endif		
	} else {
		rVIDOSD0A = win.dst_left  << 11 | win.dst_top;
		rVIDOSD0B = (win.dst_left + win.width - 1) << 11 | (win.dst_top + win.height - 1);
		
		rVIDW00ADD0B0 = winstart;
		rVIDW00ADD1B0 = winstop;
		rVIDW00ADD2B0 = ((winoffsize * BYTE_PER_PIXEL) <<13) | ((win.width * BYTE_PER_PIXEL) << 0);
#if (DEBUG_GRAPHICS	& 3)
		cpubase1=fbi->screen_base;
#endif		
#ifdef __KERNEL__
		/* only use two different buffers when not panning */
		if (fbi->var.yoffset == 0)
		{
			winstart += screeninfo->y_res * screeninfo->x_res * BYTE_PER_PIXEL;
			winstop  += screeninfo->y_res * screeninfo->x_res * BYTE_PER_PIXEL;
		}
#endif
		rVIDW00ADD0B1 = winstart;
		rVIDW00ADD1B1 = winstop;
		rVIDW00ADD2B1 = ((winoffsize * BYTE_PER_PIXEL) <<13) | ((win.width * BYTE_PER_PIXEL) << 0);
	}
	
	local_irq_restore(flags);
}

static void s3c2443_tft_show_color_screen(unsigned color)
{
	// Show black screen
	rWIN0MAP = (1<<24)|color;
	rWIN1MAP = (1<<24)|color;
}

static void s3c2443_tft_show_normal_screen(void)
{
	// Show normal screen
	rWIN0MAP = (0<<24);
	rWIN1MAP = (0<<24);
}

static int s3c2443_tft_get_number_of_layers(void)
{
	return 2;
}

static unsigned s3c2443_tft_get_base(int layer)
{
	return (layer==0) ? rVIDW01ADD0 : rVIDW00ADD0B0;
}

static void s3c2443_tft_start_lcd(void)
{
	// Show black screen
	rWIN0MAP = (1<<24)|0x000000;
	rWIN1MAP = (1<<24)|0x000000;
	
	// Start display controller
	rVIDCON0 |= 3;

  // Show normal screen
  rWIN0MAP = (0<<24);
	rWIN1MAP = (0<<24);
}

static void s3c2443_tft_stop_lcd(void)
{
	rVIDCON0 &= ~3;
}

static void s3c2443_tft_set_colorkey(int enable, unsigned key)
{
	if (!enable) {
		rW1KEYCON0 = 0;
		rW1KEYCON1 = 0;
	} else {
		rW1KEYCON0 = (1 << 26) | (1 << 25) | (0x7 << 16) | (0x3 << 8) | (0x7 << 0);
		rW1KEYCON1 = key & 0x00ffffff;
	}
}

static void s3c2443_tft_set_alpha(unsigned alpha)
{
	rVIDOSD1C = alpha;
}

static const char *s3c2443_tft_get_id(void)
{
	return "s3c2443-tft-lcd";
}

/* Returns the actual set clock divider value. */
static unsigned short s3c2443_tft_set_clkval( struct lcd_screen_info *screeninfo, unsigned short clkval )
{
	unsigned long int regval;
	unsigned long int new_clkval = 1 + 2 * clkval;   /* new_clkval + 1 = (old_clkval + 1) * 2  */
	
	/* Is this a legal value? */
	if( clkval > 31 )
		clkval= 31;
	
	/* Set the CLK value only if the divider is within bounds. */
	regval=rVIDCON0 & 0xFFFFF03F;
	rVIDCON1 =regval | (((unsigned long int) new_clkval) << 6);	
	
	/* Return the actual divider calculated. */
	return clkval;
}

static int s3c2443_tft_ioctl(unsigned int cmd, unsigned long arg, struct fb_info *info)
{
	switch(cmd) {
	case TTGFB_IOCTL_BLEND_DISABLE:
	        resume_colorkey_defined = 0;
		s3c2443_tft_set_colorkey(0, 0);
		break;
	case TTGFB_IOCTL_BLEND_COLOR_KEY:
		s3c2443_tft_set_colorkey(1, arg);
		resume_colorkey = arg;
	        resume_colorkey_defined = 1;
		break;
	case TTGFB_IOCTL_SET_ALPHA:
		resume_alpha = arg;
		s3c2443_tft_set_alpha(arg);
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static void s3c2443_tft_resume(void)
{
	if (resume_colorkey_defined)
	{
		s3c2443_tft_set_colorkey(1, resume_colorkey);
		s3c2443_tft_set_alpha(resume_alpha);
	}
	else
	{
		s3c2443_tft_set_colorkey(0, 0);
	}
}

struct lcd_controller s3c2443_tft_controller = {
	.initialize           = s3c2443_tft_initialize,
	.is_initialized       = s3c2443_tft_is_initialized,
	.update_base          = s3c2443_tft_update_base,
	.show_color_screen    = s3c2443_tft_show_color_screen,
	.show_normal_screen   = s3c2443_tft_show_normal_screen,
	.get_number_of_layers = s3c2443_tft_get_number_of_layers,
	.get_base             = s3c2443_tft_get_base,
	.start_lcd            = s3c2443_tft_start_lcd,
	.stop_lcd             = s3c2443_tft_stop_lcd,
	.get_id               = s3c2443_tft_get_id,
	.set_clkval           = s3c2443_tft_set_clkval,
	.ioctl                = s3c2443_tft_ioctl,
	.resume               = s3c2443_tft_resume
};

#ifdef __KERNEL__
#ifdef CONFIG_VIDEO_S3C241X_CAMIF
/* Video overlaying support */
static int s3c2443_overlay_get_number_of_buffers(void)
{
	return 2;
}

static void s3c2443_overlay_set_buffer(int b)
{
	/* manual buffering, buffer b */
	rWINCON0 = ((rWINCON0 & 0xff3fffff) | (b << 23));
	
}

static void s3c2443_overlay_flip_buffer(void)
{
	/* auto buffering */
	rWINCON0 = ((rWINCON0 & 0xff3fffff)  | (1 << 22));
	/* flip */
	rCPUTRIGCON2 = (1<<0);
}

static void s3c2443_overlay_set_global_alpha(unsigned char alpha)
{
	unsigned alpha4  = alpha >> 4;
	unsigned alpha24 = (alpha4 << 20) | (alpha4 << 16) |(alpha4 << 12);
	
	s3c2443_tft_set_alpha(alpha24);
}

static void s3c2443_overlay_dump_regs(void)
{
	printk(KERN_DEBUG "VIDW00ADD0B0 = 0x%08X\n", rVIDW00ADD0B0);
	printk(KERN_DEBUG "VIDW00ADD0B1 = 0x%08X\n", rVIDW00ADD0B1);
	printk(KERN_DEBUG "VIDW01ADD0   = 0x%08X\n", rVIDW01ADD0);
	
	printk(KERN_DEBUG "VIDOSD1C     = 0x%08X\n", rVIDOSD1C);
	printk(KERN_DEBUG "W1KEYCON0    = 0x%08X\n", rW1KEYCON0);
	printk(KERN_DEBUG "W1KEYCON1    = 0x%08X\n", rW1KEYCON1);
}

struct video_overlay_ops s3c2443_tft_video_overlay = {
	.get_number_of_buffers  	= s3c2443_overlay_get_number_of_buffers,
	.set_buffer             	= s3c2443_overlay_set_buffer,
	.flip_buffer            	= s3c2443_overlay_flip_buffer,
	.set_color_key          	= s3c2443_tft_set_colorkey,
	.set_global_alpha       	= s3c2443_overlay_set_global_alpha,
	.dump_regs              	= s3c2443_overlay_dump_regs,
	.set_framerate          	= s3c2443_set_framerate,
	.input_sync             	= s3c2443_input_sync,
	.set_input_synchronisation	= s3c2443_set_input_synchronisation
	
};
#endif /* CONFIG_VIDEO_S3C241X_CAMIF */
#endif /* __KERNEL__ */
/* EOF */
